﻿
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;

using System.Reflection;
using System.Reflection.Emit;
using System.Windows.Forms;

using System.Data;
using System.Data.OleDb;
using vbSystemFunctions;
using VisTool;


namespace VisTool {
    /* Expression nodes have an Eval method that calculates formula values at run time, i.e. just before the visual components 
     * are shown. Errors during evaluation should simply throw an ordinary Exception with a message as parameter. The exception 
     * is caught by EvalProp, eventually the central CalculateProperties loop. They report the error and associate it with a formula
     * token through VisEngine.curTemplate, curTplProp and curExprToken. 
     * It is thus important that the Expression nodes set these global variables, anticipating errors.
     * A default expression, B Default C, catches errors in A, but doesn't report them. Instead it evalutes C and returns it as 
     * the expression value. In this way, the designer can supress expected errors.
     */

    public class RecordRef { // Record reference class, e.g.: me >- relMedType.
        // The formula terminates in the Record state. The result is a record reference.
        public Relationship relation;
        public System.Data.DataRow Row; //data row associated with the instance

        public RecordRef(System.Data.DataRow row, Relationship relation) {
            this.Row = row;
            this.relation = relation;
        }
    }

    public enum Operator {
        Unknown,
        Add, Sub, Mul, Div, IntDiv,
        BitAnd,
        StrConcat,
        Like,
        Mod,
        Exp, //exponential
        Neg,
        Neq,
        Eq, Lt, Le, Gt, Ge, Not, And, Or,
        Xor, //exclusive or,
        LJ,  //Left join
        IJ, //Inner join
        JO, //Join one
    }

    //--------------------------------------------------------------------
    public class Expression {
        public Expression() { }

        public virtual object Eval(VisComponent context, Template tpl) { return null; }
        // context: The control of the property being evaluated. Context is defined in CreateInstances and just passed on everywhere.
        // Only used in expressions such as Me, Form, Index and Last.
        // tpl: The control's template. Also passed on everywhere, but rarely used. Context.Bundle.Template would work.

        public virtual bool Assign(VisComponent context, Template tpl, object newValue, out object setValue) {
            // Must return non-null to avoid looking like an exception.
            // setValue is the value actually stored. It may have been transformed from Double to int, etc.
            setValue = null; return false;
        }

        public bool Init; // Only used for the root expression in a formula and for each expression in an expression list.
        // When true, the formula gives the initial value. It will be evaluated only at form open. User interaction may change the value later.
        // When false, the formula will be evaluated at every refresh.
        public Token token; // The token being evaluated. Used for error reporting and Studio's adjustment of for instance Width and Left.
    }

    //---------------------------------------------------------------------
    public class Assignment : Expression {
        public readonly Expression lhs;
        public readonly Expression rhs;
        public readonly TplProperty tplProp;

        public Assignment(TplProperty tplProp, Expression lhs, Expression rhs, Token t) {
            this.lhs = lhs;
            this.rhs = rhs;
            this.token = t;
            this.tplProp = tplProp;
        }

        public override object Eval(VisComponent context, Template tpl) { // Must return non-null to avoid looking like an exception.
            // Assigning to IndirectPrefixedField is for instance missing.
            VisEngine.curComponent = context; // Prepare for error messages
            VisEngine.curExprToken = token;
            object setValue;
            if (!lhs.Assign(context, tpl, rhs.Eval(context, tpl), out setValue))
                throw new Exception("Cannot assign to this.");
            return true;
        }
    }

    //---------------------------------------------------------------------
    public class ForLoop : Expression { // For ( ident = initExpr ; conditionExpr ; ident = stepExpr ) loopStatement
        public readonly Expression initExpr;
        public readonly Expression conditionExpr;
        public readonly Expression stepExpr;
        public readonly Expression loopStatement; // Null if no Else-part.
        public int level; // 0 for the outermost For loop.

        public ForLoop(Expression initExpr, Expression conditionExpr, Expression stepExpr, Expression loopStatement, int level, Token t) {
            this.initExpr = initExpr;
            this.conditionExpr = conditionExpr;
            this.stepExpr = stepExpr;
            this.loopStatement = loopStatement;
            this.level = level;
            this.token = t;
        }

        public override object Eval(VisComponent context, Template tpl) { // Must return non-null to avoid looking like an exception.
            int loopCount = 10000; // Break loop after 10,000 iterations. Test mode?
            VisEngine.curComponent = context; // Set current component being evaluated.
            VisEngine.loopVariables[this.level] = (int)this.initExpr.Eval(context, tpl); // Convert.ToInt32
            while (Convert.ToBoolean(conditionExpr.Eval(context, tpl))) {
                loopStatement.Eval(context, tpl);
                VisEngine.loopVariables[this.level] = (int)this.stepExpr.Eval(context, tpl);
                loopCount--;
                if (loopCount < 0) throw new Exception(token.val + " has looped 10,000 times. Loop interrupted");
            }
            return true;
        }
    }

    //---------------------------------------------------------------------
    public class LoopVariable : Expression {
        int level;

        public LoopVariable(int level) {
            this.level = level;
        }

        public override object Eval(VisComponent context, Template tpl) {
            //always get the up-to-date variable
            return VisEngine.loopVariables[this.level];
        }
    }

    //---------------------------------------------------------------------
    public class IfThen : Expression { // If (ifExpr) { thenStatement } [ Else {elseStatement} ]
        public readonly Expression ifExpr;
        public readonly Expression thenStatement;
        public readonly Expression elseStatement; // Null if no Else-part.

        public IfThen(Expression ifExpr, Expression thenStatement, Expression elseStatement, Token t) {
            this.ifExpr = ifExpr;
            this.thenStatement = thenStatement;
            this.elseStatement = elseStatement;
            this.token = t;
        }

        public override object Eval(VisComponent context, Template tpl) {
            VisEngine.curComponent = context; //setting current component being evaluated.
            object ifValue = this.ifExpr.Eval(context, tpl);
            if (Convert.ToBoolean(ifValue)) thenStatement.Eval(context, tpl);
            else if (elseStatement != null) elseStatement.Eval(context, tpl);
            return true; // Must return non-null to avoid looking like an exception.
        }
    }

    //---------------------------------------------------------------------
    public class ReturnExpr : Expression { // Return from event handler.
        public ReturnExpr(Token t) { this.token = t; }

        public override object Eval(VisComponent context, Template tpl) {
            throw new LoggedException(); // Terminates ExecuteHandler. 
        }
    }

    //---------------------------------------------------------------------
    public class ExpressionList : Expression {
        public List<Expression> expressions = new List<Expression>(); // The syntax ensures that there is at least one element in the list
        public ExpressionList() { }
        public void Add(Expression e) {
            expressions.Add(e);
        }
        public override object Eval(VisComponent context, Template tpl) {
            // Never used because it cannot handle Init. To do that it needs access to old values. EvalListProp does it. ??
            VisEngine.curComponent = context; //setting current component being evaluated
            object[] o = new object[expressions.Count]; // Will end up with a value for each element
            for (int i = 0; i < o.Length; i++) o[i] = expressions[i].Eval(context, tpl);
            return o;
        }
    }

    //---------------------------------------------------------------------
    public class StatementList : Expression {
        public List<Expression> expressions = new List<Expression>(); // The syntax ensures that there is at least one element in the list
        TplProperty tplProp; // Used for error messages when a statement throws an event.
        public StatementList(TplProperty p) { this.tplProp = p; }

        public void Add(Expression e) { expressions.Add(e); }

        public override object Eval(VisComponent context, Template tpl) {
            VisEngine.curComponent = context; //setting current component being evaluated
            try {
                for (int i = 0; i < expressions.Count; i++)
                    if (expressions[i] != null)
                        expressions[i].Eval(context, tpl);
            }
            catch (LoggedException) {
                // Abort the statement list. Example: The statement list ends with Refresh, which will erase the error log.
                throw new LoggedException(); // Statement lists may be nested.
            }
            // Other exceptions are caught by ExecuteHandler in order to abort the entire event handler.
            return true; // Must return non-null to avoid looking like an exception.
        }
    }

    //---------------------------------------------------------------------
    public class Constant<T> : Expression {
        public T Value;
        public Constant(T value) {
            this.Value = value;
        }
        public Constant(T value, Token t) {
            this.Value = value; this.token = t;
        }

        public override object Eval(VisComponent context, Template tpl) {
            return Value;
        }
    }

    //---------------------------------------------------------------------
    public class SystemVariableExpr : Expression {
        FieldInfo fieldInfo = null;

        public SystemVariableExpr(FieldInfo fieldInfo) {
            this.fieldInfo = fieldInfo;
        }

        public override bool Assign(VisComponent context, Template tpl, object r, out object setValue) {
            setValue = null;
            if (fieldInfo == null) return false;
            if (fieldInfo.FieldType == typeof(float))
                fieldInfo.SetValue(null, (float)(double)r);
            else 
            fieldInfo.SetValue(null, r);
            setValue = r; return true;
        }

        public override object Eval(VisComponent context, Template tpl) {
            //always get the up-to-date variable
            if (fieldInfo != null)
                return fieldInfo.GetValue(null);
            return null;
        }
    }

    //---------------------------------------------------------------------
    public class EnumExpr : Expression {
        object value = null;
        public EnumExpr() {
        }
        public EnumExpr(Type Type, string propName) {
            PropertyInfo propInfo = Type.GetProperty(propName, System.Reflection.BindingFlags.IgnoreCase | System.Reflection.BindingFlags.Public | System.Reflection.BindingFlags.Static);
            FieldInfo fieldInfo = Type.GetField(propName, System.Reflection.BindingFlags.IgnoreCase | System.Reflection.BindingFlags.Public | System.Reflection.BindingFlags.Static);

            if (propInfo != null)
                value = propInfo.GetValue(null, null);
            else if (fieldInfo != null)
                value = fieldInfo.GetValue(null);
        }

        public EnumExpr(object value) {
            this.value = value;
        }

        public override object Eval(VisComponent context, Template tpl) {
            return value;
        }
    }

    //---------------------------------------------------------------------
    public class FormsExpr : Expression { // Forms ! FormName [(cloneID)] 
        public string FormName;
        Expression cloneIDexpr;
        public FormsExpr(string formName, Expression cloneIDexpr, Token t) {
            this.FormName = formName;
            this.cloneIDexpr = cloneIDexpr;
            this.token = t;
        }

        public override object Eval(VisComponent context, Template tpl) { // Exception if the form doesn't exist as a file.
            // Returns a VisForm (VisComponent) if the form is open and has at least one clone. Returns filename is the form is closed.
            VisEngine.curComponent = context; // Prepare for errors
            VisEngine.curExprToken = token;
            int cloneID = 0; // Default cloneID.
            if (cloneIDexpr != null) {
                object clone = (int)cloneIDexpr.Eval(context, tpl);
                if (clone is int) cloneID = (int)clone;
            }
            int actualCloneID = cloneID;
            FormTemplate frmTpl = VisEngine.GetFormTemplate(FormName, ref actualCloneID);
            if (actualCloneID < 0) 
                throw new Exception("Form " + FormName + " doesn't exist as a file");
            if( frmTpl == null) return FormName; // The form has no open clones. FormName helps the caller open the form.
            if (cloneID != actualCloneID) {
                throw new Exception("Clone " + cloneID + " of " + FormName + " doesn't exist"); // return null; // The wanted clone doesn't exist.
            }
            return frmTpl.RowBundle.Instances[0].cmp as VisForm;  // Return the first form instance in the clone's bundle.
        }
    }

    //---------------------------------------------------------------------
    public class ParamExpr : Expression { // Param[index]
        Expression indexExpr;
        Expression leftCmp;
        public ParamExpr(Expression left, Expression index, Token t) {
            this.leftCmp = left;
            this.indexExpr = index;
            this.token = t;
        }

        public override object Eval(VisComponent context, Template tpl) {
            VisEngine.curComponent = context; // Prepare for errors.
            VisEngine.curExprToken = token;
            VisComponent c = leftCmp == null ? context : leftCmp.Eval(context, tpl) as VisComponent;
            int index = indexExpr == null ? 0 : (int)indexExpr.Eval(context, tpl);
            FormTemplate frm = c.tpl.FormTpl;
            if (frm.parameters == null || frm.parameters.Length == 0)
                return null;
            //throw new Exception("The form has no parameters");
            if (index < 0 || index >= frm.parameters.Length)
                throw new Exception("Index " + index.ToString() + " is out of range 0:" + (frm.parameters.Length - 1).ToString());
            return frm.parameters[index];
        }

        public override bool Assign(VisComponent context, Template tpl, object r, out object setValue) {
            VisEngine.curComponent = context; // Prepare for errors.
            VisEngine.curExprToken = token;
            VisComponent c = leftCmp == null ? context : leftCmp.Eval(context, tpl) as VisComponent;
            int index = indexExpr == null ? 0 : (int)indexExpr.Eval(context, tpl);
            FormTemplate frm = c.tpl.FormTpl;
            if (frm.parameters == null || frm.parameters.Length == 0)
                throw new Exception("The form has no parameters");
            if (index < 0 || index >= frm.parameters.Length)
                throw new Exception("Index " + index.ToString() + " is out of range 0:" + (frm.parameters.Length - 1).ToString());
            VisComponent caller = frm.callingContext;
            bool ok = frm.exprParameters[index].Assign(caller, caller.tpl, r, out setValue);
            if (ok) return true;
            throw new Exception("Cannot assign to this parameter");
        }
    }

    //---------------------------------------------------------------------
    public class TemplateRefExpr : Expression { // Examples: Me ! tplB, tplName[i] ! tplB
        Expression leftNode;    // In A ! B, A is the leftNode. If leftNode is null, use Me.
        public Template targetTpl; // targetTpl is B. May be in another clone ??

        public TemplateRefExpr(Expression e, Template tpl, Token t) {
            this.leftNode = e;
            this.targetTpl = tpl;
            this.token = t;
        }

        public override object Eval(VisComponent context, Template tpl) {
            VisEngine.curComponent = context; // Prepare for errors.
            VisEngine.curExprToken = token;
            VisComponent leftCtrl = leftNode == null ? context : (VisComponent)leftNode.Eval(context, tpl);

            if (!targetTpl.Updated) // The main template may be updated, but the secondary not.
                VisEngine.CreateInstances(targetTpl);

            if (targetTpl.RowTpl.RowBundle.Instances.Count == 0) return null; // No target cmp.
            int index = 0;
            if (leftCtrl.tpl.RowTpl == targetTpl.RowTpl)  // The same data source. Use the same index
                index = leftCtrl.Index();
            return targetTpl.RowBundle.Instances[index].cmp;
        }
    }

    //---------------------------------------------------------------------
    public class ForeignTplOrPropRef : Expression { // Examples: Forms ! frmName ! tplName, Forms ! frmName(1) [0] ! tplName ! Top
        // The target may be a template or a property in a template. 
        // When leftNode is a Form, the target name is either a template name or a Form property name. Cannot be both.
        Expression leftNode; // In A ! B, A is the leftNode and B is the target.
        string targetName; // A template name or property name.

        public ForeignTplOrPropRef(Expression e, string tplName, Token t) {
            this.leftNode = e;
            this.targetName = tplName;
            this.token = t;
        }

        public override object Eval(VisComponent context, Template tpl) {
            VisEngine.curComponent = context; // Prepare for errors.
            VisEngine.curExprToken = token;
            VisComponent leftCtrl = leftNode == null ? context : leftNode.Eval(context, tpl) as VisComponent;
            // LeftCtrl willl be in another form than context. The target is in this foreign form.
            if (leftCtrl == null) return null; // The form isn't open or the target isn't a component.

            bool exists;
            object result = VisEngine.GetValue(leftCtrl, targetName, out exists); // GetValue doesn't report circular references, 
            // but they shouldn't occur here because it is an open foreign form with updated properties.
            if (exists) return result; // May be null.
            FormTemplate targetForm = leftCtrl.tpl.FormTpl;
            Template targetTpl;
            if (!targetForm.Templates.TryGetValue(targetName, out targetTpl) || targetTpl.Ignore)
                throw new Exception("Cannot locate " + targetName + " in form " + targetForm.Name);
            if (!targetTpl.Updated)
                VisEngine.CreateInstances(targetTpl);

            if (targetTpl.RowBundle.Instances.Count == 0) return null; // No target cmp.
            int index = 0;
            if (leftCtrl.tpl.RowTpl == targetTpl.RowTpl)  // The same data source. Use the same index
                index = leftCtrl.Index();
            return targetTpl.RowBundle.Instances[index].cmp;
        }

        public override bool Assign(VisComponent context, Template tpl, object newVal, out object setValue) {
            VisEngine.curComponent = context; // Prepare for errors.
            VisEngine.curExprToken = this.token; // Prepare for errors
            VisComponent leftCtrl = leftNode == null ? context : leftNode.Eval(context, tpl) as VisComponent;// The target component.
            setValue = null;
            // LeftCtrl willl be in another form than context. The target is in this foreign form.
            if (leftCtrl == null) return false; // The form isn't open or the target isn't a component.

            Template targetTpl = leftCtrl.tpl;
            TplProperty tplProp; PropInfo propInfo;
            tplProp = TplProperty.GetTemplateProperty(tpl, targetName, out propInfo);
            if (tplProp == null) 
                throw new Exception("Cannot assign to " + targetTpl.Name + "!" + targetName + ", " + targetName + " doesn't exist");
            if (propInfo is ValueInfo) {
                (propInfo as ValueInfo).PropertySetter.SetValue(targetName, newVal, null);
                setValue = newVal; return true;
            }
            if (propInfo == null) { // Designer property.
                leftCtrl.designerPropValues[tplProp.cmpPropIndex].Value = newVal;
                setValue = newVal; return true;
            }
            // It is an event property or an function property.
            throw new Exception("Cannot assign to " + targetTpl.Name + "!" + targetName);
            throw new LoggedException();
        }
    }

    //---------------------------------------------------------------------
    public class FindExpr : Expression { // Example: Find targetTpl On targetTpl.keyT = Me.keyS And . . . Order By Me.startTime
        public TplProperty tplProp; // Holds the FindExpr.
        public Template targetTpl;
        public List<Expression> toKeyExpr, fromKeyExpr; // Extracted from the On clause.
        public List<Expression> orderBy = null; // Extracted from the Order By clause. Null when no OrderBy.
        public List<int> asc; // Whether the orderKey is ascending.

        public FindExpr(TplProperty tplProp, Template targetTpl, List<Expression> toKeyExpr, List<Expression> fromKeyExpr,
            List<Expression> orderExpr, List<int> asc, Token t) {
            this.tplProp = tplProp; this.targetTpl = targetTpl;
            this.toKeyExpr = toKeyExpr; this.fromKeyExpr = fromKeyExpr;
            this.orderBy = orderExpr; this.asc = asc; this.token = t;
        }

        public override object Eval(VisComponent context, Template tpl) { // Finds the bundle that relates to the target component, including thei index in the bundle.
            // To find the index, it is necessary to calculate and sort all the bundles. To quickly find the target component, we make a dictionary of them.
            int targetIndex; // Index in the target template's findProperties. Also the index in the target component's childBundles. 
            CmpProp sourceCmpProp;

            if (tplProp.FindBundles == null) { // Create and populate target dictionary.
                if (!targetTpl.Updated) VisEngine.CreateInstances(targetTpl);
                targetIndex = targetTpl.findProperties.IndexOf(tplProp);
                if (targetIndex < 0) {
                    ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp, "Error: Compiled targets and Find formulas don't match", token, context.Index());
                    return null;
                } // Update the list of find properties that refer to this target.

                tplProp.FindBundles = new Dictionary<List<object>, Bundle>(targetTpl.RowBundle.Instances.Count, new CmpEqualityComparer());
                for (int ix = 0; ix < targetTpl.RowBundle.Instances.Count; ix++) { // Scan all target components and put them into a dictionary.
                    VisComponent targetCmp = targetTpl.RowBundle.Instances[ix].cmp;
                    BundleRef old = VisEngine.leftBundleRef; VisEngine.leftBundleRef = targetCmp.bundleRef();
                    if (targetCmp.childBundles == null)  // Make sure all target components get an array.
                        targetCmp.childBundles = new Bundle[targetTpl.findProperties.Count];
                    List<object> toKey = new List<object>(toKeyExpr.Count);
                    for (int i = 0; i < toKeyExpr.Count; i++) { // Compute target key
                        toKey.Add(toKeyExpr[i].Eval(targetCmp, tpl));
                    }
                    VisEngine.leftBundleRef = old;
                    if (toKeyExpr.Count != 0 && tplProp.FindBundles.ContainsKey(toKey)) {
                        ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp, "Target key is not unique ", token, context.Index());
                    } else {
                        Bundle bdl = new Bundle(16, targetCmp, tplProp);
                        if (toKeyExpr.Count != 0 || ix == 0)
                            tplProp.FindBundles.Add(toKey, bdl); // If empty key, add only the first target component.
                        targetCmp.childBundles[targetIndex] = bdl; // All target components get an array of empty childBundles.
                    }
                }

                foreach (CmpProp cmpProp in tpl.RowBundle.Instances) { // Add all source components to a bundle.
                    VisComponent sourceCmp = cmpProp.cmp;
                    List<object> fromKey = new List<object>(fromKeyExpr.Count);
                    Bundle bundle;
                    BundleRef old = VisEngine.leftBundleRef; VisEngine.leftBundleRef = sourceCmp.bundleRef();
                    for (int i = 0; i < fromKeyExpr.Count; i++) {
                        fromKey.Add(fromKeyExpr[i].Eval(sourceCmp, tpl));
                    }
                    VisEngine.leftBundleRef = old;
                    if (!tplProp.FindBundles.TryGetValue(fromKey, out bundle)) { // Target is not in the dictionary.
                        continue; // Empty keys will match the single target in the dictionary.
                    }
                    sourceCmpProp = (tplProp is TplPropDesigner ? sourceCmp.designerPropValues[tplProp.cmpPropIndex] : sourceCmp.propertyValues[tplProp.cmpPropIndex]);
                    bundle.Instances.Add(sourceCmpProp);

                    object[] orderKeys = null;
                    if (orderBy != null) {
                        orderKeys = new object[orderBy.Count];
                        for (int i = 0; i < orderBy.Count; i++) {
                            object tmp = orderBy[i].Eval(sourceCmp, tpl);
                            orderKeys[i] = tmp;
                        }
                    }
                    BundleRef source = sourceCmpProp.Value as BundleRef;
                    object tag = null;
                    if (source != null) tag = source.tag; // May contain dialog values.
                    source = new BundleRef(bundle, bundle.Instances.Count - 1, orderKeys);
                    source.tag = tag;
                    sourceCmpProp.Value = source;
                }
                
                if (orderBy != null)
                foreach (var item in tplProp.FindBundles) { // Sort each bundle in the dictionary.
                    Bundle bdl = item.Value;
                    Sort(bdl);
                }
            }
            sourceCmpProp = (tplProp is TplPropDesigner ? context.designerPropValues[tplProp.cmpPropIndex] : context.propertyValues[tplProp.cmpPropIndex]);
            return sourceCmpProp.Value;
        }

        public void Sort(Bundle bdl) {
            FindExpr exp = bdl.tplProp.expr as FindExpr;
            if (exp.orderBy == null) return;
            bdl.Instances.Sort(new CmpComparer(exp.asc));
            for (int i = 0; i < bdl.Instances.Count; i++) // Adjust indexes according to the new sequence.
                (bdl.Instances[i].Value as BundleRef).index = i;
        }
    }

    //--------------------------------------------------------
    public class CmpEqualityComparer : IEqualityComparer<List<object>> { // Compares a list of target property values with a list of source property values
        public bool Equals(List<dynamic> c1, List<dynamic> c2) {
            for (int i = 0; i < c1.Count; i++)
                if (c1[i] != c2[i]) return false;
            return true;
        }

        public int GetHashCode(List<object> c) {
            int h = 0;
            for (int i = 0; i < c.Count; i++) 
                if (c[i] != null) h ^= c[i].GetHashCode();
            return h;
        }
    }

    //--------------------------------------------------------
    public class CmpComparer : IComparer<CmpProp> { // Compares two BundleRefs for Order By.
        public CmpComparer(List<int> asc) { this.asc = asc; }
        List<int> asc; // 1 for ascending, -1 for descending.

        public int Compare(CmpProp xProp, CmpProp yProp) {
            BundleRef tmpx = xProp.Value as BundleRef;
            object[] xOrder = tmpx.orderKeys;
            BundleRef tmpy = yProp.Value as BundleRef;
            object[] yOrder = tmpy.orderKeys;
            for (int i = 0; i < xOrder.Length; i++) {
                dynamic x = xOrder[i]; dynamic y = yOrder[i]; // Cover all integer types, etc.
                try {
                    int asc = this.asc[i];
                    if (x == y) continue;
                    return asc * x.CompareTo(y);

                    //Debug.Print(x.ToString() + ", " + y.ToString() + ", " + ((int)x).CompareTo((int)y));
                    //if (x is int) {
                    //    if ((int)x == (int)y) continue;
                    //    return asc * ((int)x).CompareTo((int)y);
                    //}
                    //if (x is Double) {
                    //    if ((Double)x == (Double)y) continue;
                    //    return asc * ((Double)x).CompareTo((Double)y);
                    //}
                    //if (x is string) {
                    //    if ((string)x == (string)y) continue;
                    //    return ((string)x).CompareTo((string)y);
                    //}
                    //if (x is DateTime) {
                    //    if ((DateTime)x == (DateTime)y) continue;
                    //    return asc * ((DateTime)x).CompareTo((DateTime)y);
                    //}
                    //// Other types or null match anything, but warn:
                    //ErrorLogging.Log(ErrorMode.RunTime, Priority.Warning, VisEngine.curTemplate, VisEngine.curTplProp,
                    //    "Cannot compare type " + x.GetType().ToString() + " and " + y.GetType().ToString(), VisEngine.curTplProp.startChain, 0); 
                }
                catch { // Other types or null match anything, but warn:
                    throw new Exception("Cannot compare type " + x.GetType().ToString() + " and " + y.GetType().ToString());
                }
            }
            return 0;
        }
    }

    //---------------------------------------------------------------------
    public class BundleRefExpr : Expression { // Source: tplA !findProp ! Bundle !Index, . . . !Last
        public readonly string dot;
        public readonly string fieldOrProp;
        public Expression exp; // Computes a property reference

        public BundleRefExpr(Expression e, string dot, string fieldOrProp, Token t) {
            this.dot = dot;
            this.fieldOrProp = fieldOrProp;
            this.token = t;
            exp = e;
        }

        public override object Eval(VisComponent context, Template tpl) {
            BundleRef target = this.exp.Eval(context, tpl) as BundleRef;
            if (target == null) return null;

            if (fieldOrProp.ToUpper() == "INDEX") return target.index;
            if (fieldOrProp.ToUpper() == "LAST") return target.bundle.Instances.Count -1;
            return target.bundle; // May be followed by [ ... ]
        }
    }

    //---------------------------------------------------------------------
    public class MeExpr : Expression {
        public MeExpr(Token t) { this.token = t; }

        public override object Eval(VisComponent context, Template tpl) {
            return context;
        }
    }

    //---------------------------------------------------------------------
    public class FrmCtrlExpr : Expression { // Walks to the current form control
        public FrmCtrlExpr(Token t) { this.token = t; }

        public override object Eval(VisComponent context, Template tpl) {
            return context.Form();
        }
    }

    //---------------------------------------------------------------------
    public class MainExpr : Expression { // Main !Top
        public MainExpr(Token t) { this.token = t; }

        public override object Eval(VisComponent context, Template tpl) {
            if (context.tpl == context.tpl.RowTpl) return context;
            // Return the main component with the same index as Me.
            int index = context.Index();
            return context.tpl.RowTpl.RowBundle.Instances[index].cmp;
        }
    }

    //---------------------------------------------------------------------
    public class NameExpr : Expression { // Source: Name, e.g. Form ! Name
        public Expression leftNode;
        public NameExpr(Expression e, Token t) {
            this.leftNode = e;
            this.token = t;
        }
        public override object Eval(VisComponent context, Template tpl) {
            VisEngine.curComponent = context; //Setting the current component
            VisEngine.curExprToken = token;
            if (leftNode != null)
                return (this.leftNode.Eval(context, tpl) as VisComponent).tpl.Name;
            else return tpl.Name;
        }
    }

    //---------------------------------------------------------------------
    public class IndexExpr : Expression { // LeftNode is bundleRef: tplA !findprop ! Bundle !Index, target !Bundle(source !findProp) !Index will be 0.
        // LeftNode is cmp: Index, tplA !Index, tplA !propB !Index {Indirect ref to component}, tplA !findProp !Index.
        // Inside [ ]: cmp[index+1], bundleRef[index+1]. Handled by VisEngine.leftBundleRef being != null. 
        public Expression leftNode;
        public IndexExpr(Expression e, Token t) {
            this.leftNode = e;
            this.token = t;
        }
        public override object Eval(VisComponent context, Template tpl) {
            VisEngine.curComponent = context; // Pepare for errors.
            VisEngine.curExprToken = token;
            if (VisEngine.leftBundleRef != null)  // Index without Me and inside Subscript
                return VisEngine.leftBundleRef.index;

            object left = leftNode == null ? context : this.leftNode.Eval(context, tpl);
            if (left != null) {
                BundleRef leftRef = left as BundleRef;
                if (leftRef != null)
                    return leftRef.index; // tplA !findProp !Bundle !Index, target !Bundle(source !findProp) !Index

                VisComponent comp = left as VisComponent; // tplA !Index, tplA !propB !Index, tplA !findProp !Index
                if (comp == null || comp.tpl.RowBundle == null) return -1; // Empty bundle or no bundle.
                return comp.bundleRef().index;
            }
            if (context.tpl.RowBundle == null)
                return -1; // No rows. Possible??
            return context.bundleRef().index;
        }
    }

    //---------------------------------------------------------------------
    public class LastExpr : Expression { // LeftNode is bundleRef: tplA !findprop ! Bundle !Last, target !Bundle(source !findProp) !Last.
        // LeftNode is cmp: Last, tplA !Last, tplA !propB !last {Indirect ref to component}, tplA !findProp !Index.
        // Inside [ ]: cmp[Last], bundleRef[Last]. Handled by VisEngine.leftBundleRef being != null. 
        public Expression leftNode;
        public LastExpr(Expression e, Token t) {
            this.leftNode = e;
            this.token = t;
        }
        public override object Eval(VisComponent context, Template tpl) {
            VisEngine.curComponent = context; // Prepare for errors.
            VisEngine.curExprToken = token;
            if (leftNode != null) {
                object left = this.leftNode.Eval(context, tpl);
                BundleRef leftRef = left as BundleRef;
                if (leftRef != null) 
                    return leftRef.bundle.Instances.Count - 1; //  tplA !findProp !Bundle !Last, target !Bundle(source !findProp) !Last

                VisComponent comp = left as VisComponent; // tplA !Last, tplA !propB !Last, tplA !findProp !Last
                if (comp == null || comp.tpl.RowBundle == null) return -1; // Empty bundle or no bundle..
                return comp.tpl.RowBundle.Instances.Count - 1;  
            } 
            if (VisEngine.leftBundleRef == null) { // Last without Me and not inside Subscript
                if (context.tpl.RowBundle == null) 
                    return -1; // Default bundle of one.Possible??
                return context.tpl.RowBundle.Instances.Count - 1;
            } else // Last without me inside subscript: tplA[ Last ]
                return VisEngine.leftBundleRef.bundle.Instances.Count - 1;
        }
    }

    //---------------------------------------------------------------------
    public class DefaultExpr : Expression {
        public readonly Expression eLeft, eRight;

        public DefaultExpr(Expression e1, Expression e2, Token t) {
            this.eLeft = e1;
            this.eRight = e2;
            this.token = t;
        }

        public override object Eval(VisComponent context, Template tpl) {
            VisEngine.curComponent = context; // Prepare for errors
            try {
                object vLeft = eLeft.Eval(context, tpl);
                if (vLeft != null) return vLeft; // Return
            }
            catch (Exception) { } // Treat any exception as null
            VisEngine.curExprToken = this.token; // Prepare for errors
            if (eRight == null) return null; // Default in a statement list may have eRight == null.
            return eRight.Eval(context, tpl);
        }

        public override bool Assign(VisComponent context, Template tpl, object r, out object setValue) {
            VisEngine.curComponent = context; // Prepare for errors
            try {
                return eLeft.Assign(context, tpl, r, out setValue);
            }
            catch (Exception) { } // Ignore any exception
            VisEngine.curExprToken = this.token; // Prepare for errors
            return eRight.Assign(context, tpl, r, out setValue);
        }
    }

    //---------------------------------------------------------------------
    public class CondExpr : Expression {
        public Expression condition;
        public Expression expTrue;
        public Expression expFalse;

        public CondExpr(Expression econdition, Expression etrue, Expression efalse, Token t) {
            this.condition = econdition;
            this.expTrue = etrue;
            this.expFalse = efalse;
            this.token = t;
        }

        public override object Eval(VisComponent context, Template tpl) {
            VisEngine.curComponent = context; // Prepare for errors
            bool b = false;
            object dyna = this.condition.Eval(context, tpl);
            VisEngine.curExprToken = token;
            b = Convert.ToBoolean(dyna);

            if (b) return expTrue.Eval(context, tpl);
            return expFalse.Eval(context, tpl);
        }

        public override bool Assign(VisComponent context, Template tpl, object r, out object setValue) {
            VisEngine.curComponent = context; // Prepare for errors
            bool b = false;
            object dyna = this.condition.Eval(context, tpl);
            VisEngine.curExprToken = token;
            b = Convert.ToBoolean(dyna);

            if (b) return expTrue.Assign(context, tpl, r, out setValue);
            return expFalse.Assign(context, tpl, r, out setValue);
        }
    }

    //---------------------------------------------------------------------
    public class BinOp : Expression {
        public readonly Operator op;
        public readonly Expression eLeft, eRight;

        public BinOp(Operator op, Expression e1, Expression e2, Token t) {
            this.op = op;
            this.eLeft = e1;
            this.eRight = e2;
            this.token = t;
        }

        public override object Eval(VisComponent context, Template tpl) {
            VisEngine.curComponent = context; //Setting the current component
            object vLeft = eLeft.Eval(context, tpl);
            object vRight = null;
            if (op == Operator.And || op == Operator.Or) vRight = false;
            else vRight = eRight.Eval(context, tpl);
            // Evaluate right part of And and Or only if necessary. Primarily to avoid exceptions in the right-hand part.
            VisEngine.curExprToken = this.token; // Prepare for errors

            TypeCode typeLeft = TypeCode.Object; // Will become the result type of left-hand expression.
            TypeCode typeRight = TypeCode.Object; // Will become the result type of right-hand expression.

            if (vLeft is Int32 || vLeft is Int16)
                typeLeft = TypeCode.Int32;
            else if (vLeft is double || vLeft is float)
                typeLeft = TypeCode.Double;
            else if (vLeft is string)
                typeLeft = TypeCode.String;
            else if (vLeft is DateTime)
                typeLeft = TypeCode.DateTime;
            else if (vLeft is bool)
                typeLeft = TypeCode.Boolean;

            if (vRight is Int32 || vRight is Int16)
                typeRight = TypeCode.Int32;
            else if (vRight is double || vRight is float)
                typeRight = TypeCode.Double;
            else if (vRight is string)
                typeRight = TypeCode.String;
            else if (vRight is DateTime)
                typeRight = TypeCode.DateTime;
            else if (vRight is bool)
                typeRight = TypeCode.Boolean;

            try {
                switch (op) {
                case Operator.Add:
                    if (typeLeft == TypeCode.String || typeRight == TypeCode.String) // Accept "+" for string concatenation
                        return (vLeft == null ? "" : vLeft.ToString()) + (vRight == null ? "" : vRight.ToString());
                    if (vLeft == null || vRight == null) return null;
                    if (typeLeft == TypeCode.DateTime)
                        return Convert.ToDateTime(vLeft).AddDays(Convert.ToDouble(vRight));
                    if (typeRight == TypeCode.DateTime)
                        return Convert.ToDateTime(vRight).AddDays(Convert.ToDouble(vLeft));
                    if (typeLeft == TypeCode.Double || typeRight == TypeCode.Double)
                        return Convert.ToDouble(vLeft) + Convert.ToDouble(vRight);
                    if (typeLeft == TypeCode.Int32 || typeRight == TypeCode.Int32)
                        return Convert.ToInt32(vLeft) + Convert.ToInt32(vRight);
                    throw new Exception("Cannot combine these types");

                case Operator.Sub:
                    if (vLeft == null || vRight == null) return null;
                    if (typeLeft == TypeCode.DateTime) {
                        if (typeRight == TypeCode.DateTime)
                            return (Convert.ToDateTime(vLeft) - Convert.ToDateTime(vRight)).TotalDays;
                        return Convert.ToDateTime(vLeft).AddDays(-Convert.ToDouble(vRight));
                    }
                    if (typeLeft == TypeCode.Double || typeRight == TypeCode.Double)
                        return Convert.ToDouble(vLeft) - Convert.ToDouble(vRight);
                    if (typeLeft == TypeCode.Int32 || typeRight == TypeCode.Int32)
                        return Convert.ToInt32(vLeft) - Convert.ToInt32(vRight);
                    throw new Exception("Cannot combine these types");

                case Operator.Exp:
                    if (vLeft == null || vRight == null) return null;
                    return Math.Pow(Convert.ToDouble(vLeft), Convert.ToDouble(vRight));

                case Operator.Mul:
                    if (vLeft == null || vRight == null) return null;
                    if (typeLeft == TypeCode.Int32 && typeRight == TypeCode.Int32)
                        return Convert.ToInt32(vLeft) * Convert.ToInt32(vRight);
                    return Convert.ToDouble(vLeft) * Convert.ToDouble(vRight);

                case Operator.And:
                    if (vLeft == null || vRight == null) return null;
                    bool left = Convert.ToBoolean(vLeft);
                    if (!left) return false; // Don't evaluate the right part in this case.
                    return Convert.ToBoolean(eRight.Eval(context, tpl));

                case Operator.Or:
                    if (vLeft == null || vRight == null) return null;
                    bool left2 = Convert.ToBoolean(vLeft);
                    if (left2) return true; // Don't evaluate the right part in this case.
                    return Convert.ToBoolean(eRight.Eval(context, tpl));

                case Operator.Div:
                    if (vLeft == null || vRight == null) return null;
                    if (Convert.ToDouble(vRight) == 0) {
                        throw new Exception("Trying to divide by zero");
                    }
                    return Convert.ToDouble(vLeft) / Convert.ToDouble(vRight); // Always float in VB

                case Operator.IntDiv:
                    if (vLeft == null || vRight == null) return null;
                    return Convert.ToInt32(vLeft) / Convert.ToInt32(vRight); // Gives the VB int divide result

                case Operator.Mod:
                    if (vLeft == null || vRight == null) return null;
                    return Convert.ToInt32(vLeft) % Convert.ToInt32(vRight);

                case Operator.Eq: // A = null may be true or false, not null.
                    if (vLeft == null || vRight == null) return vLeft == vRight;
                    if (typeLeft == TypeCode.Double || typeRight == TypeCode.Double)
                        return Convert.ToDouble(vLeft) == Convert.ToDouble(vRight);
                    if (typeLeft == TypeCode.Int32 || typeRight == TypeCode.Int32)
                        return Convert.ToInt32(vLeft) == Convert.ToInt32(vRight);
                    if (typeLeft == TypeCode.DateTime || typeRight == TypeCode.DateTime)
                        return Convert.ToDateTime(vLeft) == Convert.ToDateTime(vRight);
                    if (typeLeft == TypeCode.String && typeRight == TypeCode.String)
                        return string.Compare(vLeft.ToString(), vRight.ToString(), true) == 0; // Ignore case
                    BundleRef bdlLeft = vLeft as BundleRef;
                    BundleRef bdlRight = vRight as BundleRef;
                    if (bdlLeft != null) 
                        vLeft = bdlLeft.bundle.Target;
                    if (bdlRight != null) 
                        vRight = bdlRight.bundle.Target;
                    if (vLeft is VisComponent && vRight is VisComponent)
                        return vLeft == vRight;
                    return (int)vLeft == (int)vRight; // Covers enum types such as Key.F5

                case Operator.Neq: // A <> null may be true or false, not null.
                    if (vLeft == null || vRight == null) return vLeft != vRight;
                    if (typeLeft == TypeCode.Double || typeRight == TypeCode.Double)
                        return Convert.ToDouble(vLeft) != Convert.ToDouble(vRight);
                    if (typeLeft == TypeCode.Int32 || typeRight == TypeCode.Int32)
                        return Convert.ToInt32(vLeft) != Convert.ToInt32(vRight);
                    if (typeLeft == TypeCode.DateTime || typeRight == TypeCode.DateTime)
                        return Convert.ToDateTime(vLeft) != Convert.ToDateTime(vRight);
                    if (typeLeft == TypeCode.String && typeRight == TypeCode.String)
                        return string.Compare(vLeft.ToString(), vRight.ToString(), true) != 0; // Ignore case
                    bdlLeft = vLeft as BundleRef;
                    bdlRight = vRight as BundleRef;
                    if (bdlLeft != null) 
                        vLeft = bdlLeft.bundle.Target;
                    if (bdlRight != null) 
                        vRight = bdlRight.bundle.Target;
                    if (vLeft is VisComponent && vRight is VisComponent)
                        return vLeft != vRight;
                    return (int)vLeft == (int)vRight; // Covers enum types such as Key.F5

                case Operator.Gt: // string can only be compared to string. Automatic conversion is confusing.
                    if (vLeft == null || vRight == null) return null;
                    if (typeLeft == TypeCode.Double || typeRight == TypeCode.Double)
                        return Convert.ToDouble(vLeft) > Convert.ToDouble(vRight);
                    if (typeLeft == TypeCode.Int32 || typeRight == TypeCode.Int32)
                        return Convert.ToInt32(vLeft) > Convert.ToInt32(vRight);
                    if (typeLeft == TypeCode.DateTime || typeRight == TypeCode.DateTime)
                        return Convert.ToDateTime(vLeft) > Convert.ToDateTime(vRight);
                    if (typeLeft == TypeCode.String && typeRight == TypeCode.String)
                        return string.Compare(vLeft.ToString(), vRight.ToString(), true) > 0; // Ignore case
                    throw new Exception("Cannot combine these types");

                case Operator.Ge: // string can only be compared to string. Automatic conversion is confusing.
                    if (vLeft == null || vRight == null) return null;
                    if (typeLeft == TypeCode.Double || typeRight == TypeCode.Double)
                        return Convert.ToDouble(vLeft) >= Convert.ToDouble(vRight);
                    if (typeLeft == TypeCode.Int32 || typeRight == TypeCode.Int32)
                        return Convert.ToInt32(vLeft) >= Convert.ToInt32(vRight);
                    if (typeLeft == TypeCode.DateTime || typeRight == TypeCode.DateTime)
                        return Convert.ToDateTime(vLeft) >= Convert.ToDateTime(vRight);
                    if (typeLeft == TypeCode.String && typeRight == TypeCode.String) {
                        return string.Compare(vLeft.ToString(), vRight.ToString(), true) >= 0; // Ignore case
                    }
                    throw new Exception("Cannot combine these types");

                case Operator.Lt: // string can be converted to number, etc. Converting number, etc. to string is confusing.
                    if (vLeft == null || vRight == null) return null;
                    if (typeLeft == TypeCode.Double || typeRight == TypeCode.Double)
                        return Convert.ToDouble(vLeft) < Convert.ToDouble(vRight);
                    if (typeLeft == TypeCode.Int32 || typeRight == TypeCode.Int32)
                        return Convert.ToInt32(vLeft) < Convert.ToInt32(vRight);
                    if (typeLeft == TypeCode.DateTime || typeRight == TypeCode.DateTime)
                        return Convert.ToDateTime(vLeft) < Convert.ToDateTime(vRight);
                    if (typeLeft == TypeCode.String && typeRight == TypeCode.String)
                        return string.Compare(vLeft.ToString(), vRight.ToString(), true) < 0; // Ignore case
                    throw new Exception("Cannot combine these types");

                case Operator.Le: // string can be converted to number, etc. Converting number, etc. to string is confusing.
                    if (vLeft == null || vRight == null) return null;
                    if (typeLeft == TypeCode.Double || typeRight == TypeCode.Double)
                        return Convert.ToDouble(vLeft) <= Convert.ToDouble(vRight);
                    if (typeLeft == TypeCode.Int32 || typeRight == TypeCode.Int32)
                        return Convert.ToInt32(vLeft) <= Convert.ToInt32(vRight);
                    if (typeLeft == TypeCode.DateTime || typeRight == TypeCode.DateTime)
                        return Convert.ToDateTime(vLeft) <= Convert.ToDateTime(vRight);
                    if (typeLeft == TypeCode.String && typeRight == TypeCode.String)
                        return string.Compare(vLeft.ToString(), vRight.ToString(), true) <= 0; // Ignore case
                    throw new Exception("Cannot combine these types");

                case Operator.StrConcat:
                    return (vLeft == null ? "" : vLeft.ToString()) + (vRight == null ? "" : vRight.ToString());

                case Operator.Like:
                    string s = vRight.ToString().Replace('%', '*').Replace('_', '?'); // Replace database wildcard characters with VB wildcard characters.
                    return vbSystemFunctions.vbFunctions.vbLike(vLeft.ToString(), s);

                default:
                    throw new Exception("Unknown operator: " + op.ToString());
                }
            }
            catch (Exception) { throw new Exception("Cannot combine these types"); }
        }
    }

    //---------------------------------------------------------------------
    public class UnOp : Expression {
        public readonly Operator op;
        public readonly Expression e1;

        public UnOp(Operator op, Expression e1, Token t) {
            this.op = op;
            this.e1 = e1;
            this.token = t;
        }

        public override object Eval(VisComponent context, Template tpl) {
            VisEngine.curComponent = context; //Setting the current component
            object v = this.e1.Eval(context, tpl);
            VisEngine.curExprToken = this.token; // Prepare for errors
            if (v == null) return null;
            switch (op) {
            case Operator.Not:
                return !Convert.ToBoolean(v);
            case Operator.Neg:
                TypeCode t = Type.GetTypeCode(v.GetType());
                switch (t) {
                case TypeCode.Double:
                    return -Convert.ToDouble(v);
                default:
                    return -Convert.ToInt32(v);
                }

            default:
                throw new Exception("Unknown unary operator");
            }
        }
    }

    //---------------------------------------------------------------------
    public class ValPropertyRef : Expression { // E.g. Me[Index-1] ! Bottom or B ! Top. B is the target. 
        public PropertyInfo propGUI; // For the target property
        public Expression leftSideExpr; // The target control
        public int propIndex = -1; // For the target property. -1 if the target property has no formula. 0 for the name property
        public TplProperty tplProp; // Gives access to the expression and allows checking for Init in assign.
        bool bundleRef; // Whether to refer to the bundle or the component.

        public ValPropertyRef(Expression leftSideExpr, int propIndex, PropertyInfo Prop, TplProperty tplProp, bool bundleRef, Token t) {
            propGUI = Prop; this.leftSideExpr = leftSideExpr;
            this.propIndex = propIndex; // Is <0 when the property doesn't have a formula
            this.tplProp = tplProp; this.bundleRef = bundleRef;
            this.token = t;
        }

        public override object Eval(VisComponent context, Template tpl) {
            if (VisEngine.circularRef != null)
                return null; // If a circular reference has been detected, don't pursue new references. Error reporting will be confusing.
            object target = leftSideExpr == null ? context : this.leftSideExpr.Eval(context, tpl);
            VisComponent ctrl = target as VisComponent;
            if (ctrl == null) {
                BundleRef bdlRef = target as BundleRef;
                if (bdlRef == null)
                    return null; // There is no component. Bundle may be empty.
                ctrl = bdlRef.bundle.Target; // The target of the bundle ref. If an index is applied, the result will be the source.
            }
            Template targetTpl = ctrl.tpl;
            Template oldTpl = VisEngine.curTemplate; TplProperty oldTplProp = VisEngine.curTplProp;
            // Set new current stuff:
            VisEngine.curTemplate = targetTpl; VisEngine.curTplProp = targetTpl.valPropArray[propIndex];
            VisEngine.curExprToken = this.token; // Prepare for errors
            object result = (object)propGUI.GetValue(ctrl, null);
            VisEngine.curTemplate = oldTpl; VisEngine.curTplProp = oldTplProp;
            if (VisEngine.circularRef != null) {
                VisEngine.circularLevel++;
                ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp, "Circular reference level " + VisEngine.circularLevel, token, context.Index());
            }
            return result;
        }

        public override bool Assign(VisComponent context, Template tpl, object r, out object setValue) {
            VisComponent targetCtrl = leftSideExpr == null ? context : this.leftSideExpr.Eval(context, tpl) as VisComponent;
            setValue = null;
            if (tplProp != null && tplProp.expr != null && !tplProp.expr.Init) // Can assign to null formulas and Init formulas
                throw new Exception("Cannot assign to " + tplProp.Name + ". It must be \"Init\"");
            propGUI.SetValue(targetCtrl, r, new object[] { }); setValue = r;
            return true;
        }

        public virtual object GetValue(dynamic targetCtrl, Template targetTpl) { // Base method for non-position properties
            return propGUI.GetValue(targetCtrl, null);
        }
    }

    //---------------------------------------------------------------------
    public class DesignPropRef : Expression { // E.g. Me[Index-1] ! Level or B ! Selected. B is the target. 
        public Expression leftSideExpr;
        public int propIndex;
        public TplProperty tplProp; // Gives access to the expression and allows checking for Init in assign.
        bool bundleRef; // Whether to refer to the bundle or the component.

        public DesignPropRef(Expression e, int propIndex, TplProperty tplProp, bool bundleRef, Token t) {
            this.propIndex = propIndex; this.leftSideExpr = e;
            this.tplProp = tplProp; this.bundleRef = bundleRef;
            this.token = t;
        }
        public override object Eval(VisComponent context, Template tpl) {
            if (VisEngine.circularRef != null)
                return null; // If a circular reference has been detected, don't pursue new references. Error reporting will be confusing.
            object target = leftSideExpr == null ? context : this.leftSideExpr.Eval(context, tpl);
            VisComponent ctrl = target as VisComponent;
            if (ctrl == null) {
                BundleRef bdlRef = target as BundleRef;
                if (bdlRef == null)
                    return null; // There is no component.
                ctrl = bdlRef.bundle.Target; // The target of the bundle ref. If an index is applied, the result will be the source.
            }
            Template targetTpl = ctrl.tpl;
            Template oldTpl = VisEngine.curTemplate; TplProperty oldTplProp = VisEngine.curTplProp;
            VisEngine.curExprToken = this.token; // Prepare for errors
            if (ctrl.designerPropValues[propIndex].state < State.Updated && VisEngine.circularRef == null) {
                // If a circular reference has been detected, don't pursue new references. Error reporting will be confusing.
                // Stack current tpl and tplProp to prepare for calculating another tplProp:
                // Set new current stuff:
                VisEngine.curTemplate = targetTpl; VisEngine.curTplProp = ctrl.designerPropValues[propIndex].tplProp;
                TplPropDesigner targetTplProp = ctrl.designerPropValues[propIndex].tplProp as TplPropDesigner;
                targetTplProp.EvalAndSet(ctrl); // May throw CircularRefException
                VisEngine.curTemplate = oldTpl; VisEngine.curTplProp = oldTplProp;
                if (VisEngine.circularRef != null) {
                    VisEngine.circularLevel++;
                    ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, oldTpl, oldTplProp, "Circular reference level " + VisEngine.circularLevel, token, context.Index());
                }
            }
            object result = ctrl.designerPropValues[propIndex].Value;
            return result;
        }

        public override bool Assign(VisComponent context, Template tpl, object r, out object setValue) {
            VisComponent targetCtrl = leftSideExpr==null ? context : leftSideExpr.Eval(context, tpl) as VisComponent; // Corrected SL 11-03-13
            targetCtrl.designerPropValues[propIndex].Value = r; setValue = r;
            return true;
        }
    }

    //---------------------------------------------------------------------
    public class FunctionCallExpr : Expression { // Invokes the leftNode function, which may be a FunctionProperty or its subclass SystemFunction.
        public List<Expression> actualParams; // The parameters actually used in the call. They are expressions.
        // Depending on the formal parameters of the function, they will be evaluated before the call or left for the function to evaluate.
        public Expression leftNode;
        public MethodInfo fncMethod;
        public bool SQLsupport; // True when the function exists in SQL and has to be passed on to the database.

        public FunctionCallExpr(Expression leftNode, MethodInfo fncMethod, List<Expression> actualParams, bool SQLsupport, Token tk) {
            this.leftNode = leftNode;
            this.fncMethod = fncMethod;
            this.actualParams = actualParams;
            this.SQLsupport = SQLsupport;
            this.token = tk;
        }
        public override object Eval(VisComponent context, Template tpl) {
            VisEngine.curTemplate = tpl;
            VisEngine.curComponent = context;
            VisEngine.curExprToken = this.token; // Prepare for errors
            VisComponent source = context;
            if (leftNode != null) {
                object left = leftNode.Eval(context, tpl);
                if (left is VisComponent)
                    source = left as VisComponent;
                else {
                    BundleRef bdl = left as BundleRef;
                    if (bdl != null & bdl.bundle.Instances.Count > 0)
                        source = bdl.bundle.Instances[0].cmp; // Use left[0] as default.
                    else
                        return null; // There is no component to invoke. 
                }
            }
            ParameterInfo[] formalParams = fncMethod.GetParameters(); // The parameters expected by the method. There will always be the same number of 
            // formal and actual parameters. If there is a variable number of parameters, the last actual parameter will be an array of parameters, maybe with length = 0.
            object[] paramValues = new object[actualParams.Count]; // The parameters transferred to the function. They are values.

            for (int i = 0; i < actualParams.Count; i++) {
                Type formalType = formalParams[i].ParameterType;
                if (formalType == typeof(Expression[]))
                    paramValues[i] = (actualParams[i] as ParamArrayExpr).exprArray; // The function will evaluate the parameter array, e.g. Choose(n, exp1, exp2, exp3).
                else if (formalType == typeof(Expression))
                    paramValues[i] = actualParams[i]; // The function will evaluate the parameter.
                else
                    paramValues[i] = actualParams[i].Eval(context, tpl);
                VisEngine.curTemplate = tpl; // Evaluation may change the error location.
                VisEngine.curComponent = context;
                VisEngine.curExprToken = this.token; // Prepare for errors
            }
            try {
                return fncMethod.Invoke(source, paramValues); // The method object has a reference to the object or null if static.
            }
            catch (Exception e) { // Exceptions will be passed on as the InnerException if possible.
                if (e.InnerException == null) throw e;
                throw e.InnerException;
            }
        }
    }

    //---------------------------------------------------------------------
    public class AggregateField : Expression { // Becomes an SQL snippet, e.g. Avg([ptId]) As Aggr2. 
        // At run time, generates a Field structure inserted in RecordSetExpr.expFields. Becomes an SQL snippet at CreateInstances.
        Expression parameter;
        string aggregateName;
        string strAlias;
        public AggregateField(string name, string alias, Expression param, Token t) {
            parameter = param;
            aggregateName = name;
            this.strAlias = alias;
            this.token = t;
        }

        public override object Eval(VisComponent context, Template tpl) {
            VisEngine.curComponent = context; //Setting the current component
            string aggregateParam = parameter.Eval(context, tpl).ToString();
            VisEngine.curExprToken = this.token; // Prepare for errors
            Field field = new Field() {
                alias = strAlias,
                fieldName = aggregateName + "(" + aggregateParam + ")",
                isAggregate = true
            };
            return field;
        }
    }

    //---------------------------------------------------------------------
    public class AggregateCall : Expression { // e.g. Text: Count(a). Replace by a simple field with the proper alias name.
        string alias;
        Expression leftNode;
        public AggregateCall(Expression e, string alias, Token t) {
            this.alias = alias;
            this.leftNode = e;
            this.token = t;
        }
        public override object Eval(VisComponent context, Template tpl) {
            VisComponent leftComponent = leftNode == null ? context : this.leftNode.Eval(context, tpl) as VisComponent;
            VisEngine.curExprToken = this.token; // Prepare for errors
            if (leftComponent == null) return null; // May happen if there is an error in the data source
            object field = leftComponent.Row[alias];
            if (field is string && String.IsNullOrWhiteSpace(field as string) || field is DBNull)
                return null;
            return field;
        }
    }

    //---------------------------------------------------------------------
    public class Subscript : Expression { // Subscript is used in many states, e.g. tpl[i], Bundle[i], listProp[i]. 
        // To support indirect property reference and foreign tpl reference, resolution is done at run time. Cases to deal with:
        // LeftNode is bundleRef: tplA !findprop ! Bundle[i], target !Bundle(source !findProp) [i]. The result is a source component.
        // LeftNode is cmp: tplA[i], tplA !propB[i] {Indirect ref to component}, tplA !findProp[i]. The result is a target component.
        public Expression leftNode; // C
        public Expression expInteger; // i

        public Subscript(Expression expLeft, Expression eInt, Token t) {
            this.leftNode = expLeft;
            this.expInteger = eInt;
            this.token = t;
        }
        public override object Eval(VisComponent context, Template tpl) {
            VisEngine.curExprToken = this.token; // Prepare for errors
            VisEngine.curComponent = context; // Prepare for errors.
            Bundle bdl = null;
            int index;
            object left = leftNode == null ? context : this.leftNode.Eval(context, tpl);
            if (left == null) return null; // No bundle.
            BundleRef leftRef = left as BundleRef;
            VisComponent comp = left as VisComponent;
            if (leftRef != null)
                bdl = leftRef.bundle; //  tplA !findProp !Bundle !Last, target !Bundle(source !findProp) !Last
            else if (comp != null) {
                if (comp.tpl.RowBundle == null) return null; // Empty no bundle.
                leftRef = comp.bundleRef();
                bdl = leftRef.bundle;
            }
            if (leftRef != null) {
                BundleRef old = VisEngine.leftBundleRef; VisEngine.leftBundleRef = leftRef; // Stack leftPath
                object o = expInteger.Eval(context, tpl);
                VisEngine.leftBundleRef = old; // Unstack
                VisEngine.curExprToken = this.token; // Prepare for errors.
                index = Convert.ToInt32(o);
                if (index >= 0 && bdl.Instances.Count > index) {
                    if (comp != null)
                        return leftRef.bundle.Instances[index].cmp; // The result is a target component.
                    return bdl.Instances[index].cmp; // The result is a source component.
                }  else
                    throw new Exception("Index " + index.ToString() + " out of range 0:" + (bdl.Instances.Count - 1).ToString());
            }

            if (!left.GetType().IsArray)
                throw new Exception("The left-hand side is not a component");
            dynamic arrLeft = left;
            index = Convert.ToInt32(expInteger.Eval(context, tpl));
            if (index < 0 || index >= arrLeft.Length)
                throw new Exception("Index " + index.ToString() + " is out of range 0:" + (arrLeft.Count - 1).ToString());
            return arrLeft[index];
        }

        public override bool Assign(VisComponent context, Template tpl, dynamic newValue, out object setValue) {
            // Deals with list properties (arrays) of any type, even mixed types. Some components have list properties of integers,
            // some of colors, some of integer and strings, etc. Dynamic handles this. 
            // Other uses of subscripts, e.g. tplA[index], cannot be assigned to.
            VisEngine.curExprToken = this.token; // Prepare for errors
            setValue = null;
            dynamic vLeft = leftNode == null ? context : leftNode.Eval(context, tpl); // Becomes a reference to the array.
                if (vLeft == null) return false;
            try {
                if (!vLeft.GetType().IsArray)
                    throw new Exception("The property is not a list");
                int index = Convert.ToInt32(expInteger.Eval(context, tpl));
                if (index < 0 || index >= vLeft.Length)
                    throw new Exception("Index " + index.ToString() + " is out of range 0:" + (vLeft.Count - 1).ToString());
                vLeft[index] = newValue; // Array element changed directly.
                setValue = newValue;
            } catch {
                    throw new Exception("Cannot assign type " + newValue.GetType().ToString() + " to type " + vLeft[0].GetType().ToString());
            }
            return true;
        }
    }

    //---------------------------------------------------------------------
    public class IndirectPrefixedField : Expression { // E.g.: StartControl.tblMedOrder.medName, 
        // where StartControl: me-=MedOrder or a variable set by an event handler..
        public readonly string fieldName;
        public readonly string prefix;
        public Expression exp; // Computes a property reference

        public IndirectPrefixedField(Expression e, string prefix, string field, Token t) {
            this.fieldName = field;
            this.prefix = prefix;
            this.token = t;
            exp = e;
        }

        public override object Eval(VisComponent context, Template tpl) {
            object obj = this.exp.Eval(context, tpl);
            if (obj == null) return null;
            object field;
            VisEngine.curExprToken = this.token; // Prepare for errors
            if (obj is BundleRef) obj = (obj as BundleRef).bundle.Target;
            if (!(obj is VisComponent))
                throw new Exception("This is an indirect reference. " + token.val + " must be a component");
            VisComponent ctrl = obj as VisComponent;
            try {
                field = ctrl.Row[prefix + "." + fieldName];
            }
            catch (Exception) {
                try {
                    // In case the prefixed field doesn't exist in the query because it isn't used in a formula, the following will give a wrong result. 
                    // The correct action is an error message that the field has to be used in a formula.
                    field = ctrl.Row[fieldName]; // The database engine may insist on no table name.
                }
                catch (Exception) {
                    throw new Exception("This is an indirect reference. Must refer to a field in " + ctrl.tpl.Name
                        + ". The field must be used already in some formula.");
                }
            }
            if (field is DBNull) return null;
            return field;
        }

        public override bool Assign(VisComponent context, Template tpl, object r, out object setValue) {
            object obj = this.exp.Eval(context, tpl);
            setValue = null;
            if (obj == null) return false;
            VisEngine.curExprToken = this.token; // Prepare for errors
            bool change = false;
            if (r == null || r is string && String.IsNullOrWhiteSpace((string)r)) r = DBNull.Value; // To be able to set a foreign key to null.
            if (obj is BundleRef) obj = (obj as BundleRef).bundle.Target;
            if (!(obj is VisComponent))
                throw new Exception("This is an indirect reference. " + token.val + " must be a component");
            VisComponent targetCtrl = obj as VisComponent;
            DataRow row = targetCtrl.Row;
            if (r == null || r is string && String.IsNullOrWhiteSpace((string)r)) r = DBNull.Value; // To be able to set a foreign key to null.
            if (row.Table.Columns.Contains(fieldName)) { // The field name may be with or without the table prefix. 
                if (!object.Equals(row[fieldName], r)) { // Event handlers for Textbox ! LostFocus don't know whether 
                    // the user changed something or just clicked the text. Spare unnecessary updates.
                    row[fieldName] = r; setValue = r; change = true;
                }
            } else if (row.Table.Columns.Contains(prefix + "." + fieldName)) {
                // In case the prefixed field doesn't exist in the query because it isn't used in a formula, the following will give a wrong result. 
                // The correct action is an error message that the field has to be used in a formula.
                targetCtrl.Row[prefix + "." + fieldName] = r;
                change = true;
            } else
                throw new Exception("This is an indirect reference to a field in " + targetCtrl.tpl.Name
                    + ". But the field doesn't exist, maybe because it is not used by a formula in form " + targetCtrl.tpl.FormTpl.Name);
            if (change) {
                FormTemplate targetTpl = targetCtrl.tpl.FormTpl;
                if (!targetTpl.DirtyDataTemplates.ContainsKey(targetCtrl.tpl.Name))
                    targetTpl.DirtyDataTemplates.Add(targetCtrl.tpl.Name, targetCtrl.tpl);
            }
            return true;
        }
    }

    //---------------------------------------------------------------------
    public class IndirectPropOrField : Expression { // If a field, the field name is not prefixed. Prefixed field names are handled by IndirectPrefixedField.
        // E.g.: StartControl ! Color or StartControl.medName, where StartControl: me-=MedOrder or a variable set by an event handler.
        // StartControl may be in another form.
        public readonly string dot;
        public readonly string fieldOrProp;
        public Expression exp; // Computes a property reference

        public IndirectPropOrField(Expression e, string dot, string fieldOrProp, Token t) {
            this.dot = dot;
            this.fieldOrProp = fieldOrProp;
            this.token = t;
            exp = e;
        }

        public override object Eval(VisComponent context, Template tpl) {
            object target = this.exp.Eval(context, tpl);
            if (target == null) return null;
            //throw new Exception("The target component doesn't exist"); // Seems inconvenient, but should be an error.
            VisEngine.curExprToken = this.token; // Prepare for errors.
            BundleRef bdlRef = target as BundleRef;
            if (bdlRef != null)
                target = bdlRef.bundle.Target; // Make it a reference to the component.
            if (!(target is VisComponent))
                throw new Exception("This is an indirect reference. " + token.val + " must be a component");
            VisComponent targetCtrl = target as VisComponent;
            Template targetTpl = targetCtrl.tpl;
            object field;

            if (dot == ".") { // Try whether it is a field.
                if (targetTpl.RowTpl.RowsExpr == null)
                    throw new Exception("The dot shows that this is an indirect reference to a field, but " + targetCtrl.tpl.Name
                        + " has no row and thus no fields");
                try {
                    field = targetCtrl.Row[fieldOrProp];
                }
                catch (Exception) { // Whether a prefix is needed depends on the current target.
                    string prefix = targetCtrl.Rows.FieldPrefix(fieldOrProp);
                    try {
                        field = targetCtrl.Row[prefix + "." + fieldOrProp];
                    }
                    catch (Exception) { // The error should not be suppressed by a Default operator.
                        ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, tpl, VisEngine.curTplProp,
                        "The dot shows that this is an indirect reference to a field. " + fieldOrProp + " must be a field in " + targetCtrl.tpl.Name
                            + ". Further, " + fieldOrProp + " must be used already in some formula in " + targetCtrl.tpl.Name, this.token, context.Index());
                        throw new LoggedException();
                    }
                }
                if (field is DBNull) return null;
                return field;
            }

            if (fieldOrProp.ToUpper() == "INDEX") return targetCtrl.Index();
            if (fieldOrProp.ToUpper() == "LAST") return targetCtrl.Last();
            bool exists;
            object res = VisEngine.GetValue(targetCtrl, fieldOrProp, out exists);
            if (exists) {
                if (VisEngine.circularRef != null) {
                    VisEngine.circularLevel++;
                    ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp, "Circular reference level " + VisEngine.circularLevel, token, context.Index());
                }
                return res;
            }
            throw new Exception( "The ! shows that this is an indirect reference to a property, but " + fieldOrProp + " is not a property in " + targetCtrl.tpl.Name);
        }

        public override bool Assign(VisComponent context, Template tpl, object newVal, out object setValue) {
            object obj = this.exp.Eval(context, tpl);
            VisEngine.curExprToken = this.token; // Prepare for errors
            setValue = null;
            bool change = false;
            if (obj is BundleRef) obj = (obj as BundleRef).bundle.Target;
            if (!(obj is VisComponent))
                throw new Exception("This is an indirect reference. " + token.val + " must be a component");
            VisComponent targetCtrl = obj as VisComponent;
            Template targetTpl = targetCtrl.tpl;

            if (dot == "." && targetTpl.RowTpl.RowsExpr != null) { // Try whether it is a field.
                DataRow row = targetCtrl.Row;
                string prefix = targetCtrl.Rows.FieldPrefix(fieldOrProp);
                if (newVal == null || newVal is string && String.IsNullOrWhiteSpace((string)newVal)) newVal = DBNull.Value; // To be able to set a foreign key to null.
                if (row.Table.Columns.Contains(fieldOrProp)) { // The field name may be with or without the table prefix. 
                    if (!object.Equals(row[fieldOrProp], newVal)) { // Event handlers for Textbox ! LostFocus don't know whether 
                        // the user changed something or just clicked the text. Spare unnecessary updates.
                        row[fieldOrProp] = newVal; setValue = row[fieldOrProp]; change = true;
                    }
                } else if (row.Table.Columns.Contains(prefix + "." + fieldOrProp)) {
                    // In case the prefixed field doesn't exist in the query because it isn't used in a formula, the following will give a wrong result. 
                    targetCtrl.Row[prefix + "." + fieldOrProp] = newVal; // The database engine may insist on no table name.
                    setValue = targetCtrl.Row[prefix + "." + fieldOrProp]; change = true;
                } else
                    throw new Exception("This is an indirect reference to a field in " + targetCtrl.tpl.Name
                        + ". But the field doesn't exist, maybe because it is not used by a formula in form " + targetCtrl.tpl.FormTpl.Name);
                if (change) {
                    FormTemplate frmTpl = targetCtrl.tpl.FormTpl;
                    if (!frmTpl.DirtyDataTemplates.ContainsKey(targetTpl.Name))
                        frmTpl.DirtyDataTemplates.Add(targetTpl.Name, targetTpl);
                }
                return true;
            }

            TplProperty tplProp; PropInfo propInfo;
            tplProp = TplProperty.GetTemplateProperty(tpl, fieldOrProp, out propInfo);
            if (tplProp == null) return false;
            if (propInfo is ValueInfo) {
                (propInfo as ValueInfo).PropertySetter.SetValue(targetCtrl, newVal, null);
                return true;
            }
            if (propInfo == null) { // Designer property.
                targetCtrl.designerPropValues[tplProp.cmpPropIndex].Value = newVal;
                return true;
            }
            // It is an event property or an function property.
            if (dot == "!")
                throw new Exception("This is an indirect reference. " + fieldOrProp + " must be a property in " + targetCtrl.tpl.Name);
            else
                throw new Exception("This is an indirect reference. " + fieldOrProp + " must be a property or a field in " + targetCtrl.tpl.Name
                + ". The field must be used already in some formula.");
        }
    }

    //---------------------------------------------------------------------
    public class FieldExpr : Expression {
        public string fieldName;
        public Join join;
        public Template rowTpl; // The template where the field is retrieved.
        //public string prefix;
        public Expression leftSideExpr; // Is a component or a BundleRef.
        int parentCount; // Number of steps to walk up.

        public FieldExpr(Expression e, Join join, string field, Template rowTpl, int parentCount, Token t) {
            // The database engine may identify the column with or without prefix, although uVis always specifies the prefix.
            // Evaluation tries without prefix first. If exception, changes to full field name. Should be changed to use fieldID.
            this.fieldName = field;
            this.join = join;
            this.rowTpl = rowTpl;
            this.parentCount = parentCount;
            this.token = t;
            leftSideExpr = e;
        }

        public override object Eval(VisComponent context, Template tpl) {
            object target = leftSideExpr == null ? context : this.leftSideExpr.Eval(context, tpl);
            VisComponent ctrl = target as VisComponent;
            if (ctrl == null) {
                BundleRef bdlRef = target as BundleRef;
                if (bdlRef == null) return null; // There may not be a row. Then the field value is null. 
                ctrl = bdlRef.bundle.Target; // The target of the bundle ref. If an index is applied, the result will be the source.
            }
            VisEngine.curComponent = context; // Prepare for errors.
            VisEngine.curExprToken = this.token;
            object field;
            if (ctrl.Row == null)  // May happen when Rows evaluated with errors or when Limited Rows = 0 and there are zero rows. Ignore the field
                throw new Exception("Cannot acces the field in " + ctrl.tpl.Name + " because there is no row.");
            if (ctrl.Row.Table.Columns.Contains(fieldName)) { // The field name may be with or without the table prefix. 
                field = ctrl.Row[fieldName];
            } else if (ctrl.Row.Table.Columns.Contains(join.prefix + "." + fieldName)) {
                field = ctrl.Row[join.prefix + "." + fieldName];
            } else
                return null; // May happen when the row is in state deleted, etc.
            if (field is string && String.IsNullOrWhiteSpace(field as string) || field is DBNull)
                return null;
            return field;
        }

        public override bool Assign(VisComponent context, Template tpl, object newVal, out object setValue) {
            VisComponent ctrl = leftSideExpr == null ? context : leftSideExpr.Eval(context, tpl) as VisComponent;
            setValue = null; // May return without setting anything. Then the set value is null.
            if (ctrl == null) return false;  // There may not be a row. 
            VisEngine.curComponent = context; // Prepare for errors.
            VisEngine.curExprToken = this.token;
            bool change = false;
            setValue = newVal; // May return without setting anything because the field value hasn't changed.
            DataRow row = ctrl.Row;
            if (row == null || row.RowState == DataRowState.Detached || row.RowState == DataRowState.Deleted) // Happens when a component with focus, is deleted.
                return false;
            if (newVal == null || newVal is string && String.IsNullOrWhiteSpace((string)newVal))
                newVal = DBNull.Value; // To be able to set a foreign keys and other fields to null.
            if (row.Table.Columns.Contains(fieldName)) { // The field name may be with or without the table prefix. 
                if (!isEqual(row[fieldName], newVal)) { // Event handlers for Textbox ! LostFocus don't know whether 
                    // the user changed something or just clicked the text. Spare unnecessary updates.
                    row[fieldName] = newVal; setValue = row[fieldName]; change = true;
                    // Converts Double to int, sets default value, etc. depending on the field type. Throws exception for date to int, etc.
                }
            } else if (!isEqual(row[join.prefix + "." + fieldName], newVal)) {
                row[join.prefix + "." + fieldName] = newVal; setValue = row[join.prefix + "." + fieldName]; change = true;
            }
            if (setValue is DBNull)
                setValue = null; // Uvis language doesn't deal with DBNull.

            if (change) {
                FormTemplate targetTpl = ctrl.tpl.FormTpl;
                if (!targetTpl.DirtyDataTemplates.ContainsKey(ctrl.tpl.Name))
                    targetTpl.DirtyDataTemplates.Add(ctrl.tpl.Name, ctrl.tpl);
            }
            return true;
        }

        bool isEqual(object db, object r) { // db is the Database field value. r is the value to set in UVis, but DBNull for null.
            // See http://www.codeproject.com/Articles/18714/Comparing-Values-for-Equality-in-NET-Identity-and
            return object.Equals(db, r);
            //if (db is DBNull) return r == DBNull.Value;
            //if (db is string) return (db as string) == (r as string);
            //if (db is DateTime) return (DateTime)db == (DateTime)r;
            //if (db is int) return (int)db == (int)r;
            //return db == r; // Other numbers
        }
    }

    //---------------------------------------------------------------------
    public class RecordRefExpr : Expression { // E.g.: me >- relMedType. 
        // The formula terminates in the Record state. The result is a record reference.
        Expression leftNode; // The result of evaluation should be a ControlInstace.
        Relationship relation;
        public RecordRefExpr(Expression e, Relationship relation, Token t) {
            this.leftNode = e;
            this.relation = relation;
            this.token = t;
        }

        public override object Eval(VisComponent context, Template tpl) {
            VisComponent dynamicObj = leftNode == null ? context : leftNode.Eval(context, tpl) as VisComponent;
            VisEngine.curExprToken = this.token; // Prepare for errors
            if (dynamicObj.Row == null) // May happen when Limited Rows is 0 and there are zero rows. Ignore the field
                throw new Exception("There is no data row to access");
            return new RecordRef(dynamicObj.Row, relation);
        }
    }

    //---------------------------------------------------------------------
    public class TableExpr : Expression {
        public string strTableName = "";
        public Table table;

        public TableExpr(string tableName, Table tbl) {
            this.table = tbl;
            this.strTableName = tbl.Name;
        }

        public override object Eval(VisComponent context, Template tpl) {
            return new StringBuilder("[" + strTableName + "] ");
        }
    }

    //---------------------------------------------------------------------
    public class RelationExpr : Expression {
        public Expression next;
        public Operator Op;
        public VisTool.Relationship rel;
        public RelationExpr(Expression next, Operator op, VisTool.Relationship relation, Token t) {
            this.next = next;
            this.Op = op;
            rel = relation;
            this.token = t;
        }

        static System.Text.RegularExpressions.Regex regexNestedQuery = new System.Text.RegularExpressions.Regex(
            @"AS\s+parent(?<generation>\d+)\s+", System.Text.RegularExpressions.RegexOptions.Compiled |
            System.Text.RegularExpressions.RegexOptions.IgnoreCase);

        public override object Eval(VisComponent context, Template tpl) {
            StringBuilder sb = this.next.Eval(context, tpl) as StringBuilder;
            VisEngine.curExprToken = this.token; // Prepare for errors

            string nestedSQL = "( " + sb.ToString().Trim() + " ) ";
            switch (Op) {
            case Operator.LJ:
                nestedSQL += "LEFT ";
                break;
            case Operator.IJ:
                nestedSQL += "INNER ";
                break;
            default:
                nestedSQL += "LEFT ";
                break;
            }

            string joinSQLDefinedInRelationship = rel.JoinSnippet; // Keys in the relation
                nestedSQL += joinSQLDefinedInRelationship;
            return new StringBuilder(nestedSQL);
        }
    }

    //---------------------------------------------------------------------
    public class SQLcontrolProperty : Expression {
        public Expression e1;
        public SQLcontrolProperty(Expression e1) {
            this.e1 = e1;
        }
        public override object Eval(VisComponent context, Template tpl) {
            // Inside the query tail, a control property must become a constant. Numbers become 123.
            // However, strings have to be enclosed in quotes for the SQL engine to accept them.
            // And date-time has to be enclosed in # #.
            VisEngine.curExprToken = this.token; // Prepare for errors
            VisEngine.curComponent = context; //Setting the current component
            StringBuilder sb = new StringBuilder();
            sb.Append(" ");
            object e2 = e1.Eval(context, tpl);
            if (e2 is string)
                sb.Append("\"" + e2 + "\" ");
            else if (e2 is DateTime)
                // CodeCulture format (UK) for date is dd/MM/yyyy. This isn't accepted by the database (Access), which requires MM/dd/yyyy,
                // but also accepts MM-dd-yyyy. The ToString below generates MM/dd/yyyy.
                // Without codeCulture, ToString generates MM-dd-yyyy, using the local setting (Danish) for the dashes/slashes. 
                sb.Append("#" + ((DateTime)e2).ToString("MM/dd/yy H:mm:ss", VisEngine.codeCulture) + "# ");
            else if (e2 is Int32 || e2 is Int64 || e2 is float || e2 is double || e2 is bool)
                // Numbers must be given in code culture in SQL, not local culture.
                sb.Append(Convert.ToString(e2, VisEngine.codeCulture) + " ");
            else if (e2 == null)
                sb.Append("Null ");
            else
                throw new Exception("The operand must be text, number, bool, date or null");
            return sb;
        }
    }

    //---------------------------------------------------------------------
    public class SQL_inject : Expression { // @ expression. Calculates a string. In an SQL context, this string is inserted into the SQL string.
        public Expression e1;
        public SQL_inject(Expression e1) {
            this.e1 = e1;
        }
        public override object Eval(VisComponent context, Template tpl) {
            VisEngine.curExprToken = this.token; // Prepare for errors
            VisEngine.curComponent = context;
            object res = e1.Eval(context, tpl);
            return res == null ? "" : res.ToString();
        }
    }
    
    //---------------------------------------------------------------------
    public class CreateRows : RecordSetExpr { // 
        public Expression count;
        public CreateRows(Expression count, Template tpl) : base(null, tpl) {
            this.count = count;
        }
        public override object Eval(VisComponent context, Template tpl) {
            VisEngine.curExprToken = this.token; // Prepare for errors
            VisEngine.curComponent = context;
            object res = count.Eval(context, tpl);
            return (int)res;
        }
    }
   
    //---------------------------------------------------------------------
    public class SQLConcatExpr : Expression {
        public Expression eLeft;
        public Expression eRight;
        public SQLConcatExpr(Expression left, Expression right) {
            this.eLeft = left;
            this.eRight = right;
        }

        public override object Eval(VisComponent context, Template tpl) {
            VisEngine.curComponent = context; //Setting the current component
            StringBuilder sb = new StringBuilder();
            sb.Append(eLeft.Eval(context, tpl).ToString());
            sb.Append(eRight.Eval(context, tpl).ToString());
            return sb;
        }
    }

    //---------------------------------------------------------------------
    public class Join { // An element in the join list.
        public string prefix; // The prefix for addressing a field in this join. TableName for first join in a path.
        public Table table; // The table behind this join.
        public Relationship relation; // Null for first join in a path and for a control join that starts with the primary key of this table.
        public bool leftJoin;

        public Join(string prefix, Table table, Relationship relation, bool leftJoin) {
            this.prefix = prefix;
            this.table = table;
            this.relation = relation;
            this.leftJoin = leftJoin;
        }
    }

    //---------------------------------------------------------------------
    public partial class RecordSetExpr : Expression {
        public DataTable RecordSet; // A query result - a memory table
        public Template tpl; // Convenience for testing.

        public Expression JoinChain; // A next-next chain of RelationExpr objects, terminating with a TableExpr or ParentJoin. Replace by joinList and delete JoinChain??
        // When RecordSetExpr is evaluated, it uses the JoinChain to produce the Join part of the SQL statement.

        public string primaryTable = null; // The relation with the highest cardinality. Used for assignment and record creation.
        public string lastTable = null; // The last table in the join. Used at present to determine the primary table for Group By.

        public List<Field> fields = new List<Field>(); // Collects simple fields from all formulas.
        public List<Expression> aggregateFields = new List<Expression>(); // SL 07-07-12: Aggregate fields

        public List<Join> joinList = new List<Join>(10); // The elements in the join list. Can generate the SQL join. Tells compiler which fields are available.

        public Expression exprWhere; //expression for the where clause
        public Expression exprOrderby; //expression for the orderby clause
        public List<Expression> exprGroupby = new List<Expression>(10); // One expression for each groupby element.
        // Each element will either be a FieldExpr or a string expression. Field expressions are used for checking and adding groupby elements.
        public Expression exprHaving;  //expression for the Having clause
        public Expression exprTop;  //expression for the Top clause
        public string strTop; // The value of the Top clause as a string.
        public Expression exprCreate;  // Expression for the Create clause. A create clause adds a number of new, empty rows at the end of Rows.
        public int createCount = 0; // The number of rows to add.

        OleDbDataAdapter adapter; // An SQL string (SelectCommand) and a connection (the same across the datamap).
        // Has no Dispose()
        DataSet dataSet; // A collection of DataTables and relationships. The same one is used for all RecordSetExpressions (i.e. adapters).
        public string Name; // The name of the table created in the dataSet. Since all record sets are in the same dataset, the name must include the form name.

        public RecordSetExpr(string name, Template tpl) {
            adapter = new OleDbDataAdapter(null, VisEngine.DataMaps[0].Connection);
            this.Name = name;
            this.dataSet = VisEngine.DataMaps[0].DataSet; // global data repository
            this.tpl = tpl;
        }

        public bool NeedsGroupBy(Field field) {
            // Check that there is GroupBy, but this field is not in the list, thus needs GroupBy, First(), etc.
            if (this.exprGroupby.Count == 0) return false;
            for (int i = 0; i < exprGroupby.Count; i++) {
                FieldExpr e = exprGroupby[i] as FieldExpr;
                if (e != null && e.fieldName == field.fieldName && e.join.prefix == field.tableName)
                    return false;
            }
            return true;
        }
        public override object Eval(VisComponent context, Template tpl) { // Returns the full SQL-string for the query.
            // Build up the SQL string from left to right. In case of errors, return the string generated as far. Helps the designer see where the error was.
            StringBuilder select = new StringBuilder(250);
            select.Append("SELECT ");
            try {
                if (exprCreate != null) createCount = (int)exprCreate.Eval(context, tpl); // Not part of SQL. Rows will be added after querying the database.

                int top = int.MaxValue;
                if (exprTop != null) top = (int)exprTop.Eval(context, tpl); // TOP is right after SELECT
                if (VisEngine.LimitedRows) {
                    if (VisEngine.MaxRows == 0) top = 1;
                    else if (VisEngine.MaxRows < top) top = VisEngine.MaxRows;
                }
                if (top < int.MaxValue) {
                    this.strTop = top.ToString(); // Used by parent join.
                    select.Append("TOP " + this.strTop + " ");
                }

                bool first = true;
                foreach (Field f in this.fields) {
                    if (first) first = false; else select.Append(", ");
                    select.Append(f.prefix + "." + f.fieldName);
                }
                foreach (Expression e in this.aggregateFields) {
                    if (first) first = false; else select.Append(", ");
                    Field f = (Field)e.Eval(context, tpl); // Evaluate dynamic fields - not possible with Context == null ??
                    select.Append(f.fieldName + " AS " + f.alias);
                }
                if (first) select.Append(" 1 "); // Dummy value when no field is used. Faster than the traditional *.

                select.Append(" FROM ");
                select.Append(JoinChain.Eval(context, tpl).ToString());

                if (exprWhere != null) select.Append(" WHERE " + exprWhere.Eval(context, tpl).ToString());
                if (exprGroupby.Count > 0) select.Append(" GROUP BY " + CalculateGroupBy(context, tpl)); // Is a mixture of field expressions and string expressions.
                if (exprHaving != null) select.Append(" HAVING " + exprHaving.Eval(context, tpl).ToString());
                if (exprOrderby != null) select.Append(" ORDER BY " + exprOrderby.Eval(context, tpl).ToString());
                select.Append(";");
            }
            catch (Exception e) { // Make sure something is returned as a string to help debugging.
                ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, tpl, VisEngine.curTplProp, e.Message, VisEngine.curExprToken, 0);
            }
            return select.ToString();
        }

        public string CalculateGroupBy(VisComponent context, Template tpl) {
            StringBuilder s = new StringBuilder(50);
            bool first = true;
            foreach (Expression e in exprGroupby) {
                if (first) first = false; else s.Append(", ");
                FieldExpr f = e as FieldExpr;
                if (f == null)
                    s.Append(e.Eval(context, tpl).ToString());
                else
                    s.Append("[" + f.join.prefix + "." + f.fieldName + "]"); // Include fields explicitly mentioned in group by
            }
            return s.ToString();
        }

        public bool HasPrefix(string relName, out Join join, PathInfo pi) {
            join = null;
            Join last = null;
            if (!pi.inTailExpr && this.exprGroupby.Count > 0) { // A Group By query outside the query tail. Look only at the last join.
                last = this.joinList[this.joinList.Count - 1];
                if (relName.ToUpper() == last.prefix.ToUpper()) join = last;
            } else
                join = this.joinList.Find(delegate(Join j) { return relName.ToUpper() == j.prefix.ToUpper(); });
            return (join != null);
        }

        public int HasField(string fieldName, out Join join) { // Looks for the field in all the joins. 
            join = null;
            int count = 0;
            foreach (Join j in this.joinList) {
                if (j.table.Columns.ContainsKey(fieldName)) {
                    join = j;
                    count++;
                }
            }
            return count;
        }

        public string FieldPrefix(string fieldName) { // Looks for the field in the query. 
            // Indirect references use this and rely on the field being used in a formula already.
            foreach (Field f in this.fields) {
                if (f.fieldName.ToLower() == fieldName.ToLower())
                    return f.prefix;
            }
            return null;
        }

        public int HasRelation(string relationName, out Relationship rel) {
            rel = null;
            int count = 0;
            foreach (Join j in this.joinList) {
                Relationship r;
                if (j.table.Relationships.TryGetValue(relationName, out r)) {
                    rel = r;
                    count++;
                }
            }
            return count;
        }

        public int AddKeyFields(Join join) { // Join may be a table or a relation.
            // (Earlier used only by assign-statements to add key fields when assigning to a field in a local event handler. )
            // All join steps now include the primary key of the target. This ensures that Assign can always be made, e.g. from Assign in a local event handler,
            // from Assign in a foreign form event handler, or from a Text component that has a field reference in the Source property.
            // This must not be done for group-by queries, since it would group also on the primary key fields - maybe except the primary table.
            // In practice, adding all the primary keys means 1-2 additional fields in the queries.
            if (this.exprGroupby.Count > 0)
                throw new Exception("Don't add key fields to Group By: " + VisEngine.curTemplate + "!" + VisEngine.curTplProp.Name);

            foreach (Column f in join.table.Keys.Values) {
                this.AddField(new Field() { fieldName = f.Name, prefix = join.prefix, tableName = primaryTable }); // Add field if it is not in the list.
            }
            return join.table.Keys.Count;
        }

        public int AddField(Field field) { // Result: Number of existing fields with this name. Doesn't add a field that exists already.
            int oldCount = fields.Count(p => p.prefix.ToUpperInvariant() == field.prefix.ToUpperInvariant()
                && p.fieldName.ToUpperInvariant() == field.fieldName.ToUpperInvariant() && p.isAggregate == field.isAggregate);
            if (oldCount == 0) fields.Add(field);
            return oldCount;
        }

        public void AddAggregateField(Expression field) { // TODO SLa: Expressions are pointers. Makes no sense to compare them. The contents must be compared
            if (!aggregateFields.Contains(field)) aggregateFields.Add(field);
        }

        public void Unload() {
            if (RecordSet != null) {
                RecordSet.Columns.Clear();
                RecordSet.Clear();
            }
        }

        public void Load(string sql, Template tpl) { 
            // If the query fails, the error is reported and the RecordSet will be null. Fields are unknown. 
            //VisEngine.sw1.Stop(); VisEngine.sw2.Start(); VisEngine.count1++;
            Unload(); this.RecordSet = null;
            tpl.RowTpl.addedRows.Clear();
            if (!tpl.FormTpl.updating)
                Debug.Print("Should be updating1");
            //long s1 = s.ElapsedMilliseconds;
            try {
                adapter.SelectCommand = new OleDbCommand(sql, VisEngine.DataMaps[0].Connection);
                //string dataSetName = tpl.FormTpl.Name+tpl.FormTpl.cloneID.ToString();
                //VisEngine.sw2.Restart(); 
                long count = adapter.Fill(dataSet, Name); // Returns the number of rows changed or added, but we unloaded all of them.
                //VisEngine.sw2.Stop(); //VisEngine.sw1.Start();
                //long s2 = s.ElapsedMilliseconds;
                //Debug.Print(tpl.Name + ": " + sql + " Fill: " + VisEngine.sw2.ElapsedMilliseconds);
                RecordSet = dataSet.Tables[Name]; // if (RecordSet.Columns["a"].
                for (int i = 0; i < createCount; i++) {
                    System.Data.DataRow row = tpl.RowTpl.RowsExpr.RecordSet.NewRow();
                    RecordSet.Rows.Add(row);
                    tpl.RowTpl.addedRows.Add(row); // Ensure that the new rows will be added to the database.
                }
            }
            catch (Exception e) {
                ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, tpl, tpl.propRows, "Cannot load the rows: " + e.Message, tpl.propRows.startChain, 0);
            }
            if (!tpl.FormTpl.updating)
                Debug.Print("Should be updating2");
        }

        private Column getColumn(Field field) {
            Table table = VisEngine.DataMaps[0].getTable(field.tableName);
            if (table != null)
                return table.getColumn(field.fieldName);
            else return null;
        }

        private List<Column> getColumns(List<Field> fields) {
            List<Column> columns = new List<Column>();
            Column column = null;
            foreach (Field field in fields) {
                column = getColumn(field);
                if (column != null)
                    columns.Add(column);
            }
            return columns;
        }
    }

    //---------------------------------------------------------------------
    public class ParamArrayExpr : Expression { // Holds an array of actual parameters, e.g. for Choose(n, exp1, exp2 ... ) or OpenForm(s, p0, p1 ... )
        public Expression[] exprArray; // Used as it is when the formal parameter is Expression[]. Evaluated into object[] when the formal parameter is object[].
        public ParamArrayExpr(List<Expression> nodes, Token t) {
            exprArray = new Expression[nodes.Count];
            for (int i = 0; i < exprArray.Length; i++) exprArray[i] = nodes[i];
            this.token = t;
        }

        public override object Eval(VisComponent context, Template tpl) {
            object[] dynamicValues = new object[exprArray.Length];
            for (int i = 0; i < dynamicValues.Length; i++) {
                dynamicValues[i] = exprArray[i].Eval(context, tpl);
            }
            return dynamicValues;
        }
    }

    //---------------------------------------------------------------------
    public class FormulaExpr : Expression { // Templates ! formName ! propName = formula
        public string formula; // Allows showing the formula, even in closed vis-files.
        public string formName; // For addressing form ! template ! property, e.g. to set the formula.
        public string tplName;
        public string propertyName;
        public FormulaExpr(string formName, string tplName, string propertyName, string formula) {
            this.formName = formName;
            this.tplName = tplName;
            this.propertyName = propertyName;
            this.formula = formula;
        }

        public override object Eval(VisComponent context, Template tpl) {
            return formula;
        }

        public override bool Assign(VisComponent context, Template tpl, object r, out object setValue) {
            setValue = null;
            int cloneID= 0;
            FormTemplate frmTpl = VisEngine.GetFormTemplate(formName, ref cloneID); // Get any form clone.
            if (frmTpl == null) { throw new Exception("Form " + formName + " not open"); }
            Template targetTpl;
            if (!frmTpl.Templates.TryGetValue(tplName, out targetTpl) || targetTpl.Ignore)  
                throw new Exception("Template " + tplName + " does not exist"); 
            if (VisEngine.CreateOrSetTemplateProperty(targetTpl, propertyName, r.ToString()) == null)
                throw new Exception("Template " + tplName + " cannot be edited");
            setValue = r.ToString(); return true;
        }
    }
}
