/* SL 23-07-2011: New scanner and Compiler (Parser). Several reasons to abandon CoCoR:
 (1) With the VB line continuation (space _ ) we need to detect texts that look like it, e.g. "abc _" and "abc""de _"  .
       As a result even loading a vis-file needs detection of text strings. The logical consequence is to treat all constants (text, number and date) as tokens. 
 (2) CocoR syntax for VB comments and blank space was very hard and didn't work properly. (Below we use a state-transition program for it).
 (3) CocoR syntax is inconvenient when VisTool resolves ambiguity such as me.Like (Like is also a keyword).
 (4) Maintaining syntax and semantics is much easier in one file rather than the 10 used with CocoR.
 * 
 * SL 16-10-2013: We have abandoned the VB line continuation, but look for formula end in an even more complex way:
 * Formula end is triggered by EOF, or EOL followed by either "--" or ident:
 * This is handled by NextToken which may call ReadToken several times to check for formula end.
 * 
 * The disadvantage of this solution is ambiguity with formulas like this: a ? b : c. When b: occurs at the beginning of the line, the developer will see an error message.
 * */

using System.Collections.Generic;
using System.Collections;
using System.Reflection;
using System;
using System.Text;
using System.IO;
using System.Diagnostics;
//using System.Globalization;
using VisTool;

namespace VisTool {
    public class Token {
        public Token(int line) { this.line = line; }

        public Parser.TkKind kind;    // token kind: Ident, Long ... EOL
        public Parser.KeyType keyType; // The type of keyword. Some keywords have two types (with binary | ), e.g. Top (SQL keyword and property).
        public bool EOF = false; // True if the last character was EOF. Only EOL tokens can have EOF true.
        public bool cutToken = false; // When true, this is the last token of a formula piece cut out for Intellisense. Any kind of token can be a cutToken.
        public bool comment = false; // True inside a comment, i.e. from ' to EOL
        public string space; // Preceding white space and comments. The space may contain \r\n.
        public string val; // Raw value for the token. "" for EOL.
        public string newVal; // If val is a keyword, newVal is the correct caps. The Parser decides whether to use newVal.
        public Object obj; // Value for numbers, text and time values
        public enum SqlBracket { left = 1 << 1, right = 1 << 0 }; // Tells the sql snippet generator to surround the token with brackets.
        public SqlBracket sqlBracket = 0;
        public string error = null; // An error message to be shown when the Parser handles the token, e.g. an uncompleted text.
        public Token next;  // Link to next token.
        public int pos; // char pos of the val part. The first char of the formula has pos = 0. Used for error highlight. Defined only when the formula has an error.
        public int col;     // char column within the line (starting at 1). Helps user locate the error. Defined only when the formula has an error.
        public int line;    // line number for the start of the token. Increases during scanning of the stream (TextReader).
        public int lines = 0; // Number of lines in token. Preceding space may have its own lines. 
    }

    public class LoggedException : ApplicationException {
        public LoggedException() { } // The error has been logged. Continue with recovery action
    }
    public class CutException : ApplicationException {
        public CutException() { } // Parsing reached a cut token. Return to intellisense.
    }

    public partial class Parser {
        public Parser(TextReader r, bool cut) { // Used by Visparser and Vism prser
            cutToken = cut; // When reading a file, the last token may be considered cut (terminated with EOF)
            rd = r; la = tokens = pt = NextToken(); // There is an la-token. No token read.
            Peek(); // Look at first token - for performance measurements to ensure that file is open. Doesn't harm.
        }

        public int lineCount = 1; // Grows as the stream is scanned.

        public enum KeyType {
            None = 0, Constant = 1 << 0, Operator = 1 << 1, Sql = 1 << 2, By = 1 << 3,
            AscDesc = 1 << 4, Aggregate = 1 << 5, Init = 1 << 6, Property = 1 << 7, Prefix = 1 << 8,
            StatementOnly = 1 << 9, // Function that can be used only in statements (event handlers), e.g. Refresh( )
            AutoChange = 1 << 10, // Function that delivers a new value at each call, e.g. Rnd (random) and Time.
        };
        public static Dictionary<string, KeyType> Keywords = new Dictionary<string, KeyType>(StringComparer.OrdinalIgnoreCase) { 
        {"True", KeyType.Constant}, {"False", KeyType.Constant}, {"Null", KeyType.Constant},

        {"Mod", KeyType.Operator}, {"Not", KeyType.Operator}, {"And", KeyType.Operator}, {"Or", KeyType.Operator},
        {"Xor", KeyType.Operator}, {"Eqv", KeyType.Operator}, {"Imp", KeyType.Operator}, {"Like", KeyType.Operator},
        {"Default", KeyType.Operator},

        {"Create", KeyType.Sql}, {"Top", KeyType.Sql | KeyType.Property}, {"Where", KeyType.Sql}, {"Having", KeyType.Sql}, 
        {"Group", KeyType.Sql}, {"By", KeyType.By}, {"Order", KeyType.Sql}, {"In", KeyType.Sql}, 
        {"Asc", KeyType.AscDesc}, {"Desc", KeyType.AscDesc},

        {"Sum", KeyType.Aggregate}, {"Avg", KeyType.Aggregate}, {"Max", KeyType.Aggregate}, {"Min", KeyType.Aggregate},
        {"Var", KeyType.Aggregate}, {"VarP", KeyType.Aggregate}, {"StDev", KeyType.Aggregate}, {"StDevP", KeyType.Aggregate},
        {"Count", KeyType.Aggregate}, {"First", KeyType.Aggregate},

        {"Init", KeyType.Init},
        {"Main", KeyType.Prefix | KeyType.Property}, {"Me", KeyType.Prefix}, {"Form", KeyType.Prefix},
        {"Forms", KeyType.Prefix}, {"Templates", KeyType.Prefix}, {"Map", KeyType.Prefix}, {"System", KeyType.Prefix},

        {"Index", KeyType.Property}, {"Last", KeyType.Property}, {"Height", KeyType.Property}, {"Bottom", KeyType.Property},
        {"Left", KeyType.Property}, {"Width", KeyType.Property}, {"Right", KeyType.Property}, {"Name", KeyType.Property},
        {"Rows", KeyType.Property}, {"zOrder", KeyType.Property}, {"Canvas", KeyType.Property}, {"Visible", KeyType.Property},
        {"Param", KeyType.Property}, {"Style", KeyType.Property},

        // StatementOnly and AutoChange should be specified as attributes to the System items (VisSystemAtt), because SystemItems may be added by third party. ??
        //{"Refresh", KeyType.StatementOnly}, {"RefreshForm", KeyType.StatementOnly}, {"OpenForm", KeyType.StatementOnly}, 
        //{"MsgBox", KeyType.StatementOnly}, {"MsgLog", KeyType.StatementOnly}, {"Requery", KeyType.StatementOnly}, {"RequeryForm", KeyType.StatementOnly}, 
        //{"Time", KeyType.AutoChange}, {"Now", KeyType.AutoChange}, {"Date", KeyType.AutoChange},
        //{"ToDay", KeyType.AutoChange}, {"TimeOfDay", KeyType.AutoChange}, {"Rnd", KeyType.AutoChange}, {"RndInt", KeyType.AutoChange}, 
        {"If", KeyType.StatementOnly}, {"Else", KeyType.StatementOnly}, {"For", KeyType.StatementOnly}
        };
        public static string[] Operators = new string[] {"^", "+", "-", "*", "/", "\\", "&", 
            "=", "<", ">", "<>", "<=", ">=", "=<", "-<", ">-", "@",  "( )", "[ ]", "#24-12-2012 10:55#", "\"abc\"", "123.45", "1234.5E-1"};

        public enum TkKind { Ident, Long, Double, String, Time, Operator, EOL }; // Token kinds
        // Identifier (or keyword), number constant, string constant, time constant, operator (or seperator), end-of-line (or EOF)

        // EOL is treated as space characters inside multi-line formulas. The \r\n characters are part of the space property in the next token.
        // At the end of the formula, EOL becomes an EOL token.
        // End of formula is tricky: It occurs when an EOL is followed by "Ident:" (start of a new formula) or when EOL is followed by "--".
        // This is done in the NextToken method, which "repairs" the tokens it gets from ReadToken.
        // Only the last token in a formula is an EOL. When parsing a vis-file, only the last EOL token in the file has EOF.
        // When parsing a formula in a template property, the last token will usually be EOL with an EOF indication. 
        // The exception is when the designer types - - or EOL Ident: This will show as an EOL without EOF.

        public Token token; // last recognized token
        public Token la;    // look-ahead token. Not reliable after EOF. Apparently reads spurious characters
        public Token La {
            get {
                if (la.comment) pi.state = PathState.FreeText; // The token may hold comments, unterminated strings, etc.
                if (!la.cutToken) return la;
                if (la.error != null) pi.state = PathState.Value; // Will take care of showing " etc. as expected.
                pi.la = la; throw new CutException();
            }
            set { la = value; }
        }
        Token tokens;     // list of tokens already peeked. 
        Token pt;         // current token peeked
        StringBuilder bOut = new StringBuilder(); // Builds a new source text from token.val and la.space when the current token has been processed.
        // This allows the Parser to modify tokens (also peeked ones) and build a new source text. E.g. capitalize idents correctly  and correct join operators.
        Token startOut = null; // Token where new formula text will start. E.g. at la or at token.

        public static string frmName; // context identification for error messages
        public static string tplName; // context identification for error messages
        public static string propName; // context identification for error messages
        FormTemplate frm;
        public Template tpl; // The template being compiled.
        public TplProperty tplProp;
        bool formulaError = false; // True when an error is detected and parsing continues to find more errors.
        bool IsStatement; // True while parsing a statement. Some functions are only allowed for statements, e.g. Refresh or OpenForm.

        TextReader rd;
        char c, p;       // Read char, peek char
        int cInt, pInt;   // c and p as int to catch eof. -1 if no more characters.

        public void error(string message, Token t) { error(message, t, Priority.Error); }

        public void error(string message, Token t, bool Warning) { error(message, t, Warning ? Priority.Warning : Priority.Error); }

        public void error(string message, Token t, Priority p) {
            //if (t != null && t.kind == tk.EOL && t.next != null && t.next.kind != tk.EOL) { // This happens when tokens look like end-of-formula, e.g. -- or ident: 
            //    t = t.next; // Highlight the next token. 
            //    message = "Looks like the beginning of a new formula";
            //}
            if (cutToken) { // If scanner is in cutToken mode, ignore errors at the end of the formula.
                if (la == t) { // Some errors report token as the error location, others report la. If la, check for next being EOL.
                    ResetPeek();
                    t = Peek(); if (t.EOF) return;
                }
                if (la.cutToken) {
                    la.kind = TkKind.Operator; return; // Don't return EOL when cutToken. Confuses Intellisense who cannot look ahead.
                }
            }
            ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, tpl, tplProp, message, t, 0);
        }

        char Read(Token t) {
            if (c == '\n') {
                lineCount++;  
                t.lines++;
            }
            cInt = rd.Read(); c = (char)cInt; pInt = rd.Peek(); p = (char)pInt;
            return c;
        }

        /* ReadToken: Gets the next token from the text reader. Stores the raw chars of the token in token.val. Stores the semantic value in token.obj.
        Stores the preceding WhiteSpace and comments in token.space. 
        Error handling: NextToken will always return a token. It will not report an error, but set token.error and let the Parser report the error.
        Why? Because the scanner peeks ahead of the Parser and doesn't know which property the error belongs to.
        Errors such as number format or end-of-line inside a text, will treat the rest of the formula as a comment, including possible line 
        continuations. Instead it might be possible to generate a number token or text token with token.error, and continue parsing ??  */

        Token ReadToken() {
            Token t = new Token(lineCount);
            StringBuilder bSpace = new StringBuilder(); // Preceding white space including a possible continuation mark
            StringBuilder bVal = new StringBuilder(); // String for ident and operators
            StringBuilder bText = new StringBuilder(); // String for text value

/* Space: Scanning the preceding space is amazingly complex. We use a state-transition program with these states: 
 * initial, afterEOL, comment ... continued line. A goto-style is actually more readable in this case.
 * In contrast to VB, we allow a comment after the continuation mark (space _ ).
 * 
 * SL 16-10-13: We now allow multi-line formulas without continuation mark. The formula-end is EOF, "--" or EOL ident ":".
 * This requires peeking ahead. ReadToken just returns EOL tokens, but when NextToken gets the EOL, it peeks ahead. 
 * When it is not formula-end, it adds the space part to the following token and deletes the EOL-token. 
 */

    initial: // Last read char: end of previous token
            if (pInt < 0 && cutToken) { t.cutToken = true; t.EOF = true; goto endToken; }
            if (Read(t) == '\n') { bSpace.Append(c); goto afterEOL; }
            if (cInt < 0) {
                t.kind = TkKind.EOL; t.EOF = true;
                if (cutToken) { t.cutToken = true; }
                goto endToken;
            }
            if (Char.IsWhiteSpace(c)) { bSpace.Append(c); goto initial; }
            if (c == '\'') { bSpace.Append(c); goto comment; }
            goto afterSpace; // Other character

    afterEOL: // Last read char: EOL. We allow blank lines after any formula. They must be included in the EOL token.
            while (true) {
                if (p == '\n' || Char.IsWhiteSpace(p)) { Read(t); bSpace.Append(c); continue; }
                if (pInt < 0) {
                    Read(t); t.kind = TkKind.EOL; t.EOF = true;
                    if (cutToken) { t.cutToken = true; }
                    goto endToken;
                }
                if (p == '\'') { Read(t); bSpace.Append(c); goto comment; }
                t.kind = TkKind.EOL; goto endToken; // Other character, not read yet. Belongs to next token.
            }
comment: // Last read char: ' (single quote). Or after error.
            t.comment = true;
            while (Read(t) != '\n' && cInt >= 0) bSpace.Append(c);
            if (c == '\n') { bSpace.Append(c); t.comment = false; }
            goto afterEOL;

afterSpace: // Last read char: a non-space character
            if (Char.IsLetter((char)c)) {  // Identifier or operator name
                t.kind = TkKind.Ident; bVal.Append(c);
                while (Char.IsLetterOrDigit(p) || p == '_') { // Underscore allowed in VB identifiers after letter
                    Read(t); bVal.Append(c);
                }
                if (pInt < 0 && cutToken) { t.cutToken = true; } // Uncompleted ident will be marked cutToken to help Intellisense.

            } else if (Char.IsDigit((char)c)) {  // Number
                t.kind = TkKind.Long; bVal.Append(c);
                while (Char.IsDigit(p) || p == '.' || p == 'E' || p == 'e') {
                    Read(t); bVal.Append(c);
                    if (c == 'E' || c == 'e') {
                        t.kind = TkKind.Double;
                        if (p == '-' || p == '+') { Read(t); bVal.Append(c); }
                    }
                    if (c == '.') t.kind = TkKind.Double;
                }
                t.val = bVal.ToString();
                try {
                    t.obj = (t.kind == TkKind.Double ? Convert.ToDouble(t.val, VisEngine.codeCulture) : t.obj = Convert.ToInt32(t.val));
                }
                catch {
                    if (t.kind == TkKind.Double) t.obj = 1.0; else t.obj = 1; // Set default value
                    t.error = "Wrong number"; goto endToken;
                }

            } else if (c == '#') { // DateTime
                t.kind = TkKind.Time; bVal.Append(c);
                while (p != '#') {
                    if (p == '\n' || pInt < 0) {
                        t.obj = new DateTime(1990, 1, 1); // Set default date 
                        t.error = "DateTime cannot continue across line break";
                        goto endToken;
                    }
                    Read(t); bVal.Append(c);
                }
                Read(t); bVal.Append(c);
                t.val = bVal.ToString();
                try {
                    t.obj = Convert.ToDateTime(t.val.Substring(1, t.val.Length - 2), VisEngine.codeCulture); // Strip # # away
                }
                catch {
                    t.obj = new DateTime(1990, 1, 1); // Set default date 
                    t.error = "Wrong DateTime format"; goto endToken;
                }

            } else if (c == '"') { // Text string = " {any including "" except EOL} "
                t.kind = TkKind.String; bVal.Append(c);
                while (true) {
                    if (p == '\n' || pInt < 0) {
                        if (cutToken) t.cutToken = true;
                        t.error = "Text cannot continue across line break";
                        break;
                    }
                    Read(t); bVal.Append(c);
                    if (c == '"') { // Double quote
                        if (p != '"') break; // Single double quote
                        Read(t); bVal.Append(c); // Twin double quotes
                    }
                    bText.Append(c);
                }
                t.obj = bText.ToString();

            } else { // Operator or separator
                t.kind = TkKind.Operator; bVal.Append(c);
                if (c == '-' && (p == '-' || p == '<' || p == '=') ||   // --, -<
                c == '=' && (p == '<') ||                               // =<
                c == '>' && (p == '-' || p == '=') ||                 // >-, >=  
                c == '<' && (p == '=' || p == '>')) {               // <=
                    Read(t); bVal.Append(c);
                }
            }

endToken: // Last read char: end of token
            t.val = bVal.ToString(); t.space = bSpace.ToString();
            if (t.kind == TkKind.Ident) Keywords.TryGetValue(t.val, out t.keyType);
            return t;
        } // NextToken

        Token NextToken() { // Reads the next token and maybe one or two more tokens to check whether this is a true EOL (end-of-formula).
            // The tokens are linked through token.next.
            Token t = ReadToken();
            if (t.kind != TkKind.EOL) return t; // Non-EOL token
            Token peek = t.next = ReadToken(); // The EOL may be end-of-formula or white space.
            if (peek.val == "--" || peek.EOF || peek.kind == TkKind.Ident && (peek.next = ReadToken()).val == ":")
                // We have end-of-formula. Leave it as an EOL. The vis-parser will trim blank space at the end when setting the template formula.
                return t;

            // We have an EOL that is not part of End-of-formula. Treat it as white space.
            peek.space = t.space + peek.space; // Add the EOL's space to the next token.
            return peek; // Drops the t-token.
        }

        public static bool IsIdent(string s) { // ident = letter {letter |digit |_ }
            if (s.Length == 0) return false;
            Char[] chars = s.ToCharArray();
            if (!Char.IsLetter(chars[0])) return false;
            for (int i = 1; i < chars.Length; i++)
                if (!(Char.IsLetter(chars[i]) || Char.IsDigit(chars[i]) || chars[i] == '_')) return false;
            return true;
        }

        // Output the possibly modified token and get the next token (possibly a token already seen during peeking).
        // Build val and trailing space.
        public void Get() {
            if (token != null) { // Null initially
                if (la == startOut) {
                    bOut = new StringBuilder(); bOut.Append(la.space);
                    // Output preceding space. The val may be changed and will be output later.
                } else {
                    bOut.Append(token.val); bOut.Append(la.space); // Output token and trailing space
                    // When la is EOL, output the preceding space is needed, but not EOL itself
                }
            }
            if (la.cutToken) {
                if (la.comment) pi.state = PathState.FreeText;
                pi.la = la; throw new CutException(); 
            }           
            token = la;
            if (token.EOF) return; // Stop generating tokens
            if (tokens.next == null) tokens.next = la = pt = tokens = NextToken(); else la = pt = tokens = tokens.next;
            // pt = tokens resets Peek.
        }

        public void StartOut(Token t) { // Record the first token of the formula
            startOut = t;
        }

        public void Expect(TkKind n) { // At someting unexpected, Logs an error with a state-dependent text.
            if (la.kind == n) {
                Get(); return; 
            }
            if (la.cutToken) {
                VisIntelliSense.expectedSymbol = n.ToString();
                formulaError = true; // Makes the parser not expecting EOL.
                if (la.comment) pi.state = PathState.FreeText;
                pi.la = la; throw new CutException();
            }
            //VisIntelliSense.expectedSymbol = n.ToString();
            error(ExpectText(), la);
            formulaError = true; // Makes the parser not expecting EOL.
            throw new LoggedException(); // Was: "Expects " + s
        }

        void Expect(string s, string message) { // At someting unexpected, Logs an error with the text "message".
            //VisIntelliSense.expectedSymbol = s;
            if (la.cutToken) {
                if (la.comment) 
                    pi.state = PathState.FreeText;
                else 
                    VisIntelliSense.expectedSymbol = s;
                pi.la = la; throw new CutException();
            }
            if (s.ToUpper() == la.val.ToUpper()) {
                Get();
            } else if (la.cutToken) {
                formulaError = true; // Makes the parser not expecting EOL.
                throw new LoggedException();
            } else {
                error(message, la);
                formulaError = true; // Makes the parser not expecting EOL.
                throw new LoggedException();
            }
        }

        void Expect(string s) { // At someting unexpected, Logs an error with the text "Expects s".
            //VisIntelliSense.expectedSymbol = s;
            if (la.cutToken) {
                if (la.comment)
                    pi.state = PathState.FreeText;
                else
                    VisIntelliSense.expectedSymbol = s;
            }
            if (s.ToUpper() == la.val.ToUpper()) {
                Get();
                //} else if (la.cutToken) {
                //    VisIntelliSense.expectedSymbol = s;
                //    formulaError = true; // Makes the parser not expecting EOL.
                //    throw new LoggedException();
            } else {
                error("Expects " + s, la);
                formulaError = true; // Makes the parser not expecting EOL.
                throw new LoggedException();
            }
        }

        string ExpectText() { // Returns a string  corresponding to what is expected in the current pi.state. 
            //if (la.cutToken) {
            //    if (la.comment) pi.state = PathState.FreeText;
            //    pi.la = la; throw new CutException();
            //}
            string test = ""; // pi.state.ToString() + ": ";
            switch (pi.state) {
            case PathState.Init: return test + "Expects an operand (field, property, 123, \"abc\") or a unary operator (Init, -, Not)";
            case PathState.Not: return test + "Expects an operand (field, property, 123, \"abc\") or a unary operator (-, Not)";
            case PathState.Unary: return test + "Expects an operand (field, property, 123, \"abc\") or - ";
            case PathState.ExpectPrimary: return test + "Expects  an operand, e.g. field, property, 123, \"abc\"";
            case PathState.Unknown:
                if (!pi.inTailExpr && (tplProp.Name.ToUpper() == "ROWS"))
                    return test + "Expects a table or \"Main\"";
                else
                    return test + "Expects a field, property, table, function, etc.";
            case PathState.Forms: return test + "Expects a Form name (a short file name)";
            case PathState.Control: if (!pi.inTailExpr && (tplProp.Name.ToUpper() == "ROWS"))
                    return test + "Expects a relation name";
                else
                    return test + "Expects a field, property, relation, function or [ ]";
            case PathState.Property: return test + "Expects [ ], property, or an operator, e.g. =, +, &, And" + Bracket();
            //case PathState.Indirect: return test + "Expects an operator or [ ]";
            case PathState.Function: return test + "Expects ( )";
            case PathState.Map: return test + "Expects a table";
            case PathState.RecordSet: return test + "Expects a relation name or Where, Group By . . .";
            case PathState.Record: return "Expects a field name or relation";
            case PathState.Field: return test + "Expects an operator, e.g. =, +, &, And" + Bracket();
            case PathState.Aggregate: return test + "Expects (field)";
            case PathState.Templates: return test + "Expects a Form name";
            case PathState.FormTpl: return test + "Expects a property name or template name";
            case PathState.Template: return test + "Expects a property name";
            case PathState.System: return test + "Expects a function, e.g. Refresh or Sin";
            case PathState.Value: return test + "Expects an operator, e.g. =, +, &, And" + Bracket();
            case PathState.ExpectSymbol: return "Expects " + VisIntelliSense.expectedSymbol;
            default: return test + "Unexpected symbol";
            }
        }

        string Bracket() {
            if (pi.brackets.Count == 0) return "";
            return ", " + pi.brackets.Peek();
        }

        public Token Peek() {
            if (pt.next == null) { pt = pt.next = NextToken(); } else { pt = pt.next; }
            return pt;
        }

        // Start peeking from la
        public void ResetPeek() { pt = tokens; }

        public static string GetFormulaFromTokens(Token t) {
            StringBuilder s = new StringBuilder(200);
            while (t != null) {
                s.Append(t.space); s.Append(t.val);
                if (t.EOF) break;
                t = t.next;
            }
            return s.ToString();
        }

    } // end class Compiler

} // end namespace VisTool.Kernel
