﻿using System;
using System.IO;
using System.Linq;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Data;
using System.ComponentModel;
using System.Reflection;
using System.Data.OleDb;
using VisTool;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Drawing2D;

namespace VisTool {
    /* VisComponent is the base class for all visible components, e.g. Box, Glyph, PieSlice, TimeScale, Textbox, Canvas and VisForm.
     * 
     * Some components are guiComponents. They have a Windows.Forms control as a local object. Examples: a VisTool.Textbox holds a Windows.Forms.TextBox 
     * and a VisTool.VisForm holds a Windows.Forms.Form. Why not make these guiComponents subclasses of Windows.Forms.TextBox and Windows.Forms.Form? 
     * Because VisTool needs to attach formulas to the properties and Windows doesn't allow overriding control properties. 
     * 
     * Other components are Shapes. They are painted by VisTool on a canvas using System.Drawing. Painting means drawing polygones, texts, etc. 
     * Why doesn't VisTool define new Forms controls instead? Because Forms controls are slow when there are many. Above 500 Textbox controls, time grows faster than linear. 
     * VisTool is intended for data visualization with thousands of components and can show 5,000 labels within a second.
     * 
     * Some components are Anonymous (Anonymous = true). They are created automatically by some ordinary component, e.g. TreeNode with lines and icon for each node in a Tree, 
     * extended tags and dragable borders for TimeScale  The designer cannot select them. They refer to an existing bundle and don't have a template of their own. 
     * They are assigned to a canvas.. 
     * 
     * The drawing canvas is either a VisTool Canvas or a VisForm. They are subclasses of ContainerComponent.
     * They keep track of the components drawn on them by means of 
     *      Dictionary<VisComponent, VisComponent> dictComponents
     *      VisComponent[] visComponents and 
     * Components are inserted and deleted from dictComponents during refresh. At the beginning of painting, the container creates an Array version of dictComponents in visComponents
     * and sorts the components according to zOrder. The container then paints the components in sequence so that the high zOrders are painted on top of those with lower zOrder.
     * 
     * Event handling for guiComponents: During construction, the component attaches common event handlers to the Control, using AttachControlEvents(ctrl).
     * The events are MouseMove, KeyDown, etc., which most components support. The event handlers are simple methods in VisComponent with mostly one statement, like this:
     *                      this.ExecuteHandler("MouseMove", true); // DefaultHandled is true.
     * ExecuteHandler simply looks for a property with the name, e.g. "MouseMove", and then evaluates the formula. When there is no such property, the event is ignored.
     * When "this" has got a chance to handle the event, ExecuteHandler tries the canvas hierarchy from "this" and to the form. Any handler on the way may stop or continue the propagation 
     * by setting Handled to true or false. Most events have Handled = true as default. Key-events have Handled = false as default. 
     * As an example, a Textbox handles keys and as a default passes them on to the canvas, which for instance may respond to tab and escape keys.
     * 
     * Some components attach additional event handlers that call ExecuteHandler in the same way. As an example, Textbox attaches a TextChanged event handler.
     * 
     * Event handling for Shapes: No event handlers are attached to the component. The ContainerComponent just calls the common event handlers at the right moments.
     * As an example, when the ContainerComponent gets a MouseMove event from Windows, it detects which Shape the mouse is over, and calls 
     * shape.ExecuteHandler("MouseMove"). The ContainerComponent also checks whether the Shape is the same as for the last MouseMove. If not, 
     * it calls ExecuteHandler("MouseLeave") for the old Shape and ExecuteHandler("MouseEnter") for the new Shape. It generates the other common events in the same way.
     * 
     * The key events (KeyDown, KeyPress, KeyUp) are tricky. The ContainerComponent keeps track of which component is in focus (i.e. gets the key-events).
     * It is the latest one that got a click event. When the ContainerComponent receives a key event from Windows, it calls the corresponding event handler for the shape in focus.
     * However, the guiComponents interfere. As an example, a Windows TextBox keeps its focus even if VisTool sets the focus to the Windows Form. 
     * So if there is just one textbox on the Form, it gets all key events and the ContainerComponent gets none. The remedy is to set Form.KeyPreview = true 
     * Now the form gets all key-events. When a shape is in focus, the form passes the event to it, otherwise it lets Windows handle the event. 
     * The same is done when a container is in focus since it doesn't handle the event and doesn't call the uVs event handler.
     * 
     * To do this, a ContainerComponent uses this state variable: 
     *     VisComponent mouseFocus; // The component that latest saw the mouse. Used to generate MouseEnter, MouseLeave, MouseHover.
     * A form uses also this state variable:
     *     private VisComponent keyFocus;      // The component that will receive KeyDown, KeyPress and KeyUp when the form previews them.
     *     (Canvases do the same?? SL 02-10-14.)
     */

    public class VisComponent : IDisposable { // This is the base class for all visual Components.
        // Components are created in VisEngine.UpdateBundle, using reflection to point to the proper component subclass.
        public Bundle [] childBundles = null; // One bundle for each Find-formula with my tpl as target. The bundles have me as Target.
        // Will be null until the first find-formula with me as target has been evaluated. 
        public VisComponent() { } // Used as base class for some components.

        public Template tpl;

        public int Index() { // We use property values rather than RowBundle to prepare for user re-ordering.
            BundleRef val = this.propertyValues[VisComponent.RowsID].Value as BundleRef;
            if (val != null) return val.index;
            return 0; // May happen during Rows evaluation because Rows hasn't been evaluated.
        }

        public int Last() { // We use property values rather than RowBundle to prepare for user re-ordering.
            BundleRef val = this.propertyValues[VisComponent.RowsID].Value as BundleRef;
            if (val != null) return val.bundle.Instances.Count -1;
            return -1; // May happen during Rows evaluation because Rows hasn't been evaluated.
        }

        public BundleRef bundleRef() { // We use property values rather than RowBundle to prepare for user re-ordering.
            return this.propertyValues[VisComponent.RowsID].Value as BundleRef;
        }

        public CmpProp[] propertyValues = new CmpProp[0]; // Keeps track of formula value and tags indicating updated, visited
        public CmpProp[] designerPropValues = new CmpProp[0]; // Keeps track of formula value and tags indicating updated, visited
        public DataRow Row;

        public static bool loadRecords = true;
        public bool Disposed = false;
        public bool Anonymous = false; // True for auxilary components that designers cannot select, e.g. TreeNode. 
        public enum Move { No, Init, Moved }; // No: Moving is not attempted. Init: Mouse is down and the component is movable. Moved: Also a MouseMove has been done.
        public Move Moving = Move.No;
        public Control GuiInstance;

        public VisComponent(int targetCount, DataRow row, Template tpl) {
            this.tpl = tpl;

            propertyValues = new CmpProp[tpl.valPropArray.Length];
            for (int pi = 0; pi < tpl.valPropArray.Length; pi++) {
                propertyValues[pi] = new CmpProp(tpl.valPropArray[pi], this);
            }

            designerPropValues = new CmpProp[tpl.DesignProperties.Count]; // Initialize all of them.
            int dpi = 0;
            foreach (KeyValuePair<string, TplPropDesigner> designProp in tpl.DesignProperties) {
                designerPropValues[dpi++] = new CmpProp(designProp.Value, this);
            }
            this.Row = row;
        }

        public VisForm Form() {
            if (this.tpl.FormTpl.RowBundle.Instances.Count == 0) return null;
            return this.tpl.FormTpl.RowBundle.Instances[0].cmp as VisForm;
        }

        public virtual void ResetUpdated() { // Called by kernel at refresh to set internal dependent properties to not-updated.
            // Components with additional dependencies override ResetUpdated and call base.ResetUpdated.
        }

        public virtual void ComponentUpdated() { // Called when VisEngine has calculated all properties of the component.
            // Allows the component to make more complex calculations on the properties, e.g. Tree.
        }

        public virtual void BundleUpdated() { // Called when VisEngine has calculated all properties of the bundle.
            // Allows the component to make more complex calculations on the bundle, e.g. Connector.
        }

        public void SetUpdated(ref CmpProp state) { // Sets state to Updated unless it is Initialized. Used by property setters.
            //  When a property has been calculated by an Init-formula, its state is already Initialized. 
            // For other formulas and internal calculations, the state must become Updated.
            if (state.state < State.Updated) state.state = State.Updated;
        }

        public object EvalProp(int propID, object defaultVal) {
            /* Evaluates the formula. Handles circular references. Assumes that propState is Old or Visited.
             * Makes the property state Initialized if there is no formula or an Init-formula. Makes it Updated otherwise.
             * If there is no formula or the formula returns null or an exception, returns the defaultValue. 
             * The property may be the same as the one being evaluated (VisEngine.curTplProp) or different. This has influence on how cyclical
             * references are reported.
            */
            TplProperty targetProp = tpl.valPropArray[propID];
            CmpProp instProp = this.propertyValues[propID];
            object res = defaultVal;
            if (targetProp != VisEngine.curTplProp && VisEngine.circularRef != null) // This is an internal reference from property to property in the same component.
                // A circular reference has been detected earlier. Don't pursue new references. Error reporting will be confusing. 
                return defaultVal;
            if (targetProp.expr != null) {
                if (targetProp.expr is ExpressionList)
                    ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, tpl, targetProp, "This property cannot be a list", null, this.Index());
                else {
                    if (instProp.state == State.Visited) {
                        ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, tpl, targetProp, "End of circular reference, level 0", null, this.Index());
                        VisEngine.circularRef = instProp; VisEngine.circularLevel = 0;
                    } else {
                        instProp.state = State.Visited;
                        Template oldTpl = VisEngine.curTemplate;
                        TplProperty old = VisEngine.curTplProp;
                        VisEngine.curTemplate = tpl;
                        VisEngine.curTplProp = targetProp; // Prepare for errors.
                        try {
                            res = targetProp.expr.Eval(this, tpl);
                        }
                        catch (Exception ex) {
                            ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp,
                                ex.InnerException != null ? ex.InnerException.Message : ex.Message, VisEngine.curExprToken, this.Index());
                            res = null;
                        }
                        if (res == null) res = defaultVal;
                        VisEngine.curTemplate = oldTpl; VisEngine.curTplProp = old;
                        if (VisEngine.circularRef != null) {
                            if (targetProp != VisEngine.curTplProp) { // This is an internal reference from property to property in the same component.
                                VisEngine.circularLevel++;
                                ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, tpl, VisEngine.curTplProp,
                                    "Internal reference to " + targetProp.Name + ", level " + VisEngine.circularLevel, null, this.Index()); // Token = null: Refer to the entire property.
                            }
                        }
                    }
                }
                if (targetProp.expr.Init) instProp.state = State.Initialized; // Property setters set state to Updated in other cases.
            } else
                instProp.state = State.Initialized; // No formula. Don't recalcuate at refresh. 
            return res;
        }

        public object[] EvalListProp(TplProperty targetProp, CmpProp instProp, dynamic oldList) {
            /* Evaluates the formula. Updates and returns oldList. Handles circular references. Assumes that propState is Old or Visited.
             * Makes the property state Updated. In soft refresh states (ValueChange and Requery), recalculates all elements except Init-elements.
             * In a hard refresh state (FormulaChange) recalculates all elements.
             * If there is no expression, returns a list with zero elements. If there is a simple expression, returns a list with one element. 
             * The code attempts to deal with internal dependencies between the elements by means of a list of artificial CmpProps, but this has not been tested.
            */

            List<Expression> expList;
            if (instProp.state == State.Visited) {
                ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, tpl, targetProp, "End of circular reference, level 0", null, this.Index());
                VisEngine.circularRef = instProp; VisEngine.circularLevel = 0;
                instProp.state = State.Updated;
                return new object [0] ; // Treat it as a list with zero elements.
            } 
            instProp.state = State.Visited;
            
            if (targetProp.expr == null) { // No list
                expList = new List<Expression>(1); // Treat it as a list with zero elements.
            } else if (targetProp.expr is ExpressionList)
                expList=(targetProp.expr as ExpressionList).expressions;
            else { // Simple expression. Treat it as a list of one.
                expList = new List<Expression>(1);
                expList.Add(targetProp.expr);
            }

            object [] valList = new object[expList.Count]; // Holds the new values as an array of objects. Caller will make type conversion.
            CmpProp [] instPropList = new CmpProp[expList.Count]; // Holds the States of the elements to allow crossrefs between them.
            instProp.Value = instPropList; // During evaluation of the expression list, keep the value states in the instance property to handle circular references.

            Template oldTpl = VisEngine.curTemplate;
            TplProperty old = VisEngine.curTplProp;
            VisEngine.curTemplate = tpl;
            VisEngine.curTplProp = targetProp; // Prepare for errors.
            for (int i = 0; i < expList.Count; i++) { // At value change, set non-init values. At formula change, set all values. 
                instPropList[i] = new CmpProp(targetProp, this);
                if (i >= oldList.Length || !expList[i].Init || VisEngine.refreshMode > VisEngine.RefreshMode.Requery) {
                    if (instPropList[i].state == State.Old) { // Might have been updated through a property ref with a subscript.
                        instPropList[i].state = State.Visited;
                        valList[i] = expList[i].Eval(this, tpl);
                        if (VisEngine.circularRef != null) {
                            if (targetProp != VisEngine.curTplProp) { // This is an internal reference from property to property in the same component.
                                VisEngine.circularLevel++;
                                ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, tpl, VisEngine.curTplProp,
                                    "Internal reference to " + targetProp.Name + ", level " + VisEngine.circularLevel, null, this.Index()); // Token = null: Refer to the entire property.
                                break;
                            }
                        }
                        instPropList[i].state = State.Updated; // State initialized handled below. 
                    }
                } else { // An Init value. Reuse the old value.
                    valList[i] = (object)oldList[i];
                    instPropList[i].state = State.Initialized;
                }
            }
            instProp.Value = valList; // The result for a DesignProp.
            VisEngine.curTemplate = oldTpl; VisEngine.curTplProp = old;
            instProp.state = State.Updated;
            return valList; 
        }

        public int RelativeY(int y) { // Transforms y in uPx to Y in physical pixels. This is the address used for painting on the canvas.
            if (canvas == null)
                return (int)(y * SystemItems.Zoom2+0.5);  // This is a Form. 
            ScrollableControl panel = canvas.GuiInstance as ScrollableControl;
            if (panel == null) return (int)(y * SystemItems.Zoom2+0.5);
            return (int)(y * SystemItems.Zoom2+ 0.5) - panel.VerticalScroll.Value; 
        }

        public int RelativeX(int x) { // Transforms x in uPx to Y in physical pixels. This is the address used for painting on the canvas.
            if (canvas == null) 
                return (int)(x * SystemItems.Zoom2 + 0.5); // This is a Form. 
            ScrollableControl panel = canvas.GuiInstance as ScrollableControl;
            if (panel == null) return (int)(x * SystemItems.Zoom2 + 0.5);
            return (int)(x * SystemItems.Zoom2 + 0.5) - panel.HorizontalScroll.Value;
        }

        public static int PhyS(int s) { // Transforms size s in uPx to physical pixels. This is the value used for painting. 
            return (int)(s * SystemItems.Zoom2 + 0.5);
        }

        public static int LogS(int s) { // Transforms size s in physical pixels to uPx. This is the value used in uVis formulas. 
            return (int)(s / SystemItems.Zoom2 + 0.5);
        }

        public int LogicalY(int y) { // Transforms y in physical pixels to uPx. This is the value used in uVis formulas.
            if (canvas == null) 
                return (int)(y/SystemItems.Zoom2); // This is a Form. 
            ScrollableControl panel = canvas.GuiInstance as ScrollableControl;
            if (panel == null) return (int)(y / SystemItems.Zoom2 + 0.5);
            return (int)((y + panel.VerticalScroll.Value) / SystemItems.Zoom2 + 0.5);
        }

        public int LogicalX(int x) { // Transforms x in physical pixels to uPx. This is the value used in uVis formulas.
            if (canvas == null)
                return (int)(x / SystemItems.Zoom2); // This is a Form. 
            ScrollableControl panel = canvas.GuiInstance as ScrollableControl;
            if (panel == null) return (int)(x / SystemItems.Zoom2);
            return (int)((x + panel.HorizontalScroll.Value) / SystemItems.Zoom2 + 0.5);
        }

        public Point PointToScreen(Point point) {
            if (this is VisForm)
                return this.GuiInstance.PointToScreen(point);
            var shapePoint = new Point(point.X + this.RelativeX(this.left), point.Y + this.RelativeY(this.top));
            if (this.canvas.GuiInstance.IsDisposed)
                return point; // May happen when component is refreshed.
            return this.canvas.GuiInstance.PointToScreen(shapePoint);
        }
        
        public int toInt32(object val, int Default) {
            try {
                int res = Convert.ToInt32(val);
                return res;
            }
            catch (Exception) {
                ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp,
                    "Cannot convert " + val.GetType().ToString() + " to integer", VisEngine.curTplProp.startChain, this.Index());
                return Default;
            }
        }

        public int toPosInt32(object val, int Default) {
            try {
                int res = Convert.ToInt32(val);
                if (res < 0) {
                    if (Default >= 0) // If Default is <0, the caller expects a negative result for further error handling.
                        ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp,
                            VisEngine.curTplProp.Name + " is " + res + ". Should be >= 0", VisEngine.curTplProp.startChain, this.Index());
                    return Default;
                }
                return res;
            }
            catch (Exception) {
                ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp,
                    "Cannot convert " + val.GetType().ToString() + " to integer", VisEngine.curTplProp.startChain, this.Index());
                return Default;
            }
        }

        public double toDouble(object val, double Default) {
            try {
                return Convert.ToDouble(val);
            }
            catch (Exception) {
                ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp,
                    "Cannot convert " + val.GetType().ToString() + " to double floating point", VisEngine.curTplProp.startChain, this.Index());
                return Default;
            }
        }

        public DateTime toDateTime(object val, DateTime Default) {
            try {
                return Convert.ToDateTime(val);
            }
            catch (Exception) {
                ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp,
                    "Cannot convert " + val.GetType().ToString() + " to double floating point", VisEngine.curTplProp.startChain, this.Index());
                return Default;
            }
        }

        public string toString(object val, string Default) {
            try {
                if (val == null) return null;
                return val.ToString();
            }
            catch (Exception) {
                ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp,
                    "Cannot convert " + val.GetType().ToString() + " to string", VisEngine.curTplProp.startChain, this.Index());
                return Default;
            }
        }

        public bool toBool(object val, bool Default) {
            if (val is bool) return (bool)val;
            ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp,
                "Cannot convert " + val.GetType().ToString() + " to boolean", VisEngine.curTplProp.startChain, this.Index());
            return Default;
        }

        public Color toColor(object val, Color Default) {
            if (val is Color) return (Color)val;
            ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp,
                "Cannot convert " + val.GetType().ToString() + " to Color", VisEngine.curTplProp.startChain, this.Index());
            return Default;
        }

        public ContainerComponent toCanvas(object val, ContainerComponent Default) {
            if (val is ContainerComponent) return (val as ContainerComponent);
            ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp,
                "Cannot convert " + val.GetType().ToString() + " to Canvas", VisEngine.curTplProp.startChain, this.Index());
            return Default;
        }

        public VisComponent toComponent(object val, VisComponent Default) {
            if (val == null) return Default;
            if (val is VisComponent) return (val as VisComponent);
            ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp,
                "Cannot convert " + val.GetType().ToString() + " to VisComponent", VisEngine.curTplProp.startChain, this.Index());
            return Default;
        }

        // ---------------------------------------------------------------------------------------------------------
        #region VisibleProperties
        public static int NameID; // Property index. Set at initialization in ReflectDLL. 
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, LongHelp = "The name of my template.",
            PropertyCategory = "Common", ReadOnly = false, ShowInitially = true, Position = 1,
            ShortHelp = "The name of my template.")]
        public string Name {
            get { return tpl.Name; }
        }

        public static int RowsID; // Property index. Set at initialization in ReflectDLL. 
        [VisPropAtt(VisPropAtt.PropertyType.Property, null,
            InPopup = false, PropertyCategory = "Common", ReadOnly = false, ShowInitially = true, Position = 4,
            ShortHelp = "Create one of me for each data row.")]
        public virtual RecordSetExpr Rows {
            get { return tpl.RowTpl.RowsExpr; }
        }

        public virtual int topDefault { get {return 5; } } // Property to allow overriding.
        public static int TopID; // Property index. Set at initialization in ReflectDLL. 
        public int top;
        [VisPropAtt(VisPropAtt.PropertyType.Property, "5", ShortHelp = "Number of pixels from my canvas top to my top inside the border.",
            PropertyCategory = "Common", ReadOnly = false, ShowInitially = true,
            LongHelp = "Pixels from my canvas top to my top inside the border.", Position = 5)]
        public virtual int Top {
            get {
                CmpProp topProp = this.propertyValues[TopID];
                if (topProp.state >= State.Updated) return top;
                if (tpl.propTop.expr != null) { // Top has a formula.
                    Top = toInt32(EvalProp(TopID, topDefault), topDefault);
                    // Makes topState Initialized if there is no formula or an Init-formula, Updated otherwise. 
                } else { // Top has no formula. Calculate it from height and bottom.
                    CmpProp heightProp = this.propertyValues[HeightID];
                    CmpProp bottomProp = this.propertyValues[BottomID];
                    topProp.state = State.Visited; // Start internal calculation.
                    // Although a null formula means Init null, it still has an internal formula that is now Visited.
                    if (heightProp.state < State.Updated) {
                        if ((heightProp.tplProp as TplPropValue).PropInfo.VisAttribute.ReadOnly)
                            height = Height; // Height has an internal fomula.
                        else
                            Height = toPosInt32(EvalProp(HeightID, heightDefault), heightDefault);
                    }
                    if (bottomProp.state < State.Updated) {
                        if ((bottomProp.tplProp as TplPropValue).PropInfo.VisAttribute.ReadOnly)
                            bottom = Bottom; // Height has an internal fomula.
                        else
                            Bottom = toInt32(EvalProp(BottomID, topDefault + height), topDefault + height);
                    }
                    Top = bottom - height; // Sets state to updated. There was a "formula" without Init.
                }
                return top;
            }

            set {
                top = value; SetUpdated(ref propertyValues[TopID]); 
                // if (GuiInstance != null) GuiInstance.Top = value; // Don't know the scroll yet. Set in assignCanvas. Caused flicker.
            }
        }

        public virtual int heightDefault { get { return 30; } } // Property to allow overriding.
        public static int HeightID; // Property index. Set at initialization in ReflectDLL. 
        public int height;
        [VisPropAtt(VisPropAtt.PropertyType.Property, "30", LongHelp = "Number of pixels from my top to my bottom inside the border.",
            PropertyCategory = "Common", ReadOnly = false, ShowInitially = true, ShortHelp = "Pixels from my top to my bottom inside the border.", Position = 6)]
        public virtual int Height {
            get {
                CmpProp heightProp = this.propertyValues[HeightID];
                if (heightProp.state >= State.Updated) return height;
                CmpProp topProp = this.propertyValues[TopID];
                CmpProp bottomProp = this.propertyValues[BottomID];
                if (tpl.propHeight.expr != null) { // || tpl.propTop.expr == null || tpl.propBottom.expr == null) { // Height has a formula or Top or Bottom lack one.
                    Height = toPosInt32(EvalProp(HeightID, heightDefault), heightDefault);
                } else { // Height has no formula. Calculate it from top and bottom. //, both of which have a formula.
                    heightProp.state = State.Visited; // Start internal calculation.
                    if (topProp.state < State.Updated) {
                        if ((topProp.tplProp as TplPropValue).PropInfo.VisAttribute.ReadOnly)
                            top = Top; // Top has an internal fomula.
                        else
                            Top = toInt32(EvalProp(TopID, topDefault), topDefault);
                    }
                    if (bottomProp.state < State.Updated) {
                        if ((bottomProp.tplProp as TplPropValue).PropInfo.VisAttribute.ReadOnly)
                            bottom = Bottom; // Height has an internal fomula.
                        else
                            Bottom = toInt32(EvalProp(BottomID, top + heightDefault), top + heightDefault);
                    }
                    Height = bottom - top; // Sets state to updated. There was a "formula" without Init.
                }
                return height;
            }

            set {
                height = value; SetUpdated(ref propertyValues[HeightID]);
                if (GuiInstance != null) GuiInstance.Height = (int) (value * SystemItems.Zoom2);
            }
        }

        public static int BottomID; // Property index. Set at initialization in ReflectDLL. 
        public int bottom;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, LongHelp = "Number of pixels from my canvas top to my bottom inside the border.",
            PropertyCategory = "Common", ReadOnly = false, ShowInitially = false,
            ShortHelp = "Pixels from my canvas top to my bottom inside the border.", Position = 7)]
        public virtual int Bottom {
            get {
                CmpProp bottomProp = this.propertyValues[BottomID];
                if (bottomProp.state >= State.Updated) return bottom;
                if (tpl.propBottom.expr != null) { // Bottom has a formula.
                    Bottom = toInt32(EvalProp(BottomID, topDefault + heightDefault), topDefault + heightDefault);
                } else { // Bottom has no formula. Calculate it from top and height.
                    CmpProp heightProp = this.propertyValues[HeightID];
                    CmpProp topProp = this.propertyValues[TopID];
                    bottomProp.state = State.Visited; // Start internal calculation.
                    if (topProp.state < State.Updated) {
                        if ((topProp.tplProp as TplPropValue).PropInfo.VisAttribute.ReadOnly)
                            top = Top; // Top has an internal fomula.
                        else
                            Top = toInt32(EvalProp(TopID, topDefault), topDefault);
                    }
                    if (heightProp.state < State.Updated) {
                        if ((heightProp.tplProp as TplPropValue).PropInfo.VisAttribute.ReadOnly)
                            height = Height; // Height has an internal fomula.
                        else
                            Height = toPosInt32(EvalProp(HeightID, heightDefault), heightDefault);
                    }
                    Bottom = top + height; // There was a "formula" without Init.
                }
                return bottom;
            }

            set {
                bottom = value; SetUpdated(ref propertyValues[BottomID]); 
            }
        }

        public virtual int leftDefault { get { return 5; } } // Property to allow overriding.
        public static int LeftID; // Property index. Set at initialization in ReflectDLL. 
        public int left;
        [VisPropAtt(VisPropAtt.PropertyType.Property, "5", LongHelp = "Number of pixels from my canvas left to my left inside the border.",
            PropertyCategory = "Common", ReadOnly = false, ShowInitially = true,
            ShortHelp = "Pixels from my canvas left to my left inside the border.", Position = 8)]
        public virtual int Left {
            get {
                CmpProp leftProp = this.propertyValues[LeftID];
                if (leftProp.state >= State.Updated) return left;
                if (tpl.propLeft.expr != null) { // Left has a formula.
                    Left = toInt32(EvalProp(LeftID, leftDefault), leftDefault);
                } else { // Left has no formula. Calculate it from width and right.
                    CmpProp widthProp = this.propertyValues[WidthID];
                    CmpProp rightProp = this.propertyValues[RightID];
                    leftProp.state = State.Visited; // Start internal calculation.
                    if (widthProp.state < State.Updated) {
                        if ((widthProp.tplProp as TplPropValue).PropInfo.VisAttribute.ReadOnly)
                            width = Width; // Width has an internal fomula.
                        else
                            Width = toPosInt32(EvalProp(WidthID, widthDefault), widthDefault);
                    }
                    if (rightProp.state < State.Updated) {
                        if ((rightProp.tplProp as TplPropValue).PropInfo.VisAttribute.ReadOnly)
                            right = Right; // Height has an internal fomula.
                        else
                            Right = toInt32(EvalProp(RightID, leftDefault + width), leftDefault + width);
                    }
                    Left = right - width; // There was a "formula" without Init.
                }
                return left;
            }

            set {
                left = value; SetUpdated(ref propertyValues[LeftID]); 
                // if (GuiInstance != null) GuiInstance.Left = value; // Don't know the scroll yet. Set in assignCanvas. Caused flicker.
            }
        }

        public virtual int widthDefault { get { return 50; } } // Property to allow overriding.
        public static int WidthID; // Property index. Set at initialization in ReflectDLL. 
        public int width;
        [VisPropAtt(VisPropAtt.PropertyType.Property, "50", PropertyCategory = "Common", ReadOnly = false, ShowInitially = true,
            ShortHelp = "Pixels from my left to my right inside the border.", Position = 9)]
        public virtual int Width {
            get {
                CmpProp widthProp = this.propertyValues[WidthID];
                if (widthProp.state >= State.Updated) return width;
                CmpProp leftProp = this.propertyValues[LeftID];
                CmpProp rightProp = this.propertyValues[RightID];
                if (tpl.propWidth.expr != null) { // || tpl.propLeft.expr == null || tpl.propRight.expr == null) { // Width has a formula or Left or Right lack one.
                    Width = toPosInt32(EvalProp(WidthID, widthDefault), widthDefault);
                } else { // Width has no formula. Calculate it from left and right, both of which have a formula
                    widthProp.state = State.Visited; // Start internal calculation.
                    if (leftProp.state < State.Updated) {
                        if ((leftProp.tplProp as TplPropValue).PropInfo.VisAttribute.ReadOnly)
                            left = Left; // Height has an internal fomula.
                        else
                            Left = toInt32(EvalProp(LeftID, leftDefault), leftDefault);
                    }
                    if (rightProp.state < State.Updated) {
                        if ((rightProp.tplProp as TplPropValue).PropInfo.VisAttribute.ReadOnly)
                            right = Right; // Height has an internal fomula.
                        else
                            Right = toInt32(EvalProp(RightID, left + widthDefault), left + widthDefault);
                    }
                    Width = right - left; // There was a "formula" without Init.
                }
                return width;
            }

            set {
                width = value; SetUpdated(ref propertyValues[WidthID]);
                if (GuiInstance != null) GuiInstance.Width = (int)(value * SystemItems.Zoom2);
            }
        }

        public static int RightID; // Property index. Set at initialization in ReflectDLL. 
        public int right;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Common", ReadOnly = false, ShowInitially = false,
            ShortHelp = "Pixels from my canvas left to my right inside the border.", Position = 10)]
        public virtual int Right {
            get {
                CmpProp rightProp = this.propertyValues[RightID];
                if (rightProp.state >= State.Updated) return right;
                if (tpl.propRight.expr != null) { // Right has a formula.
                    Right = toInt32(EvalProp(RightID, 100), 100);
                } else { // Right has no Formula. Calculate it from left and width.
                    CmpProp widthProp = this.propertyValues[WidthID];
                    CmpProp leftProp = this.propertyValues[LeftID];
                    rightProp.state = State.Visited; // Start internal calculation.
                    if (leftProp.state < State.Updated) {
                        if ((leftProp.tplProp as TplPropValue).PropInfo.VisAttribute.ReadOnly)
                            left = Left; // Height has an internal fomula.
                        else
                            Left = toInt32(EvalProp(LeftID, leftDefault), leftDefault);
                    }
                    if (widthProp.state < State.Updated) {
                        if ((widthProp.tplProp as TplPropValue).PropInfo.VisAttribute.ReadOnly)
                            width = Width; // Height has an internal fomula.
                        else
                            Width = toPosInt32(EvalProp(WidthID, widthDefault), widthDefault);
                    }
                    Right = left + width; // There was a "formula" without Init.
                }
                return right;
            }

            set {
                right = value; SetUpdated(ref propertyValues[RightID]); 
            }
        }

        /* The following methods are used by Studio drag-drop, resize, copy, etc. to set position formulas. Some components have 
         * read-only positions, e.g Polygon. They may set a related formula, e.g. Ypositions and leave the Top formula null.
         * The code uses reflection to find the subclass since a static method cannot be overridden, but has to be new.
         * 
         * tpl and oldTpl are the same except for copy.
         */
        static bool UseSubclass(ComponentInfo componentInfo, string methodName, Template tpl, Template oldTpl, int pos, int oldPos) {
            Type type = componentInfo.Type;
            while (type.GetMethod(methodName) == null) // Find a base type that can set a formula, at least VisComponent.
                type = type.BaseType;
            if (type == typeof(VisComponent)) return false;
            type.GetMethod(methodName).Invoke(null, new object[] { componentInfo, tpl, oldTpl, pos, oldPos });
            return true;
        }

        public static bool SetTopFormula(ComponentInfo componentInfo, Template tpl, Template oldTpl, int pos, int oldPos) {
            if (UseSubclass(componentInfo, "SetTopFormula", tpl, oldTpl, pos, oldPos)) return true;
            return ChangePositionFormula(tpl.propTop, oldTpl.propTop, pos, oldPos);
        }

        public static bool SetHeightFormula(ComponentInfo componentInfo, Template tpl, Template oldTpl, int pos, int oldPos) {
            if (UseSubclass(componentInfo, "SetHeightFormula", tpl, oldTpl, pos, oldPos)) return true;
            return ChangePositionFormula(tpl.propHeight, oldTpl.propHeight, pos, oldPos);
        }

        public static bool SetBottomFormula(ComponentInfo componentInfo, Template tpl, Template oldTpl, int pos, int oldPos) {
            if (UseSubclass(componentInfo, "SetBottomFormula", tpl, oldTpl, pos, oldPos)) return true;
            return ChangePositionFormula(tpl.propBottom, oldTpl.propBottom, pos, oldPos);
        }

        public static bool SetLeftFormula(ComponentInfo componentInfo, Template tpl, Template oldTpl, int pos, int oldPos) {
            if (UseSubclass(componentInfo, "SetLeftFormula", tpl, oldTpl, pos, oldPos)) return true;
            return ChangePositionFormula(tpl.propLeft, oldTpl.propLeft, pos, oldPos);
        }

        public static bool SetWidthFormula(ComponentInfo componentInfo, Template tpl, Template oldTpl, int pos, int oldPos) {
            if (UseSubclass(componentInfo, "SetWidthFormula", tpl, oldTpl, pos, oldPos)) return true;
            return ChangePositionFormula(tpl.propWidth, oldTpl.propWidth, pos, oldPos);
        }

        public static bool SetRightFormula(ComponentInfo componentInfo, Template tpl, Template oldTpl, int pos, int oldPos) {
            if (UseSubclass(componentInfo, "SetRightFormula", tpl, oldTpl, pos, oldPos)) return true;
            return ChangePositionFormula(tpl.propRight, oldTpl.propRight, pos, oldPos);
        }

        public static string AdjustConstant( TplProperty tplProp, int newValue, int old) { // newValue and old in logical pixels.
            // Test for formula = [init] [ + | - ] intConstant [ ( + | - ) expression ]. Replace [+|-] intConstant with new value.
            int logChange = newValue - old;
            Token newTerm = null;
            string formula = null;
            bool init = false;
            int term;
            Token token = tplProp.startChain;
            if (token.val.ToUpper() == "INIT") {
                token = token.next; init = true;
            }
            if (token.val == "+" || token.val == "-") {
                newTerm = token; token = token.next;
            }
            if (token.kind == Parser.TkKind.Long && (token.next.kind == Parser.TkKind.EOL || token.next.val == "+" || token.next.val == "-")) {
                term = (int)token.obj;
                if (newTerm != null && newTerm.val == "-") term = -term; // value of present constant term.
                term = term + logChange; // Value of new term.
                if (newTerm == null) newTerm = token;
                else newTerm.next = token.next; // Skip the old term token.
                newTerm.val = term.ToString();
                formula = Parser.GetFormulaFromTokens(tplProp.startChain).Trim();
            } else {

                // Test for formula = expression ( + | - ) intConstant EOL
                while (token.kind != Parser.TkKind.EOL) {
                    if (token.val == "+" || token.val == "-") {
                        newTerm = token; token = token.next;
                        if (token.kind == Parser.TkKind.Long && token.next.kind == Parser.TkKind.EOL) {
                            term = (int)token.obj;
                            if (newTerm != null && newTerm.val == "-") term = -term; // value of present constant term.
                            term = term + logChange; // Value of new term.
                            newTerm.val = (term < 0 ? "-" : "+");
                            token.val = Math.Abs(term).ToString();
                            formula = Parser.GetFormulaFromTokens(tplProp.startChain).Trim();
                            break;
                        }
                    }
                    token = token.next;
                }
            }
            if (formula == null) { // Default: generate [Init] [ - ] intConstant + expression
                if (init)
                    formula = tplProp.startChain.val.TrimEnd() + " " + logChange.ToString() + " + "; // Add constant after Init
                else
                    formula = logChange.ToString() + " + "; // Add constant first.
                formula = formula + Parser.GetFormulaFromTokens(tplProp.startChain).Trim();
            }
            return formula;
        }

        public static bool ChangePositionFormula(TplProperty tplProp, TplProperty oldProp, int newValue, int old) {
            if (newValue == old) return true;
            Expression expr = oldProp.expr;
            string formula = null;
            if (expr == null) { // Space only. Insert value before space.
                VisEngine.CreateOrSetTemplateProperty(tplProp.tpl, tplProp.Name, newValue.ToString() +
                    (oldProp.startChain == null ? "" : " " + oldProp.startChain.space));
                return true;
            }
            formula = AdjustConstant(tplProp, newValue, old);
            VisEngine.CreateOrSetTemplateProperty(tplProp.tpl, tplProp.Name, formula);
            return true;
        }

        public virtual Color backColorDefault { get { return Color.White; } } // Property to allow overriding.
        public static int BackColorID; // Property index. Set at initialization in ReflectDLL. 
        public Color backColor;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Common", ReadOnly = false, ShowInitially = false,
            ShortHelp = "The color of my inner area. Default: White.", Position = 11)]
        public virtual Color BackColor {
            get {
                CmpProp backColorProp = this.propertyValues[BackColorID];
                if (backColorProp.state >= State.Updated) return backColor;
                BackColor = toColor(EvalProp(BackColorID, backColorDefault), backColorDefault);
                return backColor;
            }
            set {
                backColor = value; SetUpdated(ref propertyValues[BackColorID]); 
                if (GuiInstance != null) { GuiInstance.BackColor = value; } 
            }
        }

        public virtual Color borderColorDefault { get { return Color.Black; } } // Property to allow overriding.
        public static int BorderColorID; // Property index. Set at initialization in ReflectDLL. 
        public Color borderColor;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Common", ReadOnly = false, ShowInitially = false,
            ShortHelp = "The color of my border. Default: Black.", Position = 12)]
        public virtual Color BorderColor {
            get {
                CmpProp borderColorProp = this.propertyValues[BorderColorID];
                if (borderColorProp.state >= State.Updated) return borderColor;
                BorderColor = toColor(EvalProp(BorderColorID, borderColorDefault), borderColorDefault);
                return borderColor;
            }
            set { borderColor = value; SetUpdated(ref propertyValues[BorderColorID]); }
        }

        public virtual int borderWeightDefault { get { return 1; } } // Property to allow overriding.
        public static int BorderWeightID; // Property index. Set at initialization in ReflectDLL. 
        public int borderWeight;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Common", ReadOnly = false, ShowInitially = false,
            ShortHelp = "Pixels across my border. Default: 1.", Position = 13)]
        public virtual int BorderWeight {
            get {
                CmpProp borderWeightProp = this.propertyValues[BorderWeightID];
                if (borderWeightProp.state >= State.Updated) return borderWeight;
                BorderWeight = toPosInt32(EvalProp(BorderWeightID, borderWeightDefault), borderWeightDefault);
                return borderWeight;
            }
            set { borderWeight = value; SetUpdated(ref propertyValues[BorderWeightID]); }
        }

        public static int EnabledID; // Property index. Set at initialization in ReflectDLL. 
        public bool enabled = true;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Common", ReadOnly = false, Position = 14,
            ShortHelp = "When false, I ignore key press and delete/backspace. For text boxes it means that the user cannot edit the text. Default true.")]
        public bool Enabled {
            get {
                CmpProp enabledProp = this.propertyValues[EnabledID];
                if (enabledProp.state >= State.Updated) return enabled;
                Enabled = toBool(EvalProp(EnabledID, true), true);
                return enabled;
            }
            set { enabled = value; SetUpdated(ref propertyValues[EnabledID]); }
        }

        public static int VisibleID; // Property index. Set at initialization in ReflectDLL. 
        public bool visible = true;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, LongHelp = "False if I am not visible.",
            PropertyCategory = "Common", ReadOnly = false, ShowInitially = false, ShortHelp = "False if I am not visible.", Position = 15)]
        public virtual bool Visible {
            get {
                CmpProp visibleProp = this.propertyValues[VisibleID];
                if (visibleProp.state >= State.Updated) return visible;
                Visible = toBool(EvalProp(VisibleID, true), true);
                return visible;
            }
            set {
                visible = value; SetUpdated(ref propertyValues[VisibleID]); 
                if (GuiInstance != null) GuiInstance.Visible = value; }
        }

        public static int CanvasID; // Property index. Set at initialization in ReflectDLL. 
        public ContainerComponent canvas = null;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null,
            LongHelp = "The component (canvas) where I am painted. Default: the Form.\nIt may scroll and clip me, but my Left and Top don't change.",
            PropertyCategory = "Common", ReadOnly = false, ShowInitially = false,
            ShortHelp = "The component (canvas) where I am painted. Default: the Form.", Position = 16)]
        public virtual ContainerComponent Canvas {
            get {
                CmpProp canvasProp = this.propertyValues[CanvasID];
                if (canvas != null && canvasProp.state >= State.Updated) return canvas; // Canvas is null right after refresh.
                VisForm frm = this.Form(); // When opening a form with zero instances, frm will be null.
                Canvas = toCanvas(EvalProp(CanvasID, frm), frm);
                // A null canvas becomes Initialized, but recalculation is needed since the component has been removed from the canvas.
                return canvas;
            }
            set {
                // To avoid cyclical references, don't assign to canvas and adjust virtual size until end of property calculation.
                canvas = value;
                SetUpdated(ref propertyValues[CanvasID]);
            }
        }

        public void AssignAndAdjustCanvas(ContainerComponent newCanvas) { // Used at end of property calculation. 
            // Assumes that all properties have been calculated and that a null canvas has been replaced by VisForm.
            AddToCanvas(newCanvas);

            // Adjust virtual size to make the scrollable area include this shape.
            if (visible) { // May cause circular references, so do it only after calculating all properties of the component.
                if (bottom > canvas.virtualHeight)
                    canvas.virtualHeight = Bottom;
                if (right > canvas.virtualWidth)
                    canvas.virtualWidth = Right;
            }
            // Don't set corner.Top = virtualHeight here. It causes the canvas to scroll whenever virtualHeight is set. 
            // Furthermore, the canvas loses track of the scrolls.
        }

        public void AddToCanvas(ContainerComponent newCanvas) { 
            // Used by ordinary components, but also anonymous components that don't need to adjust canvas size.
            if (newCanvas == null) newCanvas = this.Form();
            if (newCanvas != canvas) {
                if (canvas != null)
                    canvas.RemoveVisComponent(this);
                canvas = newCanvas;
            }
            canvas.AddVisComponent(this); // Adds component and maybe GuiInstance. Idempotent.
        }

        public virtual int zOrderDefault { get { return 0; } } // Property to allow overriding.
        public static int ZOrderID; // Property index. Set at initialization in ReflectDLL. 
        public int zOrder;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, LongHelp = "On my canvas, I am above components with lower ZOrder.",
            PropertyCategory = "Common", ReadOnly = false, ShowInitially = false, Position = 17,
            ShortHelp = "On my canvas, I am above components with lower ZOrder.")]
        public virtual int ZOrder {
            get {
                CmpProp zOrderProp = this.propertyValues[ZOrderID];
                if (zOrderProp.state >= State.Updated) return zOrder;
                ZOrder = toInt32(EvalProp(ZOrderID, zOrderDefault), zOrderDefault);
                return zOrder;
            }
            set { zOrder = value; SetUpdated(ref propertyValues[ZOrderID]); }
        }

        public static int TabOrderID; // Property index. Set at initialization in ReflectDLL. 
        private int tabOrder = 0; // GuiInstance.TabStop is set to false in the appropriate Control constructors.
        [VisPropAtt(VisPropAtt.PropertyType.Property, null,
            LongHelp = @"> 0. When tabbing through components on my canvas, I will be reached before components with higher TabOrder. Default: 0 (no tab stop). 
A cLabel with TabOrder will pass the tab on to the next TabOrder, e.g. a text box. A cLabel can have a short-cut key.",
            PropertyCategory = "Common", ReadOnly = false, ShowInitially = false, Position = 18,
            ShortHelp = "> 0. When tabbing through components on my canvas, I will be reached before components with higher TabOrder. Default: 0 (no tab stop).")]
        public virtual int TabOrder {
            get {
                CmpProp tabOrderProp = this.propertyValues[TabOrderID];
                if (tabOrderProp.state >= State.Updated) return tabOrder;
                TabOrder = toPosInt32(EvalProp(TabOrderID, 0), 0);
                return tabOrder;
            }
            set {
                SetUpdated(ref propertyValues[TabOrderID]); 
                tabOrder = value; 
                if (this.GuiInstance == null) return;
                if (tabOrder <= 0)
                    this.GuiInstance.TabStop = false;
                else {
                    this.GuiInstance.TabIndex = (int)tabOrder;
                    this.GuiInstance.TabStop = true;
                }
            }
        }

        public static int MovableID; // Property index. Set at initialization in ReflectDLL. 
        public bool movable = false;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, LongHelp = "True when the user can move the component.", Position = 19,
            PropertyCategory = "Common", ReadOnly = false, ShowInitially = false, ShortHelp = "True when the user can move me.")]
        public bool Movable {
            get {
                CmpProp movableProp = this.propertyValues[MovableID];
                if (movableProp.state >= State.Updated) return movable;
                Movable = toBool(EvalProp(MovableID, false), false);
                return movable;
            }
            set { movable = value; SetUpdated(ref propertyValues[MovableID]); }
        }

        public virtual bool InClickArea(Point point) {
            int Htrim = PhyS(width) >= 4 ? 0 : (5 - PhyS(width)) / 2; // Make sure that large components such as canvas don't steal neighbor areas.
            int Vtrim = PhyS(height) >= 4 ? 0 : (5 - PhyS(height)) / 2;
            // Notice that left+width is just outside the client area.
            if (point.X >= RelativeX(left - Htrim) && point.X < RelativeX(left) + PhyS(width) + Htrim && point.Y < RelativeY(top) + PhyS(height) + Vtrim && point.Y >= RelativeY(top) - Vtrim)
                return true;
            return false;
        }

        public virtual bool BiggerSize(VisComponent comp1, VisComponent comp2) { // Which of two overlapping components to select.
            int size1 = Math.Max(comp1.width, comp1.height); // A long line comes after a small rectangle. SLa 23-01-16.
            int size2 = Math.Max(comp2.width, comp2.height);
            return size1 < size2;
        }

        [VisPropAtt(VisPropAtt.PropertyType.Method, null, ShowInitially = false,
            ShortHelp = @"Bundle(sourceTpl !findProp) Find the bundle that refers to me through tpl !findProp, 
e.g. Me.Bundle(boxItem !heading)[Last]")]
        public BundleRef Bundle(Expression propRef) {
            TplProperty tplProp = null;
            DesignPropRef dRef = propRef as DesignPropRef;
            ValPropertyRef vRef = propRef as ValPropertyRef;
            if (dRef != null) tplProp = dRef.tplProp;
            else if (vRef != null) tplProp = vRef.tplProp;
            if (tplProp == null || !(tplProp.expr is FindExpr)) {
                Token t = VisEngine.curExprToken;
                ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, this.tpl, VisEngine.curTplProp, "The parameter doesn't refer to a Find-property", t, this.Index());
                return null;
            }
            FindExpr findExpr = tplProp.expr as FindExpr;

            int targetCount = this.tpl.findProperties.IndexOf(tplProp);
            if (targetCount < 0) {
                Token t = VisEngine.curExprToken;
                ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, this.tpl, VisEngine.curTplProp, "The Find-property doesn't refer to me", t, this.Index());
                return null;
            }

            if (!tplProp.tpl.Updated) VisEngine.CreateInstances(tplProp.tpl); // Make sure that source components and bundles are created 
            Bundle source = tplProp.tpl.RowBundle;
            if (source.Instances.Count == 0) return null; // No source components.
            VisComponent sourceCmp = source.Instances[0].cmp; // Any component will do for initializing bundles.
            findExpr.Eval(sourceCmp, sourceCmp.tpl);
            
            BundleRef result = new BundleRef(this.childBundles[targetCount], 0, null);
            return result; // Will be null if the source doesn't refer to me.
        }

        #endregion VisibleProperties
        // ---------------------------------------------------------------------------------------

        public void ExecuteHandler(string handlerName, bool DefaultHandled) { // DefaultHandled isn't used. The default for Handled is always false
            // Tests whether I have an event handler with this name, then evaluates it. Repeats it through the canvas hierarchy until Handled is true. 
            Template curTpl = this.tpl;
            if (curTpl.FormTpl.updating)
                return; // Don't perform event handlers during refresh.
            SystemItems.Handled = false;
            VisComponent current = this;
            while (current != null) {
                curTpl = current.tpl;
                EvtInfo evtInfo = null;
                if (curTpl.Info.EventByName.TryGetValue(handlerName, out evtInfo)) {
                    TplPropEvent tplProp = curTpl.evtPropArray[evtInfo.propIndex] as TplPropEvent;
                    if (tplProp.expr != null) {
                        try {
                            VisEngine.curTemplate = curTpl; // Prepare for errors.
                            VisEngine.curTplProp = curTpl.evtPropArray[evtInfo.propIndex];
                            tplProp.expr.Eval(current, curTpl); // May change Handled.
                            if (SystemItems.Handled) return;
                        }
                        catch (LoggedException) { }
                        catch (Exception e) { 
                            Token t = VisEngine.curExprToken;
                            ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, curTpl, tplProp, e.Message, t, this.Index());
                            return; // Don't propagate errors.
                        }
                        if (SystemItems.Handled) return;
                    }
                }
                current = current.canvas; // Try next canvas in the hierarchy.
                // The next two lines are for passing events through the z-order stack.
                //if (current.canvas == null) break; // Form.
                //current = current.canvas.getNextTarget();
            }
            return;
        }

        public void AttachControlEvents(Control ctrl) { // Routes events from Windows controls to the VisComponent
            ctrl.MouseEnter += MouseEnter;
            ctrl.MouseMove += MouseMove;
            ctrl.MouseLeave += MouseLeave;
            ctrl.MouseHover += MouseHover;
            ctrl.MouseDown += MouseDown;
            ctrl.MouseUp += MouseUp;
            ctrl.GotFocus += GotFocus;
            ctrl.LostFocus += LostFocus;
            ctrl.Click += Click;
            ctrl.DoubleClick += DoubleClick;
            /* LocationChanged, SizeChanged: Windows generates these events whenever Left, Width, etc. are set, e.g. during calculation and move.
             * UVis uses Moved and Resized instead. They occur only after user actions. Move must be at least a few pixels. Otherwise it will be a Click.
             * In order for the event handler to test stuff, SystemItems.OldLeft, etc. keep track of the values before the move.
             */
            ctrl.KeyDown += KeyDown;
            ctrl.KeyPress += KeyPress;
            ctrl.KeyUp += KeyUp;
        }

        public void SetKeyFocus( ) {
            VisForm frm = this.Form(); frm.keyFocus = this;
            //Debug.Print("SetKeyFocus to " + frm.Name + "." + this.tpl.Name);
        }

        public VisComponent GetKeyFocus(VisComponent cmp) {
            if (cmp is VisForm) return ((VisForm)cmp).keyFocus;
            return this.Form().keyFocus; 
        }

        protected void updateKeyVariables(KeyEventArgs e) {
            SystemItems.Key = e.KeyData;
            SystemItems.AltPressed = e.Alt;
            SystemItems.CtrlPressed = e.Control;
            SystemItems.ShiftPressed = e.Shift;
        }

        protected void updatePressed() {
            SystemItems.AltPressed = (Control.ModifierKeys == System.Windows.Forms.Keys.Alt);
            SystemItems.CtrlPressed = (Control.ModifierKeys == System.Windows.Forms.Keys.Control);
            SystemItems.ShiftPressed = (Control.ModifierKeys == System.Windows.Forms.Keys.Shift);
        }

        protected void updateMouseVariables(MouseEventArgs e) {
            updatePressed();
            SystemItems.MouseX = LogicalX(e.X); // Compensate for scroll. Works for a Form too.
            SystemItems.MouseY = LogicalY(e.Y); // Compensate for scroll.
            SystemItems.MouseButton = e.Button;
            SystemItems.MouseDelta = e.Delta;
        }

        public virtual void Paint(PaintEventArgs e) {
        }

        public virtual void Dispose() {
            this.Disposed = true;
            if (this.GuiInstance != null && this.GuiInstance.Parent != null) {
                //this.GuiInstance.Parent.Controls.Remove(GuiInstance);
                this.GuiInstance.Dispose(); // Also  removes GuiInstance from Parent.Controls.
            }
            if (canvas != null) {
                canvas.dictComponents.Remove(this);
                canvas.scrollAndOrderUpdated = false;
            }
        }

        #region EVENTHANDLING
        // The following methods serve as event handlers. They are called directly from the subclasses, e.g. Box and ContainerComponent.
        // They are not called through yet another event.

        // MouseEnter occurs before the first MouseMove in the component.
        [VisPropAtt(VisPropAtt.PropertyType.Eventhandler, null, Position = 10,
            LongHelp = @"The mouse pointer points at me.
Is followed by MouseLeave for the previous component and MouseMove for me.",
            PropertyCategory = "EventHandler", ReadOnly = false, ShortHelp = "The mouse pointer points at me.")]
        public virtual void MouseEnter(object sender, EventArgs e) {
            this.ExecuteHandler("MouseEnter", true);
        }

        [VisPropAtt(VisPropAtt.PropertyType.Eventhandler, null, Position = 11,
            LongHelp = @"The mouse pointer moved over me.
Depending on the move speed, the system generates more or less mouse move events.",
            PropertyCategory = "EventHandler", ReadOnly = false, ShortHelp = "The mouse pointer moved over me.")]
        public virtual void MouseMove(object sender, MouseEventArgs e) {
            if (this.Moving != Move.No) {
                //Debug.Print("MouseMove OldX, e.x: " + SystemItems.OldLeft + " " + e.X);
                if ((Math.Abs(SystemItems.OldLeft - SystemItems.MouseX) + Math.Abs(SystemItems.OldTop - SystemItems.MouseY) > 1))
                    this.Moving = Move.Moved;
                else
                    this.Moving = Move.Init;
                if (Canvas == null) return; //  Cannot invalidate the screen.
                int bw = PhyS(borderWeight);
                System.Drawing.Rectangle oldRect = new System.Drawing.
                    Rectangle(RelativeX(left) - bw, RelativeY(top) - bw, PhyS(width) + 2 * bw, PhyS(height) + 2 * bw);
                Canvas.GuiInstance.Invalidate(oldRect);
                // Optimize by invalidating only a rectangle of the Canvas. Border may create shadows behind the drag.
                this.left += LogicalX(e.X) - SystemItems.MouseX; this.top += LogicalY(e.Y) - SystemItems.MouseY;
                System.Drawing.Rectangle newRect = new System.Drawing.Rectangle(RelativeX(left) - bw, RelativeY(top) - bw, PhyS(width) + 2 * bw, PhyS(height) + 2 * bw);
                Canvas.GuiInstance.Invalidate(newRect);
                Canvas.GuiInstance.Update(); // Paints the invalidated area. 
            }
            updateMouseVariables(e);
            this.ExecuteHandler("MouseMove", true);
        }

        [VisPropAtt(VisPropAtt.PropertyType.Eventhandler, null, Position = 12,
            LongHelp = @"The mouse pointer rests on me.
At present, uVis components such as Box and PieSlice cannot generate MouseHover reliably.",
            PropertyCategory = "EventHandler", ReadOnly = false, ShowInitially = false, ShortHelp = "The mouse pointer rests on me.")]
        public virtual void MouseHover(object sender, EventArgs e) {
            this.ExecuteHandler("MouseHover", true);
        }

        [VisPropAtt(VisPropAtt.PropertyType.Eventhandler, null, Position = 13,
            LongHelp = @"The mouse pointer moved to another component.
Before this event, the new component gets a MouseEnter.",
            PropertyCategory = "EventHandler", ReadOnly = false, ShowInitially = false,
            ShortHelp = "The mouse pointer moved to another component.")]
        public virtual void MouseLeave(object sender, EventArgs e) {
            this.ExecuteHandler("MouseLeave", true);
        }

        [VisPropAtt(VisPropAtt.PropertyType.Eventhandler, null, Position = 14,
            LongHelp = "A mouse button is pressed on me.",
            PropertyCategory = "EventHandler", ReadOnly = false,
            ShortHelp = "A mouse button is pressed on me.")]
        public virtual void MouseDown(object sender, MouseEventArgs e) {
            SystemItems.OldLeft = this.left; SystemItems.OldWidth = this.width; SystemItems.OldTop = this.top; SystemItems.OldHeight = this.height;
            this.ExecuteHandler("MouseDown", true);
            if (this.movable) Moving = Move.Init;
        }

        [VisPropAtt(VisPropAtt.PropertyType.Eventhandler, null, Position = 15,
            LongHelp = "A mouse button is released over me.",
            PropertyCategory = "EventHandler", ReadOnly = false,
            ShortHelp = "A mouse button is released over me.")]
        public virtual void MouseUp(object sender, MouseEventArgs e) {
            updateMouseVariables(e);
            if (this.Moving == Move.Moved) {
                this.ExecuteHandler("Moved", true);
            } else if (this.Moving == Move.Init) { // Moved a bit only. Reset the position.
                if (Canvas == null) return; //  Cannot invalidate the screen.
                int bw = PhyS(borderWeight);
                Canvas.GuiInstance.Invalidate(new System.Drawing.Rectangle(RelativeX(left) - bw, RelativeY(top) - bw, PhyS(width) + 2 * bw, PhyS(height) + 2 * bw));
                // Optimize by invalidating only a rectangle of the Canvas. Border may create shadows behind the drag.
                this.left = SystemItems.OldLeft; this.top = SystemItems.OldTop;
                Canvas.GuiInstance.Invalidate(new System.Drawing.Rectangle(RelativeX(left) - bw, RelativeY(top) - bw, PhyS(width) + 2 * bw, PhyS(height) + 2 * bw));
                // Optimize by invalidating only a rectangle of the Canvas.
            }
            this.Moving = Move.No;
            this.ExecuteHandler("MouseUp", true);
        }

        // GotFocus and LostFocus might be replaced by Enter and Leave, except for Forms, which uses Activate and DeActivate.
        [VisPropAtt(VisPropAtt.PropertyType.Eventhandler, null, Position = 30,
            LongHelp = "I got keyboard focus and can receive keystrokes.",
            PropertyCategory = "EventHandler", ReadOnly = false,
            ShortHelp = "I got keyboard focus and can receive keystrokes.")]
        public virtual void GotFocus(object sender, EventArgs e) {
            SetKeyFocus();
            this.ExecuteHandler("GotFocus", true);
        }

        [VisPropAtt(VisPropAtt.PropertyType.Eventhandler, null, Position = 31,
            LongHelp = "I stopped receiving keystrokes.",
            PropertyCategory = "EventHandler", ReadOnly = false,
            ShortHelp = "I stopped receiving keystrokes.")]
        public virtual void LostFocus(object sender, EventArgs e) {
            this.ExecuteHandler("LostFocus", true);
        }

        // MouseClick: Click is sufficient. MouseClick is only neededfor RichTextbox, which doesn't generate Click for right click. SL 14-12-13.
        [VisPropAtt(VisPropAtt.PropertyType.Eventhandler, null, Position = 1, PropertyCategory = "EventHandler", ReadOnly = false,
             ShowInitially = true, ShortHelp = "The user clicked me with the mouse, Enter, etc.")]
        public virtual void Click(object sender, EventArgs e) {
            if (Moving != Move.Moved) {
                //if (this is ContainerComponent) (this as ContainerComponent).keyFocus = this;
                //else if (this.canvas != null && this.GuiInstance==null) {
                //    this.canvas.keyFocus = this;  // Ensures that key events are routed to the shape component.
                //    this.canvas.GuiInstance.Focus(); // Let the ContainerComponent receive the key events.
                //}
                if (this != GetKeyFocus(this)) this.GotFocus(sender, e);
                this.ExecuteHandler("Click", true);
            }
        }

        // MouseDoubleClick: DoubleClick is sufficient. MSDN says it is on a "higher level". SL 14-12-13.
        [VisPropAtt(VisPropAtt.PropertyType.Eventhandler, null, Position = 2,
            LongHelp = "The user double-clicked me. Preceded by a Click event.",
            PropertyCategory = "EventHandler", ReadOnly = false,
            ShortHelp = "The user double-clicked me. Preceded by a Click event.")]
        public virtual void DoubleClick(object sender, EventArgs e) {
            this.ExecuteHandler("DoubleClick", true);
        }

        // Windows triggers LocationChanged and SizeChanged for programmed changes as well as user changes.
        // UVis generates Move and Resized instead.
        [VisPropAtt(VisPropAtt.PropertyType.Eventhandler, null, Position = 20,
            LongHelp = @"The user has moved me.
System.OldLeft and OldTop show the previous location.",
            PropertyCategory = "EventHandler", ReadOnly = false,
            ShortHelp = "The user has moved me.")]
        public virtual void Moved(object sender, EventArgs e) {
            //Debug.Print(this.tpl.Name + "! Moved to x, y: "  + SystemItems.MouseX + " " + SystemItems.MouseY);
            this.ExecuteHandler("Moved", true);
        }

        [VisPropAtt(VisPropAtt.PropertyType.Eventhandler, null, Position = 21,
            LongHelp = @"The user has resized me.
System.OldWidth and OldHeight show the previous size.",
            PropertyCategory = "EventHandler", ReadOnly = false,
            ShortHelp = "The user has resized me.")]
        public void Resized(object sender, EventArgs e) {
            //Debug.Print(this.tpl.Name + "! Resized to x, y: " + SystemItems.MouseX + " " + SystemItems.MouseY);
            this.ExecuteHandler("Resized", true);
        }

        // Key events occur in this sequence: KeyDown, KeyPress, KeyUp.
        [VisPropAtt(VisPropAtt.PropertyType.Eventhandler, null, Position = 34,
            LongHelp = @"The user pressed a key while I have focus.
Responds to visible keys (e.g. B and +) as well as invisible keys (e.g. Key.Enter and Key.Delete).",
            PropertyCategory = "EventHandler", ReadOnly = false,
            ShortHelp = "The user pressed a key while I have focus.")]
        public virtual void KeyDown(object sender, KeyEventArgs e) {
            if (!Enabled && (e.KeyCode == Keys.Delete || e.KeyCode == Keys.Back)) {
                e.Handled = true; SystemItems.Handled = true; return;
            }
            updateKeyVariables(e);
            this.ExecuteHandler("KeyDown", false);
        }

        [VisPropAtt(VisPropAtt.PropertyType.Eventhandler, null, Position = 35,
            LongHelp = @"The user clicked a visible character key (e.g. B or +) while I have focus.
Occurs repeatedly when the user keeps the key down.",
            PropertyCategory = "EventHandler", ReadOnly = false,
            ShortHelp = "The user clicked a visible character key (e.g. B or +) while I have focus.")]
        public virtual void KeyPress(object sender, KeyPressEventArgs e) {
            if (!Enabled && !SystemItems.CtrlPressed && !SystemItems.AltPressed) { 
                e.Handled = true; SystemItems.Handled = true; return; 
            }
            this.ExecuteHandler("KeyPress", false);
        }

        [VisPropAtt(VisPropAtt.PropertyType.Eventhandler, null, Position = 36,
            LongHelp = "The user released a key while I have focus.\nResponds to all kinds of keys.",
            PropertyCategory = "EventHandler", ReadOnly = false,
            ShortHelp = "The user released a key while I have focus.")]
        public virtual void KeyUp(object sender, KeyEventArgs e) {
            updateKeyVariables(e);
            this.ExecuteHandler("KeyUp", false);
        }

        #endregion EVENTHANDLING
    }

    //-------------------------------------------------------------------------------------------------
    public class VisComponentEventArgs : EventArgs {
        public VisComponent Component;
        public VisComponentEventArgs(VisComponent Component) {
            this.Component = Component;
        }
    }

    //-----------------------------------------------------------------------------------
    [VisClassAtt(ToolboxCategory = "Control", ShortHelp = "Windows Command Button - slow when many. Provides a short-cut key.",
        LongHelp = @"Windows Command Button - slow when many. Provides a short-cut key when there is & in the text property.
When activated, it ""clicks"" the button.", Icon = "button.bmp")]
    public class cButton : VisComponent {

        public cButton(int targetCount, DataRow row, Template tpl)
            : base(targetCount, row, tpl) {
            Button button = new System.Windows.Forms.Button();
            button.TabStop = false; // Will be set to true as a result of setting TabOrder.
            Font font = button.Font;
            button.Font = new Font(font.FontFamily, font.SizeInPoints * SystemItems.Zoom2);
            GuiInstance = button;
            GuiInstance.Tag = this;
            AttachControlEvents(button); // Attach the common control events.
        }

        public virtual string textDefault { get { return this.tpl.Name; } } // Property to allow overriding.
        public static int TextID; // Property index. Set at initialization in ReflectDLL. 
        private string text;
        [VisPropAtt(VisPropAtt.PropertyType.Property, "\"abc\"", Position = 0,
            LongHelp = @"My text. When an & is present, the following character is shown with underline and works as a shortcut key.
When activated, it ""clicks"" me.",
            ReadOnly = false, ShowInitially = true, ShortHelp = "My text. When an & is present, the following character works as a shortcut key.")]
        public string Text {
            get {
                CmpProp textProp = this.propertyValues[TextID];
                if (textProp.state >= State.Updated) return text;
                Text = toString(EvalProp(TextID, textDefault), textDefault);
                return text;
            }
            set {
                text = value; SetUpdated(ref propertyValues[TextID]);
                GuiInstance.Text = value;
                Font font = GuiInstance.Font;
                GuiInstance.Font = new Font(font.FontFamily, FontSize * SystemItems.Zoom2);
            }
        }

        public static int FontSizeID; // Property index. Set at initialization in ReflectDLL. 
        private float fontSize = 8.50F; // Same as MS default for TextBox and Label.
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, Position = 21, ShowInitially = false,
            ShortHelp = "The size of my text in points. Default: 8.50")]
        public float FontSize {
            get {
                CmpProp fontSizeProp = this.propertyValues[FontSizeID];
                if (fontSizeProp.state >= State.Updated) return fontSize;
                FontSize = (float)toDouble(EvalProp(FontSizeID, 8.50), 8.50);
                return fontSize;
            }
            set {
                fontSize = value; // GuiInstance physical font size will be set in Text, since there may not be a FontSize attribute. 
                SetUpdated(ref propertyValues[FontSizeID]);
            }
        }
    }


    }

    //-----------------------------------------------------------------------------------
[VisClassAtt(ToolboxCategory = "Control", ShortHelp = "Windows CheckBox - slow when many.", Icon = "checkbox.bmp")]
public class Checkbox : VisComponent {
    CheckBox checkBox;

    public Checkbox(int targetCount, DataRow row, Template tpl)
        : base(targetCount, row, tpl) {
        checkBox = new System.Windows.Forms.CheckBox();
        checkBox.TabStop = false; // Will be set to true as a result of setting TabOrder.
        checkBox.Width = 30;
        checkBox.BackColor = Color.Yellow;
        checkBox.ForeColor = Color.Red;
        GuiInstance = checkBox;
        GuiInstance.Tag = this;
        AttachControlEvents(checkBox); // Attach the common control events.
    }

    public static int CheckID; // Property index. Set at initialization in ReflectDLL. 
    private bool check;
    [VisPropAtt(VisPropAtt.PropertyType.Property, null, Position = 0, ReadOnly = false, ShowInitially = true,
        ShortHelp = @"True when the check mark is shown.")]
    public bool Check {
        get {
            CmpProp checkProp = this.propertyValues[CheckID];
            if (checkProp.state >= State.Updated) return checkBox.Checked;
            Check = toBool(EvalProp(CheckID, false), false);
            return check;
        }
        set {
            check = value; SetUpdated(ref propertyValues[CheckID]);
            checkBox.Checked = value;
        }
    }

    [VisPropAtt(VisPropAtt.PropertyType.Property, " ' Always 14", PropertyCategory = "Common", ReadOnly = true, ShowInitially = true,
    ShortHelp = "Pixels from my top to my bottom inside the border. Always 14.", Position = 7)]
    public override int Height {
        get { base.Height = 14; return 14; }
    }

    [VisPropAtt(VisPropAtt.PropertyType.Property, " ' Always 14", PropertyCategory = "Common", ReadOnly = true, ShowInitially = true,
        ShortHelp = "Pixels from my left to my right inside the border. Always 14.", Position = 10)]
    public override int Width {
        get { base.Width = 14; return 14; }
    }
}

    //------------------------------------------------------------------------------------------
    [VisClassAtt(ToolboxCategory = "Control", Icon = "clabel.png", ShortHelp = "Windows Label - slow when many. Provides a short-cut key.",
        LongHelp = @"Windows Label - slow when many. Provides a short-cut key when there is & in the text property.
When activated, it passes focus on to the next component in the tab sequence.")]
    public class cLabel : VisComponent {

        public cLabel(int targetCount, DataRow row, Template tpl)
            : base(targetCount, row, tpl) {
            System.Windows.Forms.Label label = new System.Windows.Forms.Label();
            label.TabStop = false; // Will be set to true as a result of setting TabOrder.
            Font font = label.Font;
            label.Font = new Font(font.FontFamily, font.SizeInPoints * SystemItems.Zoom2);
            GuiInstance = label;
            GuiInstance.Tag = this;
            AttachControlEvents(GuiInstance); // Attach the common control events.
        }

        public override Color backColorDefault { get { return Color.Empty; } }
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Common", ReadOnly = false, ShowInitially = false,
            ShortHelp = "The color of my inner area. Default: No color.", Position = 11)]
        public override Color BackColor { get { return base.BackColor; } }

        public virtual string textDefault { get { return this.tpl.Name; } } // Property to allow overriding.
        public static int TextID; // Property index. Set at initialization in ReflectDLL. 
        private string text;
        [VisPropAtt(VisPropAtt.PropertyType.Property, "\"abc\"", ReadOnly = false, ShowInitially = true, ShortHelp = "My text", Position = 0,
            LongHelp = @"My text. When an & is present, the following character is shown with underline and works as a shortcut key.
When activated, it passes focus on to the next component in the tab sequence.")]
        public string Text {
            get {
                CmpProp textProp = this.propertyValues[TextID];
                if (textProp.state >= State.Updated) return text;
                Text = toString(EvalProp(TextID, textDefault), textDefault);
                return text; 
            }
            set { 
                text = value;
                Font font = GuiInstance.Font;
                GuiInstance.Font = new Font(font.FontFamily, FontSize * SystemItems.Zoom2);
                SetUpdated(ref propertyValues[TextID]); GuiInstance.Text = value;
            } 
        }

        public static int FontSizeID; // Property index. Set at initialization in ReflectDLL. 
        private float fontSize = 8.50F; // Same as MS default for TextBox and Label.
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, Position = 21, ShowInitially = false,
            ShortHelp = "The size of my text in points. Default: 8.50")]
        public float FontSize {
            get {
                CmpProp fontSizeProp = this.propertyValues[FontSizeID];
                if (fontSizeProp.state >= State.Updated) return fontSize;
                FontSize = (float)toDouble(EvalProp(FontSizeID, 8.50), 8.50);
                return fontSize;
            }
            set {
                fontSize = value; // GuiInstance physical font size will be set in Text, since there may not be a FontSize attribute
                SetUpdated(ref propertyValues[FontSizeID]);
            }
        }

    }

    //------------------------------------------------------------------------------------------
    [VisClassAtt(ToolboxCategory = "Control", ShortHelp = "Simple Box - fast", LongHelp = "Simple Box - fast", Icon = "box.png")]
    public class Box : VisComponent {

        public Box(int targetCount, DataRow row, Template tpl)
            : base(targetCount, row, tpl) {
        }

        public override void Paint(PaintEventArgs e) {
            e.Graphics.SmoothingMode = SmoothingMode.HighSpeed;
            if (Visible) {
                if (BackColor != Color.Empty && BackColor != Color.Transparent)
                    e.Graphics.FillRectangle(new SolidBrush(BackColor), RelativeX(left), RelativeY(top), PhyS(width), PhyS(height));
                int bw = PhyS(borderWeight);
                if (bw > 0) {
                    int offset = (bw + 1) / 2; // Make sure the border surrounds the box and that a zero-size box is shown when it has a border..
                    e.Graphics.DrawRectangle(new Pen(borderColor, bw), RelativeX(left) - offset, RelativeY(top) - offset, PhyS(width) + bw, PhyS(height) + bw);
                }
                base.Paint(e);
            }
        }
    }

    //-----------------------------------------------------------------------------------------
    [VisClassAtt(ToolboxCategory = "Control", ShortHelp = "Small or large picture from a file.", Icon = "picturebox.bmp")]
    public class Icon : VisComponent {

        public Icon(int targetCount, DataRow row, Template tpl)
            : base(targetCount, row, tpl) {
        }

        Image image = null; // The image to show.
        public static int FileNameID; // Property index. Set at initialization in ReflectDLL. 
        private string fileName;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Special", ShowInitially = true,
            ShortHelp = "File path to the picture. Relative to the application folder. Examples: recent.bmp, photos\\recent.gif", Position = 1)]
        public string FileName {
            get {
                CmpProp fileNameProp = this.propertyValues[FileNameID];
                if (fileNameProp.state >= State.Updated) return fileName;
                FileName = toString(EvalProp(FileNameID, null), null);
                try {
                    if (fileName == null) {
                        Directory.SetCurrentDirectory(VisEngine.ProgramDirectory);
                        image = Image.FromFile("Icons\\picturebox.bmp");
                        fileName = VisEngine.ProgramDirectory + "\\Icons\\picturebox.bmp ' (default picture)";
                    } else {
                        Directory.SetCurrentDirectory(VisEngine.ApplicationDirectory);
                        image = Image.FromFile(fileName);
                    }
                }
                catch (Exception e) {
                    image = null; // Paint will show nothing.
                    ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp, e.Message, null, this.Index());
                }
                // Start internal calculation. Since Filename is updated, we don't need a lock.
                if (tpl.propWidth.expr == null) { // Width has no expression.
                    if (tpl.propHeight.expr != null) { // Height has.
                        base.Width = image.Width * Height / image.Height;
                    } else { // None of them has a formula.
                        base.Height = image.Height; base.Width = image.Width;
                    }
                } else if (tpl.propHeight.expr == null) { // Height has no expression. Width has one.
                    base.Height = image.Height * Width / image.Width;
                } // Else: Use formula for Height as well as Width.

                return fileName;
            }
            set {
                fileName = value; SetUpdated(ref propertyValues[FileNameID]);
                } 
            }

        [VisPropAtt(VisPropAtt.PropertyType.Property, null,  PropertyCategory = "Common", ReadOnly = false, ShowInitially = true,
            ShortHelp = "Pixels from my top to my bottom inside the border. If no formula, preserve ratio.", Position = 7)]
        public override int Height {
            get {
                CmpProp heightProp = this.propertyValues[HeightID];
                if (heightProp.state >= State.Updated) return height;
                heightProp.state = State.Visited; // Start internal calculation. Catch cyclical references.
                string dummy = FileName; // Evaluates Height if it has no formula.
                if (heightProp.state == State.Visited)
                    heightProp.state = State.Old;
                return base.Height; // Evaluates Height if not set by UpdateInternal. 
            }
            set { base.Height = value; }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Common", ReadOnly = false, ShowInitially = true,
            ShortHelp = "Pixels from my left to my right inside the border. If no formula, preserve ratio.", Position = 10)]
        public override int Width {
            get {
                CmpProp widthProp = this.propertyValues[WidthID];
                if (widthProp.state >= State.Updated) return width;
                widthProp.state = State.Visited; // Start internal calculation. Catch cyclical references.
                string dummy = FileName; // Evaluates width if it has no formula.
                if (widthProp.state == State.Visited)
                    widthProp.state = State.Old;
                return base.Width; // Evaluates width if not set by UpdateInternal. 
            }
            set { base.Width = value; }
        }

        public override void Paint(PaintEventArgs e) {
            if (Visible) {
                base.Paint(e);
                if (image != null)
                    e.Graphics.DrawImage(image, new Rectangle(RelativeX(left), RelativeY(top), PhyS(width), PhyS(height)));
            }
        }
    }

    //----------------------------------------------------------------------------------------
    //[VisClassAtt(ToolboxCategory = "Control", ShortHelp = "Slice or ring of a slice - fast", LongHelp = "Slice or ring of a slice - fast", Icon = "pieSlice.PNG")]
    //public class PieSlice : VisComponent {

    //    public PieSlice(int numOfFormulas, int numOfChildBundle, DataRow row, Bundle bundle)
    //        : base(numOfFormulas, numOfChildBundle, row, bundle) {
    //        backColor = Color.White;
    //        SweepAngle = 50;
    //        OuterRadius = 10;
    //    }
    //    GraphicsPath Path;
    //    Rectangle Region;
    //    private int innerRadius;

    //    [VisPropAtt(VisPropAtt.PropertyType.Property, null, LongHelp = "The radius of the inner pie slice. Default 0",
    //       PropertyCategory = "Special", ReadOnly = true, ShowInitially = true,
    //       ShortHelp = "The radius of the inner pie slice", Position = 3)]
    //    public int InnerRadius { set { innerRadius = value; Created = false; } get { return innerRadius; } }
    //    private GraphicsPath ArcPath = new GraphicsPath();

    //    private int outerRadius;
    //    [VisPropAtt(VisPropAtt.PropertyType.Property, "30", LongHelp = "The the radius of the outer pie slice",
    //       PropertyCategory = "Special", ReadOnly = true, ShowInitially = true,
    //       ShortHelp = "The radius of the outer pie slice", Position = 4)]
    //    public int OuterRadius { set { outerRadius = value; Created = false; } get { return outerRadius; } }

    //    private int centerX;
    //    [VisPropAtt(VisPropAtt.PropertyType.Property, "50", LongHelp = "The horizontal position of the pie slice circle center",
    //       PropertyCategory = "Special", ReadOnly = true, ShowInitially = true,
    //       ShortHelp = "The horizontal position of the pie slice circle center", Position = 5)]
    //    public int CenterX { set { centerX = value; Created = false; } get { return centerY; } }

    //    private int centerY;
    //    [VisPropAtt(VisPropAtt.PropertyType.Property, "50", LongHelp = "The vertical position of the pie slice circle center",
    //       PropertyCategory = "Special", ReadOnly = true, ShowInitially = true,
    //       ShortHelp = "The vertical position of the pie slice circle center", Position = 6)]
    //    public int CenterY { set { centerY = value; Created = false; } get { return centerY; } }

    //    public override int Top {
    //        get {
    //            if (!Created)
    //                createPie();
    //            if (Region != null)
    //                return Region.Top;
    //            return 0;
    //        }
    //    }

    //    public override int Left {
    //        get {
    //            if (!Created)
    //                createPie();
    //            if (Region != null)
    //                return Region.Left;
    //            return 0;
    //        }
    //    }

    //    public int ArcLeft {
    //        get {
    //            if (ArcPath.PathPoints.Length > 0) {
    //                if (ArcPath.PathPoints.Length == 1)
    //                    return (int)ArcPath.PathPoints[0].X;
    //                else
    //                    return (int)(ArcPath.PathPoints[1].X + ArcPath.PathPoints[2].X) / 2;
    //            } else
    //                return 0;
    //        }
    //    }

    //    public int ArcTop {
    //        get {
    //            if (ArcPath.PathPoints.Length > 0) {
    //                if (ArcPath.PathPoints.Length == 1)
    //                    return (int)ArcPath.PathPoints[0].Y;
    //                else
    //                    return (int)(ArcPath.PathPoints[1].Y + ArcPath.PathPoints[2].Y) / 2;
    //            } else
    //                return 0;
    //        }
    //    }

    //    private double startAngle;
    //    [VisPropAtt(VisPropAtt.PropertyType.Property, "30", LongHelp = "The start angle of the pie slice in degrees",
    //       PropertyCategory = "Special", ReadOnly = true, ShowInitially = true,
    //       ShortHelp = "The start angle of the pie slice in degrees", Position = 6)]
    //    public double StartAngle {
    //        set {
    //            startAngle = value;
    //            Created = false;
    //        }

    //        get {
    //            return startAngle;
    //        }
    //    }

    //    private double sweepAngle;
    //    [VisPropAtt(VisPropAtt.PropertyType.Property, "30", LongHelp = "The angle subtended by the pie slice in degrees",
    //       PropertyCategory = "Special", ReadOnly = true, ShowInitially = true,
    //       ShortHelp = "The angle subtended by the pie slice in degrees", Position = 7)]
    //    public double SweepAngle {
    //        set {
    //            sweepAngle = value;
    //            Created = false;
    //        }
    //        get {
    //            return sweepAngle;
    //        }
    //    }

    //    private void createPie() {
    //        Path = new GraphicsPath();
    //        Rectangle outerRec = new Rectangle(CenterX - OuterRadius, CenterY - OuterRadius, OuterRadius * 2, OuterRadius * 2);
    //        Rectangle innerRec = new Rectangle(CenterX - innerRadius, CenterY - innerRadius, InnerRadius * 2, InnerRadius * 2);
    //        ArcPath = new GraphicsPath();
    //        float fStartAngle = (float)(startAngle - sweepAngle);
    //        ArcPath.AddArc(outerRec, (float)fStartAngle, (float)sweepAngle);
    //        if (OuterRadius > 0 && innerRadius <= 0)
    //            Path.AddPie(outerRec, (float)fStartAngle, (float)sweepAngle);
    //        //Path.AddString("ok", new FontFamily("Tahoma"), 1, 10, new Point(20, 20), StringFormat.GenericDefault);
    //        if (InnerRadius > 0) {
    //            Path = new GraphicsPath();
    //            Path.AddArc(outerRec, (float)fStartAngle, (float)sweepAngle);
    //            Path.Reverse();
    //            Path.AddArc(innerRec, (float)fStartAngle, (float)sweepAngle);
    //            Path.CloseAllFigures();
    //        }

    //        //define a bounding box
    //        float PositiveStartAngle = fStartAngle;
    //        while (PositiveStartAngle < 0)
    //            PositiveStartAngle += 360;
    //        if (PositiveStartAngle % 360 >= 0 && PositiveStartAngle % 360 < 90) { //starts in first square of a circle
    //            if (sweepAngle >= 0) {
    //                if (sweepAngle + PositiveStartAngle > 0 && sweepAngle + PositiveStartAngle <= 90) //exactly first square
    //                    Region = new Rectangle(CenterX, CenterY, OuterRadius, OuterRadius);
    //                else if (sweepAngle + PositiveStartAngle > 90 && sweepAngle + PositiveStartAngle <= 180) //spanned across 1st+2nd square
    //                    Region = new Rectangle(CenterX - OuterRadius, CenterY, OuterRadius * 2, OuterRadius);
    //                else
    //                    Region = new Rectangle(CenterX - OuterRadius, CenterY - OuterRadius, OuterRadius * 2, OuterRadius * 2);
    //            } else { //negative sweep angle
    //                if (sweepAngle + PositiveStartAngle == 0)
    //                    Region = new Rectangle(CenterX, CenterY, OuterRadius, OuterRadius); //exactly first square
    //                else if (sweepAngle + PositiveStartAngle >= -90 && sweepAngle + PositiveStartAngle < 0)
    //                    Region = new Rectangle(CenterX, CenterY - OuterRadius, OuterRadius, OuterRadius * 2); //spanned across 1st+4th squares
    //                else
    //                    Region = new Rectangle(CenterX - OuterRadius, CenterY - OuterRadius, OuterRadius * 2, OuterRadius * 2);
    //            }
    //        } else if (PositiveStartAngle % 360 >= 90 && PositiveStartAngle % 360 < 179) { //starts in second square of a circle
    //            if (sweepAngle >= 0) {
    //                if (sweepAngle + PositiveStartAngle > 90 && sweepAngle + PositiveStartAngle <= 180) //exactly second square
    //                    Region = new Rectangle(CenterX - OuterRadius, CenterY, OuterRadius, OuterRadius);
    //                else if (sweepAngle + PositiveStartAngle > 180 && sweepAngle + PositiveStartAngle <= 269) //spanned across 2nd+3rd square
    //                    Region = new Rectangle(CenterX - OuterRadius, CenterY - OuterRadius, OuterRadius, OuterRadius * 2);
    //                else
    //                    Region = new Rectangle(CenterX - OuterRadius, CenterY - OuterRadius, OuterRadius * 2, OuterRadius * 2);
    //            } else {
    //                if (sweepAngle + PositiveStartAngle == 90)
    //                    Region = new Rectangle(CenterX - OuterRadius, CenterY, OuterRadius, OuterRadius); //exactly second square
    //                else if (sweepAngle + PositiveStartAngle > 0 && sweepAngle + PositiveStartAngle < 90)
    //                    Region = new Rectangle(CenterX - OuterRadius, CenterY, OuterRadius * 2, OuterRadius); //spanned across 1st+2nd squares
    //                else
    //                    Region = new Rectangle(CenterX - OuterRadius, CenterY - OuterRadius, OuterRadius * 2, OuterRadius * 2);
    //            }
    //        } else if (PositiveStartAngle % 360 >= 180 && PositiveStartAngle % 360 < 269) //starts in third square of a circle
    //        {
    //            if (sweepAngle >= 0) {
    //                if (sweepAngle + PositiveStartAngle > 180 && sweepAngle + PositiveStartAngle <= 270) //exactly third square
    //                    Region = new Rectangle(CenterX - OuterRadius, CenterY - OuterRadius, OuterRadius, OuterRadius);
    //                else if (sweepAngle + PositiveStartAngle > 180 && sweepAngle + PositiveStartAngle <= 269) //spanned across 3rd+4th square
    //                    Region = new Rectangle(CenterX - OuterRadius, CenterY - OuterRadius, OuterRadius * 2, OuterRadius);
    //                else
    //                    Region = new Rectangle(CenterX - OuterRadius, CenterY - OuterRadius, OuterRadius * 2, OuterRadius * 2);
    //            } else {
    //                if (sweepAngle + PositiveStartAngle == 180)
    //                    Region = new Rectangle(CenterX - OuterRadius, CenterY - OuterRadius, OuterRadius, OuterRadius); //exactly third square
    //                else if (sweepAngle + PositiveStartAngle >= 90 && sweepAngle + PositiveStartAngle < 180)
    //                    Region = new Rectangle(CenterX - OuterRadius, CenterY - OuterRadius, OuterRadius, OuterRadius * 2); //spanned across 2nd+3rd squares
    //                else
    //                    Region = new Rectangle(CenterX - OuterRadius, CenterY - OuterRadius, OuterRadius * 2, OuterRadius * 2);
    //            }
    //        } else if (PositiveStartAngle % 360 >= 270 && PositiveStartAngle % 360 < 359) //starts in fourth square of a circle
    //        {
    //            if (sweepAngle >= 0) {
    //                if (sweepAngle + PositiveStartAngle > 270 && sweepAngle + PositiveStartAngle <= 360) //exactly fourth square
    //                    Region = new Rectangle(CenterX, CenterY - OuterRadius, OuterRadius, OuterRadius);
    //                else if (sweepAngle + PositiveStartAngle > 180 && sweepAngle + PositiveStartAngle <= 269) //spanned across 4th, 1st squares
    //                    Region = new Rectangle(CenterX, CenterY - OuterRadius, OuterRadius, OuterRadius * 2);
    //                else
    //                    Region = new Rectangle(CenterX - OuterRadius, CenterY - OuterRadius, OuterRadius * 2, OuterRadius * 2);
    //            } else {
    //                if (sweepAngle + PositiveStartAngle == 270)
    //                    Region = new Rectangle(CenterX, CenterY - OuterRadius, OuterRadius, OuterRadius); //exactly fourth square
    //                else if (sweepAngle + PositiveStartAngle >= 180 && sweepAngle + PositiveStartAngle < 270)
    //                    Region = new Rectangle(CenterX - OuterRadius, CenterY - OuterRadius, OuterRadius * 2, OuterRadius); //spanned across 4th,3rd squares
    //                else
    //                    Region = new Rectangle(CenterX - OuterRadius, CenterY - OuterRadius, OuterRadius * 2, OuterRadius * 2);
    //            }
    //        } else {
    //            Region = new Rectangle(CenterX - OuterRadius, CenterY - OuterRadius, OuterRadius * 2, OuterRadius * 2);
    //        }
    //    }

    //    public override void Paint(PaintEventArgs e) {
    //        if (Visible) {
    //            //if (!Created) {
    //            //    createPie();
    //            //    Created = true;
    //            //}
    //            if (BorderColor != Color.Empty && BorderColor != Color.Transparent)
    //                e.Graphics.DrawPath(new Pen(BorderColor), Path);
    //            if (BackColor != Color.Empty && BackColor != Color.Transparent)
    //                e.Graphics.FillPath(new SolidBrush(BackColor), Path);
    //        }
    //    }
    //}

//----------------------------------------------------------------------------------------
    [VisClassAtt(ToolboxCategory = "Control", ShortHelp = "Label with a text - fast", Icon = "label.png")]
    public class Label : VisComponent {

        public Label(int targetCount, DataRow row, Template tpl)
            : base(targetCount, row, tpl) {
        }

        public override void ResetUpdated() { // Called by kernel at refresh to set internal dependent properties to not-updated.
            sizeUpdated = false;
            base.ResetUpdated(); // Reset Height, etc.
        }

        public bool sizeUpdated = false;
        public void UpdateSize(int sourceID) { // Updates properties used to compute text size, e.g. Autosize and Bold.
            // Then computes and sets text size. If autoSize, sets Width or Height. Makes the source property Visited.
            // When setting text size, etc. makes the property Updated. Otherwise returns with source property Visited. 
            if (sizeUpdated) return;
            CmpProp sourceProp = this.propertyValues[sourceID];
            if (sourceProp.state >= State.Updated) return; // Init formulas are Initialized and don't need updates.
            sourceProp.state = State.Visited; // Start internal calculation. Catch cyclical references.

            FontStyle s = FontStyle.Regular;
            if (Bold) s = FontStyle.Bold;
            if (Italic) s = s | FontStyle.Italic;
            if (Underline) s = s | FontStyle.Underline;
            string fontName = FontName; // Sets also fontFamily.
            try {
                font = new Font(fontFamily, FontSize * SystemItems.Zoom2, s);
            }
            catch (Exception e) { // Some fonts don't support some styles, e.g. Aharoni. 
                ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp, e.Message, VisEngine.curExprToken, this.Index());
                font = DefaultFont;
            }

            SizeF size; // Compute text size. If autoSize, set Height and if not Wordwrap, also Width.
            Graphics dummy = Graphics.FromImage(new Bitmap(1, 1));
            StringFormat format = new StringFormat();
            if (!WordWrap) format.FormatFlags = StringFormatFlags.NoWrap;

            string test = Text; if (string.IsNullOrEmpty(text)) test = "A"; // Avoid that empty labels have zero size.
            if (wordWrap) { // WordWrap calculated above.
                if (sourceID == WidthID) sourceProp.state = State.Old; // Allow base.Width to work.
                size = dummy.MeasureString(test, font, PhyS(base.Width - PaddingLeft - PaddingRight), format);
                TextHeight = size.Height / SystemItems.Zoom2;
                TextWidth = size.Width / SystemItems.Zoom2;
                if (AutoSize) {
                    base.Height = (int)(textHeight + 1.0) + PaddingTop + PaddingBottom; // (int) truncates.
                }
            } else { // No wrap.
                size = dummy.MeasureString(test, font, 32000, format); // Any Width
                textHeight = size.Height / SystemItems.Zoom2; 
                textWidth = size.Width / SystemItems.Zoom2; this.propertyValues[TextWidthID].state = State.Updated;
                if (AutoSize) {
                    base.Width = (int)(textWidth + 1.0) + PaddingLeft + PaddingRight; // (int) truncates.
                    base.Height = (int)(textHeight + 1.0) + PaddingTop + PaddingBottom;
                }
            }
            this.propertyValues[TextHeightID].state = State.Updated;
            this.propertyValues[TextWidthID].state = State.Updated;

            if (sourceProp.state == State.Visited) // If it hasn't been computed, reset it to Old.
                sourceProp.state = State.Old;
            sizeUpdated = true;
        }

        public override int Top {
            get { UpdateSize(TopID); return base.Top; } // Evaluates Top if not set by UpdateSize. 
            set { base.Top = value; }
        }

        public override int Height {
            get { UpdateSize(HeightID); return base.Height; } // Evaluates Height if not set by UpdateSize. 
            set { base.Height = value; }
        }

        public override int Bottom {
            get { UpdateSize(BottomID); return base.Bottom; } // Evaluates Bottom if not set by UpdateSize. 
            set { base.Bottom = value; }
        }

        public override int Left {
            get { UpdateSize(LeftID); return base.Left; }
            set { base.Left = value; }
        }

        public override int Width {
            get { UpdateSize(WidthID); return base.Width; }
            set { base.Width = value; }
        }

        public override int Right {
            get { UpdateSize(RightID); return base.Right; }
            set { base.Right = value; }
        }

        public override int borderWeightDefault { get { return 0; } } 
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Common", ReadOnly = false, ShowInitially = false,
    ShortHelp = "Pixels across my border. Default: 0 (no border).", Position = 13)]
        public override int BorderWeight { get { return base.BorderWeight; } }

        public override Color backColorDefault { get { return Color.Empty; } } 
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Common", ReadOnly = false, ShowInitially = false,
            ShortHelp = "The color of my inner area. Default: No color.", Position = 11)]
        public override Color BackColor { get { return base.BackColor; } }

        public static int TextHeightID; // Property index. Set at initialization in ReflectDLL. 
        public float textHeight;
        [VisPropAtt(VisPropAtt.PropertyType.Property, "' Read only", ReadOnly = true, ShowInitially = false, Position = 0,
            ShortHelp = "The height of the text in pixels.")]
        public float TextHeight {
            get { UpdateSize(TextHeightID); return textHeight; }
            set { textHeight = value; SetUpdated(ref propertyValues[TextHeightID]); }
        }

        public static int TextWidthID; // Property index. Set at initialization in ReflectDLL. 
        public float textWidth;
        [VisPropAtt(VisPropAtt.PropertyType.Property, "' Read only", ReadOnly = true, ShowInitially = false, Position = 1,
            ShortHelp = "The width of the text in pixels.")]
        public float TextWidth {
            get { UpdateSize(TextWidthID); return textWidth; }
            set { textWidth = value; SetUpdated(ref propertyValues[TextWidthID]); }
        }

        public static int TextID; // Property index. Set at initialization in ReflectDLL. 
        public string text;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ReadOnly = false, ShowInitially = true, Position = 2,
            ShortHelp = "My text. May contain several lines, e.g. computed with T1 + System.NewLine() + T2.")]
        public string Text {
            get {
                CmpProp textProp = this.propertyValues[TextID];
                if (textProp.state >= State.Updated) return text;
                string textDefault = this.tpl.Name;
                Text = toString(EvalProp(TextID, textDefault), textDefault);
                return text;
            }
            set { text = value; SetUpdated(ref propertyValues[TextID]); }
        }

        public static int AlignID; // Property index. Set at initialization in ReflectDLL. 
        private int align;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Special", ReadOnly = false,
            ShortHelp = "Horizontal alignment of my text: 0 (Left), 1 (Center), 2 (Right). Default: 0 (Left)", Position = 6)]
        public int Align {
            get {
                CmpProp alignProp = this.propertyValues[AlignID];
                if (alignProp.state >= State.Updated) return align;
                Align = toPosInt32(EvalProp(AlignID, 0), 0);
                return align;
            }
            set { align = value; SetUpdated(ref propertyValues[AlignID]); }
        }

        public static int TextColorID; // Property index. Set at initialization in ReflectDLL. 
        public Color textColor;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ReadOnly = false, ShowInitially = false, Position = 5, 
            ShortHelp = "The color of my text. Default: black.")]
        public Color TextColor {
            get {
                CmpProp textColorProp = this.propertyValues[TextColorID];
                if (textColorProp.state >= State.Updated) return textColor;
                TextColor = toColor(EvalProp(TextColorID, Color.Black), Color.Black);
                return textColor;
            }
            set { textColor = value; SetUpdated(ref propertyValues[TextColorID]); }
        }

        public static int AutoSizeID; // Property index. Set at initialization in ReflectDLL. 
        public bool autoSize; // Similar Textbox.
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ShowInitially = true, Position = 7,
            LongHelp = "When true, my Height is determined by the text and if not Wordwrap, also my Width. Default: false.\n Formulas in Width and Height are ignored",
            ShortHelp = "When true, my Height is determined by the text and if not Wordwrap, also my Width. Default: false.")]
        public bool AutoSize { 
            get {
                CmpProp autoSizeProp = this.propertyValues[AutoSizeID];
                if (autoSizeProp.state >= State.Updated) return autoSize;
                AutoSize = toBool(EvalProp(AutoSizeID, false), false);
                return autoSize; 
            }
            set { autoSize = value; SetUpdated(ref propertyValues[AutoSizeID]); }
        }

        public static int WordWrapID; // Property index. Set at initialization in ReflectDLL. 
        private bool wordWrap; // Similar to Textbox.
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ShowInitially = true, Position = 8,
            ShortHelp = "When true, long lines are wrapped to the next line. Default: false.")]
        public bool WordWrap { 
            get {
                CmpProp wordWrapProp = this.propertyValues[WordWrapID];
                if (wordWrapProp.state >= State.Updated) return wordWrap;
                WordWrap = toBool(EvalProp(WordWrapID, false), false);
                return wordWrap; 
            }
            set { wordWrap = value; SetUpdated(ref propertyValues[WordWrapID]); }
        }

        public static int RotationAngleID; // Property index. Set at initialization in ReflectDLL. 
        private double rotationAngle = 0;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ShowInitially = false, Position = 9,
            ShortHelp = "Rotate the text this angle around the top-left corner. Default: 0")]
        public double RotationAngle { 
            get {
                CmpProp rotationAngleProp = this.propertyValues[RotationAngleID];
                if (rotationAngleProp.state >= State.Updated) return rotationAngle;
                RotationAngle = toDouble(EvalProp(RotationAngleID, 0.0), 0.0);
                return rotationAngle; 
            }
            set { rotationAngle = value; SetUpdated(ref propertyValues[RotationAngleID]); }  
        }

        public static int PaddingTopID; // Property index. Set at initialization in ReflectDLL. 
        private int paddingTop = 2;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ShowInitially = false, Position = 10, 
            ShortHelp = "Pixels between my top border and the top of my text. Default: 2.")]
        public int PaddingTop {
            get {
                CmpProp paddingTopProp = this.propertyValues[PaddingTopID];
                if (paddingTopProp.state >= State.Updated) return paddingTop;
                PaddingTop = toPosInt32(EvalProp(PaddingTopID, 2), 2);
                return paddingTop;
            }
            set { paddingTop = value; SetUpdated(ref propertyValues[PaddingTopID]); }
        }

        public static int PaddingLeftID; // Property index. Set at initialization in ReflectDLL. 
        private int paddingLeft = 2;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ShowInitially = false, Position = 11,
            ShortHelp = "Pixels between my left border and the text. Default: 2.")]
        public int PaddingLeft {
            get {
                CmpProp paddingLeftProp = this.propertyValues[PaddingLeftID];
                if (paddingLeftProp.state >= State.Updated) return paddingLeft;
                PaddingLeft = toPosInt32(EvalProp(PaddingLeftID, 2), 2);
                return paddingLeft;
            }
            set { paddingLeft = value; SetUpdated(ref propertyValues[PaddingLeftID]); }
        }

        public static int PaddingBottomID; // Property index. Set at initialization in ReflectDLL. 
        private int paddingBottom = 2;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ShowInitially = false, Position = 12,
            ShortHelp = "Pixels between my bottom border and the text. Default: 2.")]
        public int PaddingBottom { 
            get {
                CmpProp paddingBottomProp = this.propertyValues[PaddingBottomID];
                if (paddingBottomProp.state >= State.Updated) return paddingBottom;
                PaddingBottom = toPosInt32(EvalProp(PaddingBottomID, 2), 2);
                return paddingBottom;
            }
            set { paddingBottom = value; SetUpdated(ref propertyValues[PaddingBottomID]); }
}

        public static int PaddingRightID; // Property index. Set at initialization in ReflectDLL. 
        private int paddingRight = 2;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ShowInitially = false, Position = 13,
            ShortHelp = "Pixels between my right border and the text. Default: 2.")]
        public int PaddingRight { 
            get {
                CmpProp paddingRightProp = this.propertyValues[PaddingRightID];
                if (paddingRightProp.state >= State.Updated) return paddingRight;
                PaddingRight = toPosInt32(EvalProp(PaddingRightID, 2), 2);
                return paddingRight;
            }
            set { paddingRight = value; SetUpdated(ref propertyValues[PaddingRightID]); }
}
        public static Font DefaultFont = new System.Drawing.Font("Tahoma", 8.50F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
        private FontFamily fontFamily = DefaultFont.FontFamily;
        private static string fontNameDefault = "Tahoma";
        Font font; 

        public static int FontNameID; // Property index. Set at initialization in ReflectDLL. 
        private string fontName;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, Position = 20,
            LongHelp = "The font for my text, e.g. Font.Arial or \"Times New Roman\". Default: \"Tahoma\"\nWhen the font name is more than one word, you have to enclose it in quotes.",
            ShowInitially = false, ShortHelp = "The font for my text, e.g. Font.Arial or \"Times New Roman\". Defualt: \"Tahoma\"")]
        public string FontName {
            get {
                CmpProp fontNameProp = this.propertyValues[FontNameID];
                if (fontNameProp.state >= State.Updated) return fontName;
                FontName = toString(EvalProp(FontNameID, fontNameDefault), fontNameDefault);
                try {
                    fontFamily = new FontFamily(fontName);
                }
                catch (System.ArgumentException) {
                    ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp, "Font " + fontName + " doesn't exist", VisEngine.curExprToken, this.Index());
                }
                return fontName;
            }
            set { fontName = value; SetUpdated(ref propertyValues[FontNameID]); }
        }

        public static int FontSizeID; // Property index. Set at initialization in ReflectDLL. 
        private float fontSize = 8.50F; // Same as MS default for TextBox and Label.
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, Position = 21, ShowInitially = false, 
            ShortHelp = "The size of the text in points. Default: 8.50")]
        public float FontSize {
            get {
                CmpProp fontSizeProp = this.propertyValues[FontSizeID];
                if (fontSizeProp.state >= State.Updated) return fontSize;
                FontSize = (float)toDouble(EvalProp(FontSizeID, 8.50), 8.50);
                return fontSize;
            }
            set { fontSize = value; SetUpdated(ref propertyValues[FontSizeID]); }
        }

        public static int BoldID; // Property index. Set at initialization in ReflectDLL. 
        private bool bold = false;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, Position = 22, ShowInitially = false, 
            ShortHelp = "When true, the text is bold. Default: false.")]
        public bool Bold {
            get {
                CmpProp boldProp = this.propertyValues[BoldID];
                if (boldProp.state >= State.Updated) return bold;
                Bold = toBool(EvalProp(BoldID, false), false);
                return bold;
            }
            set { bold = value; SetUpdated(ref propertyValues[BoldID]); }
        }

        public static int ItalicID; // Property index. Set at initialization in ReflectDLL. 
        private bool italic = false;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, Position = 23, ShowInitially = false, 
            ShortHelp = "When true, the text is Italic. Default: false.")]
        public bool Italic {
            get {
                CmpProp italicProp = this.propertyValues[ItalicID];
                if (italicProp.state >= State.Updated) return italic;
                Italic = toBool(EvalProp(ItalicID, false), false);
                return italic;
            }
            set { italic = value; SetUpdated(ref propertyValues[ItalicID]); }
        }

        public static int UnderlineID; // Property index. Set at initialization in ReflectDLL. 
        private bool underline = false;
        [VisPropAtt(VisPropAtt.PropertyType.Property, "", Position = 24, ShowInitially = false, 
            ShortHelp = "When true, the text is underlined. Default: false.")]
        public bool Underline {
            get {
                CmpProp underlineProp = this.propertyValues[UnderlineID];
                if (underlineProp.state >= State.Updated) return underline;
                Underline = toBool(EvalProp(UnderlineID, false), false);
                return underline;
            }
            set { underline = value; SetUpdated(ref propertyValues[UnderlineID]); }
        }

        public override void Paint(PaintEventArgs e) {
        Brush textBrush;
            if (Visible) {
                Rectangle textRec = new Rectangle(RelativeX(left) + PhyS(paddingLeft), RelativeY(top) + PhyS(paddingTop), 
                    PhyS(width) - PhyS(paddingLeft) - PhyS(paddingRight), PhyS(height) - PhyS(paddingTop) - PhyS(paddingBottom));
                e.Graphics.SmoothingMode = SmoothingMode.HighSpeed;
                StringFormat format = new StringFormat();
                if (!wordWrap) format.FormatFlags = StringFormatFlags.NoWrap;
                format.Alignment = (align < 1 ? StringAlignment.Near : align == 1 ? StringAlignment.Center : StringAlignment.Far);
                textBrush = new SolidBrush(TextColor);
                if (RotationAngle != 0 && RotationAngle % 360 != 0) {
                    e.Graphics.TranslateTransform(RelativeX(left), RelativeY(top));
                    e.Graphics.RotateTransform((float)RotationAngle);
                    if (BackColor != Color.Transparent && BackColor != Color.Empty)
                        e.Graphics.FillRectangle(new SolidBrush(BackColor), 0, 0, PhyS(width), PhyS(height));
                    if (BorderColor != Color.Transparent && BorderColor != Color.Empty && borderWeight > 0)
                        e.Graphics.DrawRectangle(new Pen(BorderColor, (float)borderWeight), 0, 0, PhyS(width), PhyS(height));
                    e.Graphics.DrawString(Text, font, textBrush, new RectangleF(PhyS(paddingLeft), PhyS(paddingTop), textRec.Width, textRec.Height), format);
                    e.Graphics.ResetTransform();
                } else {
                    if (BackColor != Color.Transparent && BackColor != Color.Empty)
                        e.Graphics.FillRectangle(new SolidBrush(BackColor), RelativeX(left), RelativeY(top), PhyS(width), PhyS(height));
                    if (BorderColor != Color.Transparent && BorderColor != Color.Empty && borderWeight > 0)
                        e.Graphics.DrawRectangle(new Pen(BorderColor, (float)borderWeight), RelativeX(left), RelativeY(top), PhyS(width), PhyS(height));
                    e.Graphics.DrawString(Text, font, textBrush, textRec, format);
                }
                base.Paint(e);
            }
        }
    }

    //----------------------------------------------------------------------------------
    [VisClassAtt(ToolboxCategory = "Control", ShortHelp = "Windows Textbox - slow when many", Icon = "textbox.bmp")]
    public class Textbox : VisComponent {

        private bool userChange = true; // False when the Text property has been set by uVis, e.g. at Init.
        // When false, ignores the associated TextChanged event.
        private bool textChanged = false; // True from the user has changed the text until source has been updated. Used to prevent void updates at LostFocus.
        private bool error = false; // True when upDateValue has found an error.

        public Textbox(int targetCount, DataRow row, Template tpl)
            : base(targetCount, row, tpl) {
            TextBox textBox = new TextBox();
            // textBox.Margin = new Padding(4); Has no effect.
            GuiInstance = textBox;
            GuiInstance.Tag = this;

            fontFamily = textBox.Font.FontFamily;
            fontSize = textBox.Font.SizeInPoints;

            //textBox.SelectionStart = 0; No effect until some text is set in Text.
            //textBox.SelectionLength = 0;
            textBox.TextChanged += TextChanged;
            textBox.KeyDown += keyDown;
            textBox.LostFocus += lostFocus;
            AttachControlEvents(textBox); // Attach the common control events, including KeyDown, etc.
        }

        public override void ResetUpdated() { // Called by kernel at refresh to set internal dependent properties to not-updated.
            sizeUpdated = false;
            base.ResetUpdated(); // Reset Height, etc.
        }

        public bool sizeUpdated = false;
        public void UpdateSize(int sourceID) { // Updates properties used to compute text size, e.g. Autosize and Bold.
            // Then computes text size. If autoSize, sets Width or Height.
            if (sizeUpdated) return;
            CmpProp sourceProp = this.propertyValues[sourceID];
            if (sourceProp.state >= State.Updated) return; // Init formulas are Initialized and don't need updates.
            sourceProp.state = State.Visited; // Start internal calculation. Catch cyclical references.

            FontStyle s = FontStyle.Regular;
            if (Bold) s = FontStyle.Bold;
            if (Italic) s = s | FontStyle.Italic;
            if (Underline) s = s | FontStyle.Underline;
            string fontName = FontName; // Evaluates also fontFamily.
            try {
                (this.GuiInstance as TextBox).Font = font = new Font(fontFamily, FontSize * SystemItems.Zoom2, s);
            }
            catch (Exception e) { // Some fonts don't support some styles, e.g. Aharoni. 
                ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp, e.Message, VisEngine.curExprToken, this.Index());
                font = DefaultFont;
            }

            SizeF size; // Compute text size. If autoSize, set Width or Height.
            Graphics dummy = Graphics.FromImage(new Bitmap(1, 1));
            StringFormat format = new StringFormat();
            if (!WordWrap) format.FormatFlags = StringFormatFlags.NoWrap;

            string test = Text; if (string.IsNullOrEmpty(text)) test = "A"; // Avoid that empty text boxes have zero size.
            if (wordWrap) { // WordWrap calculated above.
                if (sourceID == WidthID) sourceProp.state = State.Old; // Allow base.Width to work.
                size = dummy.MeasureString(test, font , PhyS(base.Width - 4), format); // Leave 4 bit margin for the border.
                TextHeight = size.Height / SystemItems.Zoom2;
                TextWidth = size.Width / SystemItems.Zoom2;
                if (AutoSize) {
                    base.Height = (int)(textHeight + 7.0F); // (int) truncates. TextBox needs 6 bits margin, probably for the border.
                }
            } else { // No wrap.
                size = dummy.MeasureString(test, font, 32000, format); // Any Width
                TextHeight = size.Height / SystemItems.Zoom2;
                TextWidth = size.Width / SystemItems.Zoom2; 
                if (AutoSize) {
                    base.Width = (int)(textWidth + 4.0);
                    base.Height = (int)(textHeight + 7.0); // (int) truncates. TextBox needs 5 bits margin, probably depending on the borer style.
                }
            }
            if (sourceProp.state == State.Visited) // If it hasn't been computed, reset it to Old.
                sourceProp.state = State.Old;
            sizeUpdated = true;
        }

        public override int Top {
            get { UpdateSize(TopID); return base.Top; } // Evaluates Top if not set by UpdateSize. 
            set { base.Top = value; }
        }

        public override int Height {
            get { UpdateSize(HeightID); return base.Height; } // Evaluates Height if not set by UpdateSize. 
            set { base.Height = value; }
        }

        public override int Bottom {
            get { UpdateSize(BottomID); return base.Bottom; } // Evaluates Bottom if not set by UpdateSize. 
            set { base.Bottom = value; }
        }

        public override int Left {
            get { UpdateSize(LeftID); return base.Left; }
            set { base.Left = value; }
        }

        public override int Width {
            get { UpdateSize(WidthID); return base.Width; }
            set { base.Width = value; }
        }

        public override int Right {
            get { UpdateSize(RightID); return base.Right; }
            set { base.Right = value; }
        }

        public static int TextHeightID; // Property index. Set at initialization in ReflectDLL. 
        public float textHeight;
        [VisPropAtt(VisPropAtt.PropertyType.Property, "' Read only", ReadOnly = true, ShowInitially = false, Position = 0,
            ShortHelp = "The height of the text in pixels.")]
        public float TextHeight {
            get { UpdateSize(TextHeightID); return textHeight; }
            set { textHeight = value; SetUpdated(ref propertyValues[TextHeightID]); }
        }

        public static int TextWidthID; // Property index. Set at initialization in ReflectDLL. 
        public float textWidth;
        [VisPropAtt(VisPropAtt.PropertyType.Property, "' Read only", ReadOnly = true, ShowInitially = false, Position = 1,
            ShortHelp = "The width of the text in pixels.")]
        public float TextWidth {
            get { UpdateSize(TextWidthID); return textWidth; }
            set { textWidth = value; SetUpdated(ref propertyValues[TextWidthID]); }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Eventhandler, null, Position = -10,
            LongHelp = @"Occurs when the user has changed the text.
Windows generates TextChanged also when the program sets Text. UVis doesn't",
            PropertyCategory = "EventHandler", ReadOnly = false, ShortHelp = "Occurs when the user has changed the text.")]
        public void TextChanged(object sender, EventArgs e) {
            //Debug.Print("TextChanged " + this.tpl.Name + ", userChange: " + this.userChange.ToString() + ", textChanged: " + this.textChanged.ToString());
            textChanged = true; // Changed and must be updated no matter whether user or event handler.
            // Is sometimes called without any change.
            if (this.userChange) {
                this.ExecuteHandler("TextChanged", true);
            } else
                userChange = true;
        }

        [VisPropAtt(VisPropAtt.PropertyType.Eventhandler, null, Position = -9, PropertyCategory = "EventHandler", ReadOnly = false,
            LongHelp = @"Occurs when I have transformed Text to System.NewValue, but before storing it in Source. 
Validate can modify System.NewValue, or block storing by setting System.Handled to true (makes the textbox yellow).",
            ShortHelp = "Occurs when the text has been transformed to System.NewValue, to be stored in Source")]
        public void Validate(object sender, EventArgs e) {
            //this.ExecuteHandler("Validate", true); The Validate event handler is called from updateValue( ).
        }

        [VisPropAtt(VisPropAtt.PropertyType.Eventhandler, null, Position = -8, PropertyCategory = "EventHandler", ReadOnly = false,
            ShortHelp = "Occurs after LostFocus or single-line Enter when I have stored the value in Source")]
        public void Updated(object sender, EventArgs e) {
            //this.ExecuteHandler("Updated", true); The Update event handler is called from updateValue( ).
        }

        public void keyDown(object sender, KeyEventArgs e) {
            if (e.KeyCode == Keys.Enter && !(GuiInstance as TextBox).Multiline) {
                updateValue();
            }
        }

        public void lostFocus(object sender, EventArgs e) {
            updateValue(); //base.LostFocus(sender, e); Has an event handler already.
        }

        bool updateValue() { // Maybe store the value in the source component or field.
            //Debug.Print("updateValue " + this.tpl.Name + ", textChanged: " + this.textChanged.ToString());
            if (!textChanged) 
                return true;
            textChanged = false;
            TextBox txtBox = GuiInstance as TextBox;
            try {
                if (String.IsNullOrWhiteSpace(txtBox.Text)) SystemItems.NewValue = null;
                else if (format == null) SystemItems.NewValue = txtBox.Text;
                else if (numberFormat) SystemItems.NewValue = Convert.ToDouble(txtBox.Text.Replace(" ", "")); // Remove spaces
                else SystemItems.NewValue = Convert.ToDateTime(txtBox.Text); // Don't remove spaces. Needed between date and time.
            }
            catch (Exception) { // Format errors in text.
                error = true;
                txtBox.BackColor = Color.Yellow;
                SystemItems.NewValue = null;
                return false;
            }
            try {
                SystemItems.Handled = false; // Default is that the TextBox makes the conversion.
                this.ExecuteHandler("Validate", false); // May change SystemItems.NewValue.
                if (!SystemItems.Handled) { // The event handler didn't object. Assign the value. 
                    TplPropValue sourceProp = tpl.valPropArray[SourceID];
                    if (sourceProp.expr != null) {
                        object setValue;
                        bool res = sourceProp.expr.Assign(this, tpl, SystemItems.NewValue, out setValue);
                        if (res) {
                            SystemItems.NewValue = setValue; // Use a possibly transformed value as NewValue.
                            Source = setValue;
                            this.ExecuteHandler("Updated", false); // May Refresh or set related values.
                        }
                    }
                } else { // Storing blocked. Leave text as is but with yellow background.
                    txtBox.BackColor = Color.Yellow;
                    return false;                
                }
                // Something stored. Based on data entered or value set by Validate.
                txtBox.BackColor = this.backColor;
                error = false;
                if (String.IsNullOrWhiteSpace(this.Text))
                    this.Text = null;
                else if (format == null)
                    this.Text = SystemItems.NewValue.ToString(); // Text records that it isn't a user change, so TextChanged should ignore.
                else
                    this.Text = vbSystemFunctions.vbFunctions.vbFormat(SystemItems.NewValue, format);
                this.userChange = true; // Allow TextChanged to act.
                return true;
            }
            catch (Exception e) { // Unexpected errors in Assign, etc.
                ErrorLogging.Log(ErrorMode.RunTime, Priority.Warning, tpl, "Source", e.Message, null, this.Index());
                return false;
            }
        }

        public static int TextID; // Property index. Set at initialization in ReflectDLL. 
        public string text;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, Position = 2, ReadOnly = false, ShowInitially = true,
            ShortHelp = "My text. When Source has a formula, I ignore any Text formula.")]
        public string Text {
            get {
                CmpProp textProp = this.propertyValues[TextID];
                if (textProp.state >= State.Updated) return GuiInstance.Text;
                if (this.tpl.valPropArray[SourceID].expr == null) { // There is no source
                    Text = toString(EvalProp(TextID, null), null); // No default value for a textbox.
                    return GuiInstance.Text;
                }
                textProp.state = State.Visited; // Start internal calculation. 
                object dummy = Source; // Sets Text
                return GuiInstance.Text;
            }
            set {
                CmpProp textProp = this.propertyValues[TextID];
                if (error) { textProp.state = State.Updated; return; }
                TextBox txt = GuiInstance as TextBox;
                if (textProp.state < State.Updated) // Text may be set from Source.
                    textProp.state = State.Updated;
                this.userChange = false; // Setting GuiInstance.Text triggers a TextChanged event. Should not call the event handler.
                txt.Text = text = value;
                txt.SelectionStart = 0; // Must be set before SelectionLength is set.
                txt.SelectionLength = 0;
                this.userChange = true; // Future changes will call the event handler.
            }
        }

        public static int SourceID; // Property index. Set at initialization in ReflectDLL. 
        private object source;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ShowInitially = true, Position = 3,
            ShortHelp = "The value I will show as a text, e.g. Me.field2 or box3!Height Default Null. When the user has changed the text, I will try updating the source.")]
        public object Source {
            get {
                CmpProp sourceProp = this.propertyValues[SourceID];
                if (sourceProp.state >= State.Updated) return source;
                if (this.tpl.valPropArray[SourceID].expr == null) {  // There is no source.
                    sourceProp.state = State.Updated; return null;
                }
                source = EvalProp(SourceID, null); // Sets state to updated.
                sourceProp.state = State.Visited;
                object dummy = Format;
                Source = source; // Sets also Text, based on Format.
                return source;
            }
            set { // Called from Source.get when there is a source expression, as well as assign to source.
                source = value; SetUpdated(ref propertyValues[SourceID]);
                if (source == null) 
                    this.Text = null;
                else if (format == null)
                    this.Text = source.ToString(); // Text records that this isn't a user change.
                else
                    this.Text = vbSystemFunctions.vbFunctions.vbFormat(source, format);
            }
        }

        private bool numberFormat; // Used in updateValue to convert text to date or number.
        public static int FormatID; // Property index. Set at initialization in ReflectDLL. 
        private string format = null;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ShowInitially = true, Position = 4,
            ShortHelp = "The format I will use to show the value, e.g. \"0.00\" or \"yyyy-MM-dd\". No format: I assume that Source is a text.")]
        public string Format {
            get {
                CmpProp formatProp = this.propertyValues[FormatID];
                if (formatProp.state >= State.Updated) return format;
                Format = toString(EvalProp(FormatID, null), null); // Sets state to updated.
                if (String.IsNullOrWhiteSpace(format)) { format = null; return null; }
                numberFormat = false;
                for (int i = 0; i < format.Length; i++) {
                    if (!Char.IsWhiteSpace(format, i)) { // The first non-blank character determines whether it is a number format.
                        Char c = format[i];
                        numberFormat = (c == '0' || c == '#' || c == 'g' || c == 'c' || c == 'f' || c == 'p' || c == 'e' || c == 'n' || c == 'x');
                        break;
                    }
                }
                return format;
            }
            set { format = value; SetUpdated(ref propertyValues[FormatID]); }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, "' Not supported", PropertyCategory = "Common", ReadOnly = true,
            ShowInitially = false, Exposed = false, InPopup = true,
            ShortHelp = "Windows' TextBox doesn't support BorderColor. TextColor is available.", Position = 12)]
        public override Color BorderColor {
            get { return Color.Black; }
        }

        public static int BorderStyleID; // Property index. Set at initialization in ReflectDLL. 
        private int style;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Special", ReadOnly = false, Position = 13,
            ShortHelp = "The style of my border: 0 (None), 1 (FixedSingle), 2 (Fixed3D). Default: 2 (Fixed3D)")]
        public int BorderStyle {
            get {
                CmpProp borderStyleProp = this.propertyValues[BorderStyleID];
                if (borderStyleProp.state < State.Updated)
                    BorderStyle = toPosInt32(EvalProp(BorderStyleID, 2), 2);
                return style;
            }
            set {
                (GuiInstance as TextBox).BorderStyle =
                (value == 0 ? System.Windows.Forms.BorderStyle.None : value == 1 ? System.Windows.Forms.BorderStyle.FixedSingle : System.Windows.Forms.BorderStyle.Fixed3D);
                style = value; SetUpdated(ref propertyValues[BorderStyleID]);
            }
        }

        public static int AlignID; // Property index. Set at initialization in ReflectDLL. 
        private int align;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Special", ReadOnly = false,
            ShortHelp = "Horizontal alignment of my text: 0 (Left), 1 (Center), 2 (Right). Default: 0 (Left)", Position = 6)]
        public int Align {
            get {
                CmpProp alignProp = this.propertyValues[AlignID];
                if (alignProp.state < State.Updated)
                    Align = toPosInt32(EvalProp(AlignID, 0), 0);
                return align;
            }
            set {
                (GuiInstance as TextBox).TextAlign =
                (value == 0 ? System.Windows.Forms.HorizontalAlignment.Left : value == 1 ? System.Windows.Forms.HorizontalAlignment.Center : System.Windows.Forms.HorizontalAlignment.Right);
                align = value; SetUpdated(ref propertyValues[AlignID]);
            }
        }

        public static int TextColorID; // Property index. Set at initialization in ReflectDLL. 
        public Color textColor = Color.White;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Special", ReadOnly = false,
            ShortHelp = "The color of my text. Default: black.", Position = 5)]
        public Color TextColor {
            get {
                CmpProp textColorProp = this.propertyValues[TextColorID];
                if (textColorProp.state < State.Updated)
                    TextColor = toColor(EvalProp(TextColorID, Color.Black), Color.Black);
                return textColor;
            }
            set { textColor = value; SetUpdated(ref propertyValues[TextColorID]); (GuiInstance as TextBox).ForeColor = value; }
        }

        public static int AutoSizeID; // Property index. Set at initialization in ReflectDLL. 
        public bool autoSize = false; // MS TextBox uses true.
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Special", ReadOnly = false, ShowInitially = true, Position = 7,
            LongHelp = "When true, my Height is determined by the text and if not Wordwrap, also my Width. Default: false.\nHeight and maybe also Width formulas are ignored.",
            ShortHelp = "When true, my Height is determined by the text and if not Wordwrap, also my Width. Default: false.")]
        public bool AutoSize { // The autosize property doesn't appear in MS's Intellisense list, yet exists.
            get {
                CmpProp autoSizeProp = this.propertyValues[AutoSizeID];
                if (autoSizeProp.state >= State.Updated) return autoSize;
                AutoSize = toBool(EvalProp(AutoSizeID, false), false);
                return autoSize;
            }
            set {
                (GuiInstance as TextBox).AutoSize = autoSize = value; SetUpdated(ref propertyValues[AutoSizeID]);
                (this.GuiInstance as TextBox).ScrollBars = wordWrap && !value ? ScrollBars.Vertical : ScrollBars.None;
            }
        }

        public static int WordWrapID; // Property index. Set at initialization in ReflectDLL. 
        private bool wordWrap = false; // MS TextBox uses true.
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ShowInitially = true, Position = 8,
            ShortHelp = "When true, my text is word-wrapped. Default: false.")]
        public bool WordWrap {
            get {
                CmpProp wordWrapProp = this.propertyValues[WordWrapID];
                if (wordWrapProp.state >= State.Updated) return wordWrap;
                WordWrap = toBool(EvalProp(WordWrapID, false), false);
                return wordWrap;
            }
            set {
                this.wordWrap = value; SetUpdated(ref propertyValues[WordWrapID]);
                (this.GuiInstance as TextBox).WordWrap = value;
                (this.GuiInstance as TextBox).Multiline = value;
                (this.GuiInstance as TextBox).ScrollBars = value && !autoSize ? ScrollBars.Vertical : ScrollBars.None;
            }
        }

        public static Font DefaultFont = new System.Drawing.Font("Tahoma", 8.50F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
        private FontFamily fontFamily = DefaultFont.FontFamily;
        private static string fontNameDefault = "Tahoma";
        Font font; 

        public static int FontNameID; // Property index. Set at initialization in ReflectDLL. 
        private string fontName;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, Position = 20,
            LongHelp = "The font for my text, e.g. Font.Arial or \"Times New Roman\". Default: \"Tahoma\"\nWhen the font name is more than one word, you have to manually enclose it in quotes.",
            ShowInitially = false, ShortHelp = "The font for my text, e.g. Font.Arial or \"Times New Roman\". Default: \"Tahoma\"")]
        public string FontName {
            get {
                CmpProp fontNameProp = this.propertyValues[FontNameID];
                if (fontNameProp.state >= State.Updated) return fontName;
                FontName = toString(EvalProp(FontNameID, fontNameDefault), fontNameDefault);
                try {
                    fontFamily = new FontFamily(fontName);
                }
                catch (System.ArgumentException) {
                    ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp, "Font " + fontName + " doesn't exist", VisEngine.curExprToken, this.Index());
                }
                return fontName;
            }
            set { fontName = value; SetUpdated(ref propertyValues[FontNameID]); }
        }

        public static int FontSizeID; // Property index. Set at initialization in ReflectDLL. 
        private float fontSize = 8.50F; // Same as MS default for TextBox and Label.
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, Position = 21, ShowInitially = false,
            ShortHelp = "The size of my text in points. Default: 8.50")]
        public float FontSize {
            get {
                CmpProp fontSizeProp = this.propertyValues[FontSizeID];
                if (fontSizeProp.state >= State.Updated) return fontSize;
                FontSize = (float)toDouble(EvalProp(FontSizeID, 8.50), 8.50);
                return fontSize;
            }
            set { fontSize = value; SetUpdated(ref propertyValues[FontSizeID]); }
        }

        public static int BoldID; // Property index. Set at initialization in ReflectDLL. 
        private bool bold = false;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, Position = 22, ShowInitially = false,
            ShortHelp = "When true, the text is bold. Default: false.")]
        public bool Bold {
            get {
                CmpProp boldProp = this.propertyValues[BoldID];
                if (boldProp.state >= State.Updated) return bold;
                Bold = toBool(EvalProp(BoldID, false), false);
                return bold;
            }
            set { bold = value; SetUpdated(ref propertyValues[BoldID]); }
        }

        public static int ItalicID; // Property index. Set at initialization in ReflectDLL. 
        private bool italic = false;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, Position = 23, ShowInitially = false,
            ShortHelp = "When true, the text is Italic. Default: false.")]
        public bool Italic {
            get {
                CmpProp italicProp = this.propertyValues[ItalicID];
                if (italicProp.state >= State.Updated) return italic;
                Italic = toBool(EvalProp(ItalicID, false), false);
                return italic;
            }
            set { italic = value; SetUpdated(ref propertyValues[ItalicID]); }
        }

        public static int UnderlineID; // Property index. Set at initialization in ReflectDLL. 
        private bool underline = false;
        [VisPropAtt(VisPropAtt.PropertyType.Property, "", Position = 24, ShowInitially = false,
            ShortHelp = "When true, the text is underlined. Default: false.")]
        public bool Underline {
            get {
                CmpProp underlineProp = this.propertyValues[UnderlineID];
                if (underlineProp.state >= State.Updated) return underline;
                Underline = toBool(EvalProp(UnderlineID, false), false);
                return underline;
            }
            set { underline = value; SetUpdated(ref propertyValues[UnderlineID]); }
        }
    }
