﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using VisTool;
using Trace = System.Diagnostics.Trace;

namespace VisTool
{ /* Error handling (see also OverAll flow.doc):
   * An error is stored in several related data structures: 
   * ErrorLogging.Errors: One shared List<Error>. Each Error contains names of the form, template and property, the message text, ErrorMode and Priority,
   * plus references to template, template property and the token relating to the error. These references may be null for unexpected exceptions.
   * The list is cleaned before refresh and at CloseForm, so that the references are valid unless null.
   * 
   * TemplateProperty.tokenErrors: A List<TokenError> for each template property. Each TokenError contains a reference to the Error and to the token.
   * The lists are cleaned when the Errors list is cleaned.
   * 
   * Token: The scanner reads a vis-file or a formula string and transforms it into a chain of tokens. When the scanner detects an error (string not terminated in the line, 
   * dateTime error, etc.), it sets Token.error to a message string. It cannot log the error properly because the templates and properties have not yet been created.
   * The vis-file Parser gets tokens from the scanner, but the scanner may be a couple of tokens ahead. The vis-file parser creates templates and properties, and 
   * breaks up the token chain into a chain for each formula. It too finds errors and logs them and the scanner errors, with reference to the proper template and property.
   * Later the formula parser compiles each chain of tokens into expression objects. It too logs errors. Finally, when the templates are instantiated into controls, 
   * the formulas are evaluated. This may log additional errors.
   * 
   * Pos, line and col: Each token may contain three integers: pos, line and col. They indicate the first non-space character of the token. Pos is the char number in 
   * the string (0-based), line is the line number in the possibly multi-line formula (1-based), and col is the character number in the line (1-based).
   * These position numbers are calculated and set when an error in the formula is logged. This also supports the case where the compiler auto-corrects the formula, 
   * for instance changing a dot (.) to a join (-<).   
   * 
   * ErrorLog.txt: All errors are also logged in text form in ErrorLog.txt. Also debug tracing such as Debug.Print end up here. VisTool creates ErrorLog.txt when it starts up.
   */

    public enum ErrorMode {
        Vism, // Error in vis file. Not visible in property grid, only as error line.
        CompileTime, // CompileTime errors will be removed before re-compile
        RunTime // RunTime errors will be removed before any refresh
    }
    public enum Priority {
        MsgBox, // Shows a message box, e.g. File does not exist, Form does not exist
        Error, // Most errors are of this priority
        Warning // E.g. database server may not ...
    }

    public class Error {
        // Contains dual information: (1) The error data in text form - for logging and human reading.
        // (2) The error as references to template and token - for clicking and showing the property box.
        public ErrorMode mode;
        public Priority priority;
        public string formName;
        public string templateName;
        public string propertyName;
        public string ErrorMessage;
        public Template tpl; // Null if the error cannot be related to a property
        public TplProperty tplProp; // Null if the error cannot be related to a property.
        public int Index = 0; // Index of the component that had the error. RunTime errors only. Ignore if Index < 0.
        public Token token; // Null if the error cannot be related to a token

        public Error(ErrorMode mode, Priority priority, string formName, string templateName, string propertyName,
            string errorMessage, Template tpl, TplProperty tplProp, Token t, int Index) {
            this.mode = mode;
            this.priority = priority;
            this.formName = formName; // In most cases the form will be open, but may also be a closed or unknown form.
            this.templateName = templateName; // In most cases this template will exist, but some errors may refer to a template being created.
            this.propertyName = propertyName; // In most cases this property will exist.
            this.ErrorMessage = errorMessage; // May be defined by VisTool or a message from .NET, e.g. "divide by zero"
            this.tpl = tpl;
            this.tplProp = tplProp;
            this.token = t;
            this.Index = Index;
        }
    }

    public class TokenError {
        public Error error; // The logged error
        public Token t; // The token having the error

        public TokenError(Error e, Token t) {
            this.error = e; this.t = t;
        }
    }

    public class ErrorLogging
    {
        public static ObservableCollection<Error> Errors = new ObservableCollection<Error>(); // To be shown by VistoolStudio or thirdparty.
        static TextWriterTraceListener traceListener;

        static ErrorLogging() { // Initialize log and listener.
            FileStream traceLog=null;
            FileInfo logFile = null;
            int fileNumber;
            // Cyclically use ErrorLog0, 1, ... and make sure the next one in the cycle is deleted.
            for (fileNumber = 0; fileNumber < 5; fileNumber++) {
                logFile = new FileInfo(VisEngine.ProgramDirectory + "\\ErrorLog" + fileNumber + ".txt");
                if (!logFile.Exists) break;
            }
            traceLog = new FileStream(VisEngine.ProgramDirectory + "\\ErrorLog" + fileNumber + ".txt", FileMode.Create);
            traceListener = new TextWriterTraceListener(traceLog);
            Trace.AutoFlush = true;
            Trace.Listeners.Add(traceListener);
            traceListener.WriteLine(DateTime.Now.ToString() + ", " + VisEngine.ProgramDirectory);

            logFile = new FileInfo(VisEngine.ProgramDirectory + "\\ErrorLog" + (fileNumber + 1) % 5 + ".txt");
            logFile.Delete(); // Make sure there is an unused logfile name for the next run.
        }

        public static void Log(string str) {
            traceListener.WriteLine(str);
        }

        public static void Log(ErrorMode mode, Priority priority, Template tpl, string propName, string errorMessage, Token t, int Index) {
            // Used in cases where the property may have no formula, e.g. for internal dependencies.
            PropInfo propInfo;
            TplProperty tplProp = TplProperty.GetTemplateProperty(tpl, propName, out propInfo);
            if (tpl.Info.PropertyByName.TryGetValue(propName, out propInfo))
                Log(mode, priority, tpl, tplProp, errorMessage, t, Index);
            else
                Log(mode, priority, tpl, tplProp, propName + ": " + errorMessage, t, Index); // tplProp is null in this case.
        }

        public static void Log(ErrorMode mode, Priority priority, Template tpl, TplProperty tplProp, string errorMessage, Token t, int Index) {
            // tpl, tplProp and t may be null for unexpected errors. Don't crash for this reason.
            bool proper = false;
            string frmName = "Frm?", tplName = "Template?", propName = "Property?", val = "Token?"; // Default values.
            int line = 1, col = 1, pos = 0;

            if (tpl != null) { tplName = tpl.Name; frmName = tpl.FormTpl.Name; }
            if (tplProp != null) propName = tplProp.Name;
            if (tpl != null && tplProp != null && tplProp.startChain != null) { // Proper error report
                if (t == null) { // Error relates to entire property. Mark it as the end token.
                    t = tplProp.startChain;
                    while (t.next != null && t.kind != Parser.TkKind.EOL) t = t.next;
                }
                foreach (TokenError tk in tplProp.tokenErrors) if (tk.t == t && tk.error.ErrorMessage == errorMessage) return; 
                // An error with the same message is already logged for this token. (Errors may be produced for each instance of a template.)
                // A token may report more than one error, but with different messages, e.g. for cyclical references.
                proper = true;
                // Compute token positions in the formula. Done several times when there are several errors in the formula. Don't care !
                val = t.val;
                Token c = tplProp.startChain;

                ComputeTokenPosition(t, c, ref line, ref col, ref pos);

            }  // End proper error report 

            VisEngine.ErrorCount++;
            if (VisEngine.ErrorCount > VisEngine.MaxErrors) {
                if (VisEngine.MaxErrorReported) return;
                VisEngine.MaxErrorReported = true;
                errorMessage = "Reported more than " + VisEngine.MaxErrors + " errors. Stopped reporting.";
            }
            var error = new Error(mode, priority, frmName, tplName, propName, errorMessage, tpl, tplProp, t, Index);
            Errors.Add(error);
            if (proper) tplProp.tokenErrors.Add(new TokenError(error, t));
            // Compute text to log or display in a message box:
            string msgText1 = frmName + "." + tplName + (ErrorMode.RunTime == mode && Index >= 0 ? "[" + Index + "]" : "") + "! " + propName +
                (t==null ? "(pos?) " : " (pos " + t.pos + ", line " + t.line + ", col " + t.col + (t.val.Length>0? (", token "+t.val) : "") +") ");
            string msgText2 = errorMessage;
            Trace.WriteLine((priority == Priority.Warning ? "Warning: " : "Error: ") + msgText1 + msgText2);
            if (priority == Priority.MsgBox)
                MessageBox.Show(msgText1 + "\n" + msgText2, "VisTool", MessageBoxButtons.OK, MessageBoxIcon.Warning);
            // PrintErrors(); 
        }

        public static void Log(ErrorMode mode, Priority priority, String name1, String name2, String name3, String errorMessage, Token t) {
        // Error logging used by VismParser. Use null for string parameters if they are irrelevant.
            // name1: file name, name2: e.g. table name, name3: e.g. relation name.
            VisEngine.ErrorCount++;

            Error error = new Error(ErrorMode.Vism, priority, name1, name2, name3, errorMessage, null, null, t, 0);
            // Will be shown in Studio. t.line must match line numbers in Notepad.

            Errors.Add(error);
            // Compute the first message text string
            string msgText1 = name1 + (name2 == null ? "" : ", ") + name2 + (name3 == null ? "" : ", ") + name3
            + (t == null ? " (pos?)" : ", line " + t.line + ", " + ", token " + t.val);

            //compute the second message text string
            string msgText2 = errorMessage;

            //write the error to the Trace or display a messagebox 
            Trace.WriteLine((priority == Priority.Warning ? "Warning: " : "Error: ") + msgText1 + msgText2);
            if (priority == Priority.MsgBox)
                MessageBox.Show(msgText1 + "\n" + msgText2, "VisTool", MessageBoxButtons.OK, MessageBoxIcon.Warning);
        }

        /// <summary>
        /// Computes the position of the token target, from the beginning of the property (startToken) 
        /// </summary>
        /// <param name="target"> the token for which the position will be returned </param>
        /// <param name="curToken"> the token at the beginning of the statement / assignment </param>
        /// <param name="line"> uninitialized but declared beforehand</param>
        /// <param name="col"> uninitialized but declared beforehand </param>
        /// <param name="pos"> uninitialized but declared beforehand </param>
        private static void ComputeTokenPosition(Token target, Token startToken, ref int line, ref int col, ref int pos){
            // Sets pos in all tokens from startToken to target. Pos of a token, is the number of characters until the val-part of the token.
            Token curToken = startToken;
            while (true) {
                if (curToken.lines == 0) col += curToken.space.Length;
                else {
                    line += curToken.lines;
                    col = curToken.space.Length - curToken.space.LastIndexOf('\n');
                };
                curToken.col = col;
                curToken.pos = pos += curToken.space.Length;
                col += curToken.val.Length;
                pos += curToken.val.Length;
                if (curToken == target || curToken == curToken.next || curToken.next == null) break;
                curToken = curToken.next;
            }
        }

        // Retrieves any Error related to a property in an open form. Returns null when there isn't an error.
        public static Error GetError(TplProperty tplProp) {
            for (int i = 0; i < Errors.Count; i++) {
                if (Errors[i].tplProp== tplProp)
                    return Errors[i];
            }
            return null;
        }

        public static void ClearRuntimeErrors( ) { // Used at Refresh with ValueChange
            for (int i = 0; i < Errors.Count; i++) {
                Error e = Errors[i];
                if (e.mode == ErrorMode.RunTime) {
                    if (e.tplProp != null) {
                        TplProperty p = e.tplProp;
                        for (int j = 0; j < p.tokenErrors.Count; j++)
                            if (p.tokenErrors[j].error == e) {
                                p.tokenErrors.RemoveRange(j, 1); j--; break;
                            }
                    }
                    Errors.RemoveAt(i); i--;
                }
            }
            VisEngine.ErrorCount = 0; VisEngine.MaxErrorReported = false;
        }

        public static void ClearAllErrors(string frmName ) { // Used at Refresh with recompile and at form
            for (int i = 0; i < Errors.Count; i++) {
                Error e = Errors[i];
                if (e.formName == frmName) {
                    if (e.tplProp != null) e.tplProp.tokenErrors.Clear();
                    Errors.RemoveAt(i); i--;
                } else if (e.tplProp == null) {
                    Errors.RemoveAt(i); i--; // E.g. errors from vism parser
                }
            }
            VisEngine.ErrorCount = 0; VisEngine.MaxErrorReported = false;
        }

        static void PrintErrors() { // For debug only
            Debug.Print("ERROR LIST");
            string msgText1;
            for (int i = 0; i < Errors.Count; i++) {
                Error e = Errors[i];
                msgText1 = e.priority.ToString() + " " +  e.formName + "." + e.templateName + "." + e.propertyName +
                    (e.token == null ? "" : " (pos " + e.token.pos.ToString() + ", " + e.token.line.ToString() + ", " + e.token.col.ToString() + (e.token.val.Length > 0 ? (", token " + e.token.val) : "") + ") ");
                string msgText2 = e.ErrorMessage;
                Debug.Print(msgText1 + msgText2);
            }
        }
    }
}
