﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Windows.Forms;
using VisTool;

namespace VisTool {

    public partial class VisEngine { 
        // The methods with names like MmmNoLock assume that the form is locked and don't unlock it.
        public static bool LockForm(Template tpl) { // True if the Form can be updated. Tests that the Form isn't being updated. Then locks the Form.
            if (tpl.FormTpl.updating) 
                return false;
            tpl.FormTpl.updating = true;
            return true;
        }

        public static bool LockNoClones(Template tpl) { // True if the Form can be updated. Tests that the Form isn't being updated and isn't a clone. Then locks the Form.
            if (tpl.FormTpl.updating)
                return false;
            FormTemplate frmThis = tpl.FormTpl;
            for (int i = 0; i < tpl.FormTpl.clones.Count; i++) { // Don't use enumerator. Changes will take place in the collection.
                FormTemplate frmTpl = tpl.FormTpl.clones[i];
                if (frmTpl == frmThis) continue;
                if (frmTpl != null) {
                    if (frmTpl.open) {
                        VisForm frmCmp = frmTpl.RowBundle.Instances[0].cmp as VisForm;
                        (frmCmp.GuiInstance as Form).Close(); // Will close other forms in the form bundle.
                    }
                    tpl.FormTpl.clones[i] = null; // Delete loaded forms too.
                }
            }
            tpl.FormTpl.updating = true; // multiClone may show an error dialogue. FocusLost events will occur. Lock first.
            return true;
        }

        public static void UnLockForm(Template tpl) { // True if the Form can be updated. Test that the Form isn't being updated and isn't a clone. Then locks the Form.
            tpl.FormTpl.updating = false; return;
        }

        public static Template CreateTemplate(string name, ComponentInfo info, FormTemplate frm) {
            if (!LockNoClones(frm)) return null;
            Template tpl = CreateTemplateNoLock(name, info, frm);
            UnLockForm(frm);
            return tpl;
        }

        public static Template CreateTemplateNoLock(string name, ComponentInfo info, FormTemplate frm) { 
            // Adds the Name property. Caller adds other properties.
            // Might be merged with VisParser.CreateControlTemplate, but the latter has a very different error handling and cooperates with CreateParentTree.
            Template existingTemplate;
            if (frm.Templates.TryGetValue(name, out existingTemplate)) {
                MessageBox.Show("Another template has this name"); return null;
            }
            PropInfo propInfo;
            if (TplProperty.GetTemplateProperty(frm, name, out propInfo) != null) {
                MessageBox.Show(name + " is also a Form property name"); return null;
            }
            var tpl = new Template(name, info, frm);
            tpl.FormTpl = frm;
            frm.Templates.Add(name, tpl);
            tpl.Name = name;
            CreateOrSetTemplatePropertyNoLock(tpl, "Name", name); // SL 18-12-2012. Name used to hold comments.
            frm.isDirty = true;
            refreshMode = RefreshMode.FormulaChange;
            return tpl;
        }

        public static Template CreateTemplate(Template tpl, FormTemplate frm) { // Add the template to the Form
            // The properties, Name, etc. have been set already. Intended for Undo.
            if (frm.Templates.ContainsKey(tpl.Name)) {
                MessageBox.Show("Another template has this name"); return null;
            }
            if (!LockNoClones(frm)) return null;
            tpl.FormTpl = frm;
            frm.Templates.Add(tpl.Name, tpl);
            frm.isDirty = true;
            refreshMode = RefreshMode.FormulaChange;
            UnLockForm(tpl);
            return tpl;
        }

        public static Template CopyTemplate(Template tpl, FormTemplate newFrm) { 
            // Copy existing template to frm. May be another Form than tpl's form.
            if (tpl is FormTemplate) {
                MessageBox.Show("Cannot copy a Form"); return null;
            }
            if (!LockNoClones(newFrm)) return null;
            int number = 1; 
            string newTplName = tpl.Name; // Try with old name first. When copying to another form, the old name may be ok.
            while (newFrm.Templates.ContainsKey(newTplName)) // Not an existing template name.
                newTplName = tpl.Name + "_" + (number++).ToString();
            while (newFrm.Info.PropertyByName.ContainsKey(newTplName)) // Not a built in property name.
                newTplName = tpl.Name + "_" + (number++).ToString();
            while (newFrm.DesignProperties.ContainsKey(newTplName)) // Not a designer property name.
                newTplName = tpl.Name + "_" + (number++).ToString();

            Template newTpl = CreateTemplateNoLock(newTplName, tpl.Info, newFrm);
            if (newTpl == null) { UnLockForm(tpl); return null; } // Maybe multiple clones.
            for (int pi = 1; pi < tpl.valPropArray.Length; pi++) { // Skip name property. Has been set by CreateTemplate.
                TplProperty tplProp = tpl.valPropArray[pi];
                CreateOrSetTemplatePropertyNoLock(newTpl, tplProp.Name, tplProp.Formula);
            }
            foreach (TplProperty tplProp in tpl.evtPropArray)
                CreateOrSetTemplatePropertyNoLock(newTpl, tplProp.Name, tplProp.Formula);
            foreach (var item in tpl.DesignProperties) {
                TplPropDesigner tplProp = item.Value;
                CreateOrSetTemplatePropertyNoLock(newTpl, tplProp.Name, tplProp.Formula);
            }
            newFrm.isDirty = true;
            UnLockForm(newFrm);
            return newTpl;
        }

        public static bool ChangeTemplateType(Template tpl, string typeName) {
            if (!LockNoClones(tpl)) return false;
            if (tpl is FormTemplate) {
                MessageBox.Show("Cannot change the type of the Form");
                UnLockForm(tpl); return false;
            }
            ComponentInfo info;
            if (!VisEngine.ComponentByName.TryGetValue(typeName, out info)) { 
                MessageBox.Show("Cannot find control type: " + typeName);
                UnLockForm(tpl); return false; 
            }
            //Copy all properties. This may change some properties from Design to guiProperty or vice versa. 
            var oldValProp = tpl.valPropArray;
            var oldEvtProp = tpl.evtPropArray;
            var oldDesign = tpl.DesignProperties; // You cannot remove properties while scanning them, so make a copy for scanning
            tpl.CreateTplProps(info);

            foreach (var prop in oldValProp) {
                if (prop.Formula != null) CreateOrSetTemplatePropertyNoLock(tpl, prop.Name, prop.Formula);
            }
            foreach (var prop in oldEvtProp) {
                if (prop.Formula != null) CreateOrSetTemplatePropertyNoLock(tpl, prop.Name, prop.Formula); // Locks and unlocks.
            }
            foreach (var prop in oldDesign) {
                TplProperty p = prop.Value;
                CreateOrSetTemplatePropertyNoLock(tpl, p.Name, p.Formula);
            }
            refreshMode = RefreshMode.FormulaChange; // Needed to remove all instances.
            tpl.FormTpl.isDirty = true;
            UnLockForm(tpl);
            return true;
        }

        public static bool DeleteTemplate(Template tpl) {
            if (tpl is FormTemplate) { // The template is a Form
                MessageBox.Show("You cannot delete the Form this way.\r\nDelete the vis-file, e.g. from the Application explorer.");
                return false;
            }
            if (!LockNoClones(tpl)) return false;
            FormTemplate frmTpl = tpl.FormTpl;
            refreshMode = RefreshMode.FormulaChange;
            tpl.deleteInstances(); // Delete all template instances and unload their table. TreeChange doesn't do it because we remove the template.
            frmTpl.Templates.Remove(tpl.Name);
            frmTpl.isDirty = true;
            UnLockForm(frmTpl);
            return true;
        }

        // -----------------------------------------------Template properties------------------------------------
        public static bool DeleteTemplateProperty(Template tpl, TplProperty tplProp) {
            if (!LockNoClones(tpl)) return false;
            bool res = DeleteTemplatePropertyNoLock(tpl, tplProp);
            UnLockForm(tpl);
            return res;
        }

        public static bool DeleteTemplatePropertyNoLock(Template tpl, TplProperty tplProp) {
            FormTemplate frmTpl = tpl.FormTpl;
            PropInfo propInfo;
            if (tpl.Info.PropertyByName.TryGetValue(tplProp.Name, out propInfo)) {
                if (propInfo is ValueInfo)
                    tplProp = tpl.valPropArray[propInfo.propIndex];
                else if (propInfo is EvtInfo)
                    tplProp = tpl.evtPropArray[propInfo.propIndex];
                else {
                    return false; // Cannot delete Function property.
                }
                tplProp.Formula = null; tplProp.expr = null;
            } else {
                tpl.DesignProperties.Remove(tplProp.Name); 
                // Now instancePropIndex is inconsistent and too large. Renumber them. Refresh will adjust the instance properties.
                int dIndex = 0;
                foreach (var dProp in tpl.DesignProperties)
                    dProp.Value.cmpPropIndex = dIndex++;
                tplProp.Name = null; // Mark it as dangling. Name editor in Studio may use it.
            }
            refreshMode = RefreshMode.FormulaChange; // Needed to update InstancePropertyIndexes.
            frmTpl.isDirty = true;
            return true;
        }

        public static bool RenameTemplateProperty(Template tpl, TplProperty tplProp, string newName) {
            if (!LockNoClones(tpl)) return false;
            FormTemplate frmTpl = tpl.FormTpl;
            if (newName.ToUpper() == tplProp.Name.ToUpper()) {
                tplProp.Name = newName; // Upper/lower may have changed.
                if (tplProp.Name != newName)  tpl.FormTpl.isDirty = true; // Name caps have changed. Save it.
                UnLockForm(tpl);
                return true; 
            }
            if (tpl.DesignProperties.ContainsKey(newName) || newName.ToUpper()=="LAST" || newName.ToUpper()=="INDEX") {
                MessageBox.Show("Property " + newName + " already exists");
                UnLockForm(tpl);
                return false;
            }
            if (tpl is FormTemplate && frmTpl.Templates.ContainsKey(newName)) {
                MessageBox.Show("Property " + newName + " is the name of a template");
                UnLockForm(tpl);
                return false;
            }
            PropInfo propInfo;
            if (tpl.Info.PropertyByName.TryGetValue(newName, out propInfo)) {
                if (propInfo is ValueInfo && tpl.valPropArray[propInfo.propIndex].expr != null ||
                    propInfo is EvtInfo && tpl.evtPropArray[propInfo.propIndex].expr != null ||
                    propInfo is FunctInfo ) {
                        MessageBox.Show("Property " + newName + " already exists");
                        UnLockForm(tpl);
                        return false;
                    }
                }

            string formula = tplProp.Formula;
            if (!DeleteTemplatePropertyNoLock(tpl, tplProp)) {
                UnLockForm(tpl); return false; 
            }
            CreateOrSetTemplatePropertyNoLock(tpl, newName, formula); 
            refreshMode = RefreshMode.FormulaChange; // Necessary to reset the property that had the old name. SL 01-07-12
            tpl.FormTpl.isDirty = true;
            UnLockForm(tpl); 
            return true;
        }

        static bool CheckAndSetTemplateNameNoLock(Template tpl, TplProperty prop, string formula) {
            // True if the formula was set
            if (formula == null) return false; // Formula may be null.
            FormTemplate frmTpl = tpl.FormTpl;
            Parser p = null;
            if (formula != null) {
                p = new Parser(tpl, new StringReader(formula), prop, false); // Uses or sets startChain. Sets refs for error handling.
                p.Get();
            }
            if (formula == null || p.token.kind != Parser.TkKind.Ident) {
                MessageBox.Show("The name formula must start with a template name. You can add comments after the name.");
                return false;
            }

            string name = p.token.val;
            if (formula == tpl.valPropArray[VisComponent.NameID].Formula) 
                return true; // The comment may have changed
            if (tpl is FormTemplate) { // Form comments may change.
                if (name != tpl.Name) {
                    try {
                        System.IO.File.Move(tpl.Name + ".vis", name + ".vis");
                    }
                    catch (Exception e) {
                        MessageBox.Show("Form " + name + " exists already");
                        return false;
                    }
                    VisEngine.UpdateFormDirectories(tpl.Name, name);
                    frmTpl.Templates.Remove(tpl.Name); frmTpl.Templates.Add(name, tpl);
                    tpl.Name = name;
                }
                return true;
            }
            if (name != tpl.Name && tpl.FormTpl.Templates.ContainsKey(name)) {
                MessageBox.Show( "Template \""+ name + "\" already exists");
                return false; 
            }
            PropInfo propInfo;
            if (TplProperty.GetTemplateProperty(tpl.FormTpl, name, out propInfo) != null) {
                MessageBox.Show(name + " is also a Form property name"); return false;
            }
            tpl.FormTpl.Templates.Remove(tpl.Name);
            tpl.FormTpl.Templates.Add(name, tpl);
            tpl.Name = name;
            tpl.valPropArray[VisComponent.NameID].Formula = formula;
            if (p.la.val.ToUpper() == "IGNORE") {
                tpl.Ignore = true; p.Get();
            } else
                tpl.Ignore = false;
            if (p.la.kind != Parser.TkKind.EOL)
                MessageBox.Show("Unexpected symbol in Name formula: " + p.la.val);
            refreshMode = RefreshMode.FormulaChange; // Needed to reset the property that had the old name. SL 01-07-12
            tpl.FormTpl.isDirty = true;
            return true;
        }

        // Sets the property formula. Creates the property if it doesn't exist. Returns null if nothing has changed.
        // Checks for Form being updated and multiple clones.
        public static TplProperty CreateOrSetTemplateProperty(Template tpl, string propertyName, string formula) {
            if (!LockNoClones(tpl)) return null;
            TplProperty tplProp = CreateOrSetTemplatePropertyNoLock(tpl, propertyName, formula);
            UnLockForm(tpl);
            return tplProp;
        }
        
        // Sets the property formula. When formula==null, the property disappears. Returns null if nothing has changed.
        public static TplProperty CreateOrSetTemplatePropertyNoLock(Template tpl, string propertyName, string formula) {
            PropInfo propInfo;
            TplProperty tplProp;
            if (tpl.Info.PropertyByName.TryGetValue(propertyName, out propInfo)) { // The property is a built-in property.
                if (propInfo is ValueInfo)
                    tplProp = tpl.valPropArray[propInfo.propIndex];
                else if (propInfo is EvtInfo)
                    tplProp = tpl.evtPropArray[propInfo.propIndex];
                else
                    return null; // The property is a Function property and doesn't have a formula.

                FormTemplate frmTpl = tpl.FormTpl;
                if (refreshMode < RefreshMode.FormulaChange) refreshMode = RefreshMode.FormulaChange;
                tplProp.startChain = null; // Force scan of the formula.
                switch (propertyName.ToUpper()) {
                case "NAME":
                    if (!CheckAndSetTemplateNameNoLock(tpl, tplProp, formula)) return tplProp; // The old name is kept. Errors have been reported.
                    break;
                case "ROWS":
                    tpl.RowsCompiled = false; // Otherwise Formula won't compile.
                    break; 
                }
                tplProp.expr = null;
                tplProp.Formula = formula; // (formula == null ? " " : formula); // A blank is needed for error reporting. Needs a startChain.
                // Also empty formulas may have errors, e.g. a mandatory property or an internal circular reference. 
                frmTpl.isDirty = true;
                return tplProp;
            }

            // The property is an existing designer property or a new one.
            TplPropDesigner propDesigner;
            if (!tpl.DesignProperties.TryGetValue(propertyName, out propDesigner)) {
                propDesigner = new TplPropDesigner(tpl, propertyName, typeof(object), PropertyKind.Designer);
                propDesigner.cmpPropIndex = tpl.DesignProperties.Count;
                tpl.DesignProperties[propertyName] = propDesigner; // Add the property.
            }
            if (propDesigner.Formula == formula || String.IsNullOrWhiteSpace(propDesigner.Formula) && String.IsNullOrWhiteSpace(formula))
                return propDesigner;
            if (refreshMode < RefreshMode.FormulaChange) refreshMode = RefreshMode.FormulaChange;
            propDesigner.startChain = null; // Force scan of the formula.
            propDesigner.Formula = (formula == null ? " " : formula); // A blank is needed for error reporting. Needs a startChain.
            // Trim unnecessary. A possible first blank has been removed by the scanner in AdjustPos. The rest should be kept.
            tpl.FormTpl.isDirty = true;
            return propDesigner;
        }

        // This method might fit well in the VisComponent class. 
        public static object GetValue(VisComponent cmp, string propName, out bool propExists) { // Will update the value when needed.
            // Sets propExists = false if the property isn't value or designer or if the name doesn't exist. A bit slow because it uses dictionary lookup to find the property.
            // Caller must report circular references because GetValue doesn't know the token for the report
            ValueInfo info; propExists = false; 
            object res;
            if (propName == null) return null;
            Template tpl = cmp.tpl;
            if (tpl.Info.ValuePropertyByName.TryGetValue(propName, out info)) {
                res = info.PropertySetter.GetValue(cmp, null); // Covers built-in value properties, also those with an Init formula. SL 06-06-14.
                propExists = true; return res;
            }
            TplPropDesigner tplProp;
            if ( tpl.DesignProperties.TryGetValue(propName, out tplProp)) {
                tplProp.EvalAndSet(cmp);
                res = cmp.designerPropValues[tplProp.cmpPropIndex].Value;
                propExists = true; return res;
            }
            return null; // An event handler or a property function.
        }

        // --------------------------------------------- Symbolic control addressing --------------------------------------
        // Gets the template property from form name, template name and property name. Looks only for open forms.
        public static TplProperty GetProperty(string frmName, string tplName, string propName) {
            Template tpl; TplProperty tplProp; PropInfo propInfo; int cloneID = 0;
            FormTemplate frmTpl = VisEngine.GetFormTemplate(frmName, ref cloneID); // Any clone will do.
            if (frmTpl == null || !frmTpl.open) return null;
            if (tplName == null || propName == null || !frmTpl.Templates.TryGetValue(tplName, out tpl) || tpl.Ignore) return null;
            tplProp = TplProperty.GetTemplateProperty(tpl, propName, out propInfo);
            return tplProp;
        }

        // Creates a control address as a sequence of template names and indexes. Can later restore the control ref to the closest possible, e.g. a lower index.
        public class ControlAddress { // Stores a symbolic cmp address as: FormName, bundleIndex, templateName, cmp index.
            // Retrieves the closest cmp.
            string formName;
            int cloneID; // The Form bundle's index.
            string tplName;
            int index;


            public ControlAddress(VisComponent ctrl) { // Translate the control into a list of AddressSteps.
                if (ctrl == null) return; // May happen with 0 rows.
                formName = ctrl.tpl.FormTpl.Name;
                cloneID = ctrl.tpl.FormTpl.cloneID;
                tplName = ctrl.tpl.Name;
                index = ctrl.Index();
            }

            public  VisComponent GetControl( ) {
                // Translates the symbolic address to the closest cmp or to null if impossible.
                List<FormTemplate> frms = null;
                FormTemplate frmTpl;
                Template tpl;
                if (!VisEngine.Forms.TryGetValue(formName, out frms)) return null;
                if (frms.Count == 0) return null;
                if (frms.Count <= cloneID || frms[cloneID] == null) { // The original clone doesn't exist. Find an open one.
                    cloneID = - 1;
                    for (int j = 0; j < frms.Count; j++) if (frms[j] != null) { cloneID = j; break; }
                    if (cloneID < 0) return null;
                }
                frmTpl = frms[cloneID];
                if (!frmTpl.Templates.TryGetValue(tplName, out tpl)) return null;
                if (tpl.RowBundle == null) return null; 
                int count = tpl.RowBundle.Instances.Count;
                if (count == 0) return null;
                return tpl.RowBundle.Instances[index < count ? index : count-1].cmp;
            }
        }
    }
}
