﻿using System.Linq;
using System.IO;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Data;
using System.ComponentModel;
using System.Reflection;

using System.CodeDom.Compiler;
using Microsoft.CSharp;
using System.Xml.Linq;
using System.Xml.XPath;
using System;
using System.Diagnostics;
using VisTool;
using System.Drawing;
using System.Drawing.Drawing2D;

namespace VisTool {

    /* A Tree component is a kind of position translator. Its Root property specifies the component to start with. From there it traverses the tree of Find components 
     * given by Find-properties that refer to the growing tree. Each component in the tree is given by a Find component property, which has a BundleRef value. 
     * It adds an Anonymous expand/collapse node (TreeNode) to the BundleRef, using its tag property. The component formulas can use for instance Tree1!HPos(targetProperty) and 
     * Tree1!VPos(targetProperty) to get their suggested position in the tree. The HPos and VPos methods ensure that the entire tree is generated (UpdateTree) before the position is returned. 
     * 
     * Tree nodes that belong to a collapsed branch of the tree, exist, but are invisible (not added to the canvas). This means that depending components automatically align themselves 
     *  to an invisible TreeNode. The pattern is for instance used in LifeLine, where ptMedOrder aligns itself to a tree that may be collapsed. PtMedOrder can use 
     *  Tree1!Expanded(targetProperty) to decide whether to show nothing or show all MedOrders on top of each other in this state.
     * 
     * This implementation is 180 lines of code plus 150 lines to specify properties available to the designer. It relies on Find-formulas to generate bundles of visual components.
     * This allows that a component uses several such position translators (e.g. a vertical and a horizontal), each using different Find-formulas in the same component.
     * 
     *  In a previous implementation, the kernel checked for the presence of a layout property formula whenever it computed a position property. The layout formula then computed and 
     *  set the component position. This didn't generalize to situations where a component depended on more than one layout method.
     *  The implementation was around 300 lines of code, but couldn't handle mutiple levels, wasn't sequence independent, and couldn't handle variable component heights.
     *  Further, it couldn't tell components whether the parents were collapsed. Tree properties were limited to a line spacing parameter.
     */

    [VisClassAtt(ToolboxCategory = "Control", Icon = "tree.png", ShortHelp = "Arranges bundles of components as a tree with expand/collapse icons.")]
    public class Tree : VisComponent {
        public Tree(int targetCount, DataRow row, Template tpl)
            : base(targetCount, row, tpl) {
            height = CollapseImage.Height + 10; width = CollapseImage.Width + 10; // Provide reasonable initial defaults.
            ZOrder = -2; // Should be behind ordinary components.
            levelIndent = CollapseImage.Width / 2 + iconSpacing;
            rootRef = new BundleRef(null, 0, null);
        }

        public static Image ExpandImage = Image.FromFile(VisTool.VisEngine.ProgramDirectory + "\\Icons\\" + "treeview_expand.png");
        public static Image CollapseImage = Image.FromFile(VisTool.VisEngine.ProgramDirectory + "\\Icons\\" + "treeview_collapse.png");
        int lastTop, lastBottom, lastLeft; // Pos of the latest component.
        int lastConnectorBottom; // Bottom of the latest connector: lastBottom - lastHeight / 2.
        int maxRight; // Maximum right of the components.
        const int iconSpacing = 10; // Space between icon and component.
        static public int levelIndent; // Icon width + iconSpacing.
        BundleRef rootRef; // A dummy BundleRef that holds the TreeNode of the root.
        bool treeUpdated = false;

        public override void ResetUpdated() { // Called by kernel at refresh to set internal dependent properties to not-updated.
            treeUpdated = false; //nodes.Clear(); cmpToTreeNode.Clear(); 
            base.ResetUpdated();
        }

        public override void ComponentUpdated() { // Maybe nobody has referred to the tree. 
            // Update it before Paint refers to Top, etc. At that point, tree nodes cannot be added to the canvas.
            base.ComponentUpdated();
            UpdateTree(0);
        }

        public static int RootID; // Property index. Set at initialization in ReflectDLL. 
        public VisComponent root = null;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Special", ReadOnly = false, ShowInitially = true,
            ShortHelp = "The root component I start with. The next level matches components that refer to Root with a Find property, and so on.", Position = 1)]
        public VisComponent Root {
            get {
                CmpProp rootProp = this.propertyValues[RootID];
                if (rootProp.state >= State.Updated) return root;
                object temp = EvalProp(RootID, null);
                Root = temp as VisComponent; // Null root is ok. Just show an empty box.
                if (root == null && temp != null) {
                    ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, this.tpl, this.tpl.valPropArray[RootID],
                        "This is not a component", this.tpl.valPropArray[RootID].startChain, this.Index());
                }
                return root;
            }
            set { root = value; SetUpdated(ref propertyValues[RootID]); }
        }

        public static int ExcludeID; // Property index. Set at initialization in ReflectDLL. 
        public BundleRef[] exclude;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Special", ReadOnly = false, ShowInitially = true,
            ShortHelp = "I will ignore these Find properties, e.g. \"lblActivity !Start\" or \"lblActivity !Start, lblActivity !End\".", Position = 2)]
        public BundleRef[] Exclude {
            get {
                CmpProp excludeProp = this.propertyValues[ExcludeID];
                if (excludeProp.state >= State.Updated) return exclude;
                TplProperty targetProp = tpl.valPropArray[ExcludeID];

                object[] temp = EvalListProp(targetProp, excludeProp, new BundleRef[0]);
                exclude = new BundleRef[temp.Length];
                for (int i = 0; i < exclude.Length; i++) {
                    if (!(temp[i] == null || temp[i] is BundleRef)) {
                        ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, VisEngine.curTemplate, this.tpl.valPropArray[ExcludeID],
                            "This is not a Find-property", this.tpl.valPropArray[ExcludeID].startChain, this.Index());
                        exclude[i] = null;
                    } else
                        exclude[i] = temp[i] as BundleRef;
                }
                Exclude = exclude;
                return exclude;
            }
            set { SetUpdated(ref propertyValues[ExcludeID]); }
        }

        public static int VerticalSpacingID; // Property index. Set at initialization in ReflectDLL. 
        int verticalSpacing = 0;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Special", ReadOnly = false, ShowInitially = true,
            ShortHelp = "The vertical spacing between the nodes. Default 0.", Position = 3)]
        public virtual int VerticalSpacing {
            get {
                CmpProp verticalSpacingProp = this.propertyValues[VerticalSpacingID];
                if (verticalSpacingProp.state >= State.Updated) return verticalSpacing;
                VerticalSpacing = toPosInt32(EvalProp(VerticalSpacingID, 0), 0);
                return verticalSpacing;
            }
            set { verticalSpacing = value; SetUpdated(ref propertyValues[VerticalSpacingID]); }
        }

        public static int ExpandInitiallyID; // Property index. Set at initialization in ReflectDLL. 
        bool expandInitially = true;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Special", ReadOnly = false, ShowInitially = true,
            ShortHelp = "Expand the root node initially. Default true.", Position = 4)]
        public virtual bool ExpandInitially {
            get {
                CmpProp expandInitiallyProp = this.propertyValues[ExpandInitiallyID];
                if (expandInitiallyProp.state >= State.Updated) return expandInitially;
                ExpandInitially = toBool(EvalProp(ExpandInitiallyID, true), true);
                return expandInitially;
            }
            set { expandInitially = value; SetUpdated(ref propertyValues[ExpandInitiallyID]); }
        }

        public static int IndentFirstID; // Property index. Set at initialization in ReflectDLL. 
        int indentFirst = 5;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Special", ReadOnly = false, ShowInitially = true,
            ShortHelp = "Indent first level this number of pixels. Default 5.", Position = 5)]
        public virtual int IndentFirst {
            get {
                CmpProp minCellWidthProp = this.propertyValues[IndentFirstID];
                if (minCellWidthProp.state >= State.Updated) return indentFirst;
                IndentFirst = toPosInt32(EvalProp(IndentFirstID, 4), 4);
                return indentFirst;
            }
            set { indentFirst = value; SetUpdated(ref propertyValues[IndentFirstID]); }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Method, null, ShowInitially = false,
            ShortHelp = "(Find_property) Suggested left position for me in the Find property tree. Example: myTree ! HPos(Me !target)")]
        public int HPos(object cmp) {
            UpdateTree(0);
            BundleRef bdlRef = cmp as BundleRef;
            TreeNode node = null;
            if (bdlRef != null) node = bdlRef.tag as TreeNode;
            if (node == null) {
                ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp, "The parameter doesn't refer to the tree.", VisEngine.curExprToken, 0);
                return 10;
            }
            return node.nodeLeft;
        }

        [VisPropAtt(VisPropAtt.PropertyType.Method, null, ShowInitially = false,
            ShortHelp = "(Find_property) Suggested top position for me in the Find property tree. Example: myTree ! VPos(Me !target)")]
        public int VPos(object cmp) {
            UpdateTree(0);
            BundleRef bdlRef = cmp as BundleRef;
            TreeNode node = null;
            if (bdlRef != null) node = bdlRef.tag as TreeNode;
            if (node == null) {
                ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp, "The parameter doesn't refer to the tree.", VisEngine.curExprToken, 0);
                return 10;
            }
            return node.nodeTop;
        }

        [VisPropAtt(VisPropAtt.PropertyType.Method, null, ShowInitially = false,
            ShortHelp = "(Find_property) True if all ancestors of me in the Find_property tree are expanded. If false, I am invisible, but have the same top as my parent.")]
        public bool Expanded(object cmp) {
            UpdateTree(0);
            BundleRef bdlRef = cmp as BundleRef;
            TreeNode node = null;
            if (bdlRef != null) node = bdlRef.tag as TreeNode;
            if (node == null) {
                ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp, "The parameter doesn't refer to the tree.", VisEngine.curExprToken, 0);
                return true;
            }
            return node.ancestorsExpanded;
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, "' Height of the entire tree (Internal formula)",
        PropertyCategory = "Common", ReadOnly = true, ShowInitially = false, ShortHelp = "Height of the entire tree.", Position = 6)]
        public override int Height { get { UpdateTree(HeightID); return height; } }

        [VisPropAtt(VisPropAtt.PropertyType.Property, "' Bottom of the entire tree (Internal formula).", PropertyCategory = "Common", ReadOnly = true, ShowInitially = false,
        ShortHelp = "Bottom of the entire tree.", Position = 7)]
        public override int Bottom { get { UpdateTree(BottomID); return bottom; } }

        [VisPropAtt(VisPropAtt.PropertyType.Property, "' Width of the entire tree (Internal formula).",
        PropertyCategory = "Common", ReadOnly = true, ShowInitially = false,
        ShortHelp = "Width of the entire tree.", Position = 9)]
        public override int Width { get { UpdateTree(WidthID); return width; } }

        [VisPropAtt(VisPropAtt.PropertyType.Property, "' Right of the entire tree (Internal formula)",
            PropertyCategory = "Common", ReadOnly = true, ShowInitially = false,
            ShortHelp = "Right of the entire tree.", Position = 10)]
        public override int Right { get { UpdateTree(RightID); return right; } }

        public override int zOrderDefault { get { return -2; } } // Should be behind ordinary components and the TreeNodes, also for clicking.
        [VisPropAtt(VisPropAtt.PropertyType.Property, null,
            PropertyCategory = "Common", ReadOnly = false, ShowInitially = false, Position = 16,
            ShortHelp = "On my canvas, I am behind components with higher ZOrder. Default -2 to be behind normal tree node components.")]
        public override int ZOrder {
            get { return base.zOrder; }
        }

        public void UpdateTree(int sourceID) { // SourceID == 0 when there is no source property to lock, e.g. VPos.
            if (treeUpdated) return;
            CmpProp sourceProp = this.propertyValues[sourceID];
            if (sourceID != 0)
                sourceProp.state = State.Visited; // Start internal calculation. Catch cyclical references.

            // Root is a component, default the Form.
            object dummy = VerticalSpacing; dummy = ExpandInitially; dummy = IndentFirst;
            lastBottom = this.Top;
            lastLeft = this.Left + levelIndent;
            maxRight = lastLeft;
            VisComponent test = Root as VisComponent;
            if (test == null) { // No root. Create a node attached to the Form.
                TreeNode node = new TreeNode(this.Form());
                rootRef.tag = node;
                node.expanded = true;
                node.top = this.Top; node.Height = CollapseImage.Height + 10; node.bottom = node.top + node.height;
                node.Left = this.Left; node.Width = CollapseImage.Width + 10; node.right = node.left + node.width;
                node.drawExpandCollapse = true;
                node.AddToCanvas(this.Canvas); // Ensure that the node will be painted and clickable.
            } else {
                UpdateSubTree(test, rootRef, true, 0);
                this.Height = lastBottom - this.top;
                this.Bottom = lastBottom;
                this.Width = maxRight - this.left;
                this.Right = maxRight;
            }
            treeUpdated = true;
            if (sourceProp.state == State.Visited) // If it hasn't been computed, reset it to Old.
                sourceProp.state = State.Old;
        }

        void UpdateSubTree(VisComponent cmp, BundleRef bdl, bool ancestorsExpanded, int level) {
            // cmp will be null when there is not Root.
            int prevIndent = lastLeft;
            TreeNode node = bdl.tag as TreeNode;
            if (node == null) {
                node = new TreeNode(cmp);
                bdl.tag = node;
                if (cmp == root && expandInitially) node.expanded = true;
            }
            node.bottom = node.top + node.height; node.right = node.left + node.width;
            node.ancestorsExpanded = ancestorsExpanded;
            node.zOrder = zOrder + 1; // Make sure the expand/collapse icon is clicked, not the Tree.

            if (ancestorsExpanded) { // All ancestors are expanded.
                if (cmp != null) cmp.Visible = true; // Might have been false when collapsed.
                node.AddToCanvas(this.Canvas); // Ensure that the node will be painted and clickable.

                // Compute suggested position of the component.
                node.nodeTop = lastBottom;
                if (cmp == root) node.isRoot = true; else node.nodeTop += this.verticalSpacing;
                lastTop = node.nodeTop;
                node.nodeLeft = lastLeft;

                if (node.nodeLeft + cmp.Width > maxRight) maxRight = node.nodeLeft + cmp.width;

                // Compute position of the connector.
                node.connectorTop = (cmp.Index() == 0 ? lastBottom : lastConnectorBottom);
                node.connectorBottom = lastConnectorBottom = node.nodeTop + cmp.Height / 2;
                lastBottom = node.nodeTop + cmp.height;

                // Compute position of the icon, whether it will be painted or not. Paint will ask for it.
                node.drawExpandCollapse = true;
                node.top = node.connectorBottom - ExpandImage.Height / 2;
                node.height = ExpandImage.Height;
                node.left = node.nodeLeft - levelIndent; // Logically we should subtract CollapseImage.Width/2, 
                // but this makes the icon look unaligned with the parent component (typically a label). Cause: The label  starts with a bit of blank space.
                node.width = ExpandImage.Width;
            } else {
                // Parent collapsed. The node and its children will be located on top of the ancestor, but will not be visible.
                node.nodeTop = lastTop;
                node.nodeLeft = lastLeft;
                cmp.Visible = false;
            }

            Bundle nextBundle;
            if (NextLevel(cmp, level, out nextBundle)) {
                level++;
                if (node.isRoot) lastLeft += indentFirst; else lastLeft += levelIndent;
                foreach (CmpProp cmpProp in nextBundle.Instances)
                    UpdateSubTree(cmpProp.cmp, cmpProp.Value as BundleRef, ancestorsExpanded && node.expanded, level);
                lastConnectorBottom = node.connectorBottom; // Next connector must go to the top of the child bundle.
                lastLeft = prevIndent;
                return;
            }
            if (level > 0) 
                node.drawExpandCollapse = false; return; // No actual children. A terminal node.
        }

        bool NextLevel(VisComponent cmp, int level, out Bundle bundle) { // Returns true if cmp has a next level with at least one component.
            bundle = null; level++;
            BundleRef next;

            TplProperty childTplProp = null; // Becomes a Find property that refers to cmp and isn't excluded.
            int childIndex = 0; // Becomes the index in findProperties and childBundles that refer to the child bundle.
            int notExcludedCount = 0;
            for (int i = 0; i < cmp.tpl.findProperties.Count; i++) { // Look at Find properties referring to cmp.
                TplProperty test = cmp.tpl.findProperties[i];
                bool excluded = false;
                for (int j = 0; j < Exclude.Length; j++) {  // test whether it is in the exclude list.
                    BundleRef bdlRef = exclude[j];
                    if (bdlRef == null || bdlRef.bundle == null) continue;
                    if (bdlRef.bundle.tplProp == test) { // It is excluded.
                        excluded = true; break;
                    }
                }
                if (!excluded) {
                    notExcludedCount++; childTplProp = test; childIndex = i;
                }

            }

            if (notExcludedCount > 1) {
                string findList = "";
                for (int j = 0; j < cmp.tpl.findProperties.Count; j++) { // Look at Find properties referring to cmp.
                    TplProperty test = cmp.tpl.findProperties[childIndex];
                    findList += (j == 0 ? "" : ", ") + test.tpl.Name + "!" + test.Name;
                }
                ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, this.tpl, "Root", "More than one Find bundle to use. Exclude the irrelevant ones: " + findList, null, -1);
                //return false; // Or continue with an arbitrary bundle??
            } 

            if (childTplProp == null) return false; // No valid templates refer to cmp.
            if (!childTplProp.tpl.Updated)
                VisEngine.CreateInstances(childTplProp.tpl);
            List<CmpProp> childBdl = childTplProp.tpl.RowBundle.Instances;
            if (childBdl.Count == 0) return false; // Empty child bundle, no next level. 
            VisComponent child = childBdl[0].cmp; // Compute the bundle for the first child component. 
            bool exists;
            next = VisEngine.GetValue(child, childTplProp.Name, out exists) as BundleRef; // Will compute all child bundles.
            if (VisEngine.circularRef != null) {
                VisEngine.circularLevel++;
                ErrorLogging.Log(ErrorMode.RunTime, Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp, "Circular reference with Tree, level " + VisEngine.circularLevel, childTplProp.startChain, cmp.Index());
                return false;
            }

            if (cmp.childBundles == null) return false; // Has no child bundles.
            bundle = cmp.childBundles[childIndex]; // Find the child bundle of cmp.
            if (bundle == null) return false; // No child component refers to cmp.
            return bundle.Instances.Count > 0; // Refers, but the bundle may be empty. Possible?? 
        }
    }

    /* TreeNode draws a connector to the parent or previous tree node on the same level. The connector consists of a vertical line and a horizontal line.
     * For the root node, there is no previous tree node and the vertical line is absent. When the node has a sub-tree, TreeNode also draws an expand/collapse icon.
     * The Top, Height, etc. of the Tree node, is the clickable part - the icon.
     */
        public class TreeNode : VisComponent {
            public int nodeTop, nodeLeft; // Proposed position for the node component.
            public int connectorTop, connectorBottom; // For drawing the vertical line.
            // ConnectorTop is the bottom of the previous node or the middle of it. ConnectorBottom is the middle of this node.
            public bool expanded = false; // True when user has clicked expanded. Survives collapse of ancestor because the components still exist.
            public bool ancestorsExpanded = false; // True if all ancestors are expanded.
            public bool isRoot = false;
            public bool drawExpandCollapse = false;
            public bool drawCollapse = false;
            public int lineDistance;

            public TreeNode(VisComponent cmp) {
                // Borrows template and properties from cmp.
                this.Anonymous = true; // TreeNode is not available to designers, but clickable.
                this.tpl = cmp.tpl;
                propertyValues = cmp.propertyValues; // Share properties.
            }

            public override void Click(object sender, EventArgs e) {
                if (drawExpandCollapse) {
                    expanded = !expanded; 
                    VisEngine.FormRefresh(this.tpl.FormTpl); // Recomputes all properties. Keeps the tag values for components that aren't deleted during refresh.
                }
            }

            public override void Paint(PaintEventArgs e) {
                // When an ancestor is collapsed, the TreeNode is not on the canvas and will not be painted.
                e.Graphics.SmoothingMode = SmoothingMode.HighSpeed;
                Pen pen = new Pen(Brushes.Gray);
                pen.DashStyle = DashStyle.Dot;
                pen.Width = 2;
                PointF[] Points;
                int connectorLeft = this.left + this.width / 2; 
                PointF point1 = new PointF(RelativeX(this.nodeLeft), RelativeY(this.connectorBottom));
                PointF point2 = new PointF(RelativeX(connectorLeft), RelativeY(this.connectorBottom));
                if (isRoot) {
                    Points = new PointF[] { point1, point2 };
                } else {
                    PointF point3 = new PointF(RelativeX(connectorLeft), RelativeY(this.connectorTop));
                    Points = new PointF[] { point1, point2, point3 };
                }
                e.Graphics.DrawLines(pen, Points);
                if (drawExpandCollapse) {
                    e.Graphics.DrawImage(expanded ? Tree.CollapseImage : Tree.ExpandImage, new Point(RelativeX(this.left), RelativeY(this.top)));
                }
            }
        }
}