﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Linq;
using System.Runtime.InteropServices;
using System.Windows.Forms;
using System.Windows.Input;
using System.Windows.Interop;
using VisTool;
using uVisStudio.VisDesigner;
using uVisStudio.VisToolbox;
using Application = System.Windows.Application;
using Point = System.Drawing.Point;

namespace uVisStudio {

    /* All Hook events are caught by IntPtr HookMouse, HookKeyBoard, etc. They get the raw events from Windows.
     * Hook.GetClickedInstances gets a list of the components that are below the cursor.
     */
    class Hook {
        private enum MouseMessages {
            WM_LBUTTONDOWN = 0x0201,
            WM_LBUTTONUP = 0x0202,
            WM_LBUTTONDBLCLK = 0x0203,
            WM_MOUSEMOVE = 0x0200,
            WM_MOUSEWHEEL = 0x020A,
            WM_RBUTTONDOWN = 0x0204,
            WM_RBUTTONUP = 0x0205
        }
        // Only some of these are used:
        private const int WH_MOUSE = 7;
        private const int WH_MOUSE_LL = 14;
        private const int WH_KEYBOARD_LL = 13;
        private const int WH_KEYBOARD = 2;
        private const int WM_KEYDOWN = 0x0100;
        private const int WM_KEYUP = 0x0101;
        private const int WM_SYSKEYDOWN = 0x0104;
        private const int WM_SYSKEYUP = 0x0105;
        private const int WM_SYSCOMMAND = 0x0112;
        private const uint GW_HWNDNEXT = 2;

        private static LowLevelMouseProc _procMouse = HookMouse;
        private static IntPtr hookMouseID = IntPtr.Zero;
        private static LowLevelKeyboardProc _procKeyboard = HookKeyboard;
        private static IntPtr hookKeyboardID = IntPtr.Zero;
        private delegate IntPtr LowLevelMouseProc(int nCode, IntPtr wParam, IntPtr lParam);
        private delegate IntPtr LowLevelKeyboardProc(int nCode, IntPtr wParam, IntPtr lParam);

        public static Point StartPoint; // Screen point for the latest mouse down event.
        public static Form currentForm; // The latest form clicked or passed with the mouse.

        public static bool Handled; // When true, the hook has processed the event. Other threads shouldn't handle it. 
        // Default true when a HookState method is called. When handled is true, the hook returns IntPtr(-1), otherwise the next thread is called.
        public static Control mouseCapture = null; // The gui control that has captured the mouse during dragging.

        public static HookState State = new HookNeutralState();

        public static void StartHook() { // Called from Studio Window constructor.
            hookMouseID = SetHook(_procMouse);
            hookKeyboardID = SetHook(_procKeyboard);
        }

        // Hook timer removed. Wasn't used. SL 28-02-16.

        private static IntPtr SetHook(LowLevelKeyboardProc proc) {
            using (Process curProcess = Process.GetCurrentProcess())
            using (ProcessModule curModule = curProcess.MainModule) {
                return SetWindowsHookEx(WH_KEYBOARD_LL, proc, GetModuleHandle(curModule.ModuleName), 0);
            }
        }

        private static IntPtr SetHook(LowLevelMouseProc proc) {
            using (Process curProcess = Process.GetCurrentProcess())
            using (ProcessModule curModule = curProcess.MainModule) {
                return SetWindowsHookEx(WH_MOUSE, proc, IntPtr.Zero, GetCurrentThreadId());
            }
        }

        private static IntPtr HookKeyboard(int nCode, IntPtr wParam, IntPtr lParam) {
            Handled = true; // Default
            if (nCode >= 0 && wParam == (IntPtr)WM_KEYDOWN) KeyDown(nCode, wParam, lParam);
            else if (nCode >= 0 && wParam == (IntPtr)WM_KEYUP) KeyUp(nCode, wParam, lParam);
            else Handled = false;
            if (Handled) return new IntPtr(-1);
            return CallNextHookEx(hookMouseID, nCode, wParam, lParam);
        }

        private static IntPtr HookMouse(int nCode, IntPtr wParam, IntPtr lParam) { // Activated when mouse moves across visual parts of Studio and VisForms.
            Handled = true; // Default
            var message = (MouseMessages)wParam;
            if (nCode >= 0 && (MouseMessages.WM_LBUTTONDOWN == message)) { //  || MouseMessages.WM_RBUTTONDOWN == message
                var hookStruct = (MSLLHOOKSTRUCT)Marshal.PtrToStructure(lParam, typeof(MSLLHOOKSTRUCT));
                if (mouseCapture != null) { // Make sure to release mouse also at button down, e.g. when dragging a form.
                    mouseCapture.Capture = false; mouseCapture = null;
                }
                StartPoint = new Point(hookStruct.pt.x, hookStruct.pt.y);
                State.MouseLeftButtonDown(StartPoint);
            } else if (nCode >= 0 && MouseMessages.WM_MOUSEMOVE == message) {
                var hookStruct = (MSLLHOOKSTRUCT)Marshal.PtrToStructure(lParam, typeof(MSLLHOOKSTRUCT));
                var point = new Point(hookStruct.pt.x, hookStruct.pt.y);
                State.MouseMove(point);
            } else if (nCode >= 0 && (MouseMessages.WM_LBUTTONUP == message)) { // || MouseMessages.WM_RBUTTONUP == message
                var hookStruct = (MSLLHOOKSTRUCT)Marshal.PtrToStructure(lParam, typeof(MSLLHOOKSTRUCT));
                if (mouseCapture != null) { // Make sure to release mouse whenever button goes up.
                    mouseCapture.Capture = false; mouseCapture = null;
                }
                var endPoint = new Point(hookStruct.pt.x, hookStruct.pt.y);
                State.MouseLeftButtonUp(endPoint);
            } else Handled = false;
            if (Handled) return new IntPtr(-1);
            return CallNextHookEx(hookMouseID, nCode, wParam, lParam);
        }

        private static IntPtr KeyUp(int nCode, IntPtr wParam, IntPtr lParam) {
            Handled = true;
            int vkCode = Marshal.ReadInt32(lParam);
            Keys key = (Keys)vkCode;
            if (Keys.Z == key && Keyboard.Modifiers == (ModifierKeys.Control) && !StudioBase.formulaEditorFocus) {
                // Also handled by formula editor and StudioWindow.xaml.cs. SL 01-04-13
                UndoManager.Undo();
            } else if (Keys.Y == key && Keyboard.Modifiers == (ModifierKeys.Control) && !StudioBase.formulaEditorFocus) {
                UndoManager.Redo();
            } else if (Keys.Escape == key) {
                System.Windows.Forms.Cursor.Current = System.Windows.Forms.Cursors.Default;
                State = new HookNeutralState();
                VisToolboxItem.Deselect();
                SelectionBox.ResetAll(); // Move blue boxes back to the component.
                if (VisToolboxItem.IsOneSelected) VisToolboxItem.Deselect();
                Handled = false; // Allow Studio to use Esc too.
            } else
                State.KeyUp(key);
            if (Handled) return new IntPtr(-1);
            return CallNextHookEx(hookMouseID, nCode, wParam, lParam);
        }

        private static IntPtr KeyDown(int nCode, IntPtr wParam, IntPtr lParam) {
            Handled = true;
            int vkCode = Marshal.ReadInt32(lParam);
            Keys key = (Keys)vkCode;
            State.KeyDown(key);
            if (Handled) return new IntPtr(-1);
            return CallNextHookEx(hookMouseID, nCode, wParam, lParam);
        }

        public static bool IsKeyPressed(Keys key) {
            return KeyStates.Down == (GetKeyState(key) & KeyStates.Down);
        }

        public static Form GetClickedForm(Point point) { // Returns the Form of a VisForm or null.
            Form clickedForm = null;
            foreach (var form in Hook.GetForms()) {
                var clientPoint = form.PointToClient(point);
                var succes = form.ClientRectangle.Contains(clientPoint);
                if (!succes) continue;
                clickedForm = form;
            }
            return clickedForm;
        }

        private static List<Form> GetForms() {
            var forms = new List<Form>();
            foreach (var kv in VisEngine.Forms) {
                foreach (FormTemplate frmTpl in kv.Value)
                    if (frmTpl != null && frmTpl.open && !frmTpl.updating && frmTpl.RowBundle != null) // Mouse may be used during Refresh
                        for (int i = 0; i < frmTpl.RowBundle.Instances.Count; i++) {
                            VisComponent formInstance = (VisForm)frmTpl.RowBundle.Instances[i].cmp;
                            if (formInstance != null) forms.Add((Form)formInstance.GuiInstance);
                        }
            }
            forms = Hook.SortByZIndex(forms);
            return forms;
        }

        public static List<VisComponent> GetClickedInstances(Form form, Point p) {
            var instances = new List<VisComponent>();
            var clickLocation = new Point(p.X, p.Y);
            Hook.AddClickedInstances(form.Tag as ContainerComponent, ref instances, clickLocation);
            return instances;
        }

        private static void AddClickedInstances(ContainerComponent parent, ref List<VisComponent> instances, Point clickLocation) {
            if (!parent.Anonymous) instances.Add(parent);
            for (int i = 0; i < parent.visComponents.Length; i++) { // Add components in increasing zOrder.
                VisComponent ci = parent.visComponents[i];
                if (Contains(clickLocation, ci) && !instances.Contains(ci))
                    if (ci is ContainerComponent)
                        AddClickedInstances(ci as ContainerComponent, ref instances, clickLocation);
                    else {
                        if (!ci.Anonymous) instances.Add(ci);
                    }
            }
        }

        private static bool Contains(Point point, VisComponent cmp) {
            int Htrim = cmp.width > 10 ? 0 : 5 - cmp.width / 2; // Make sure that large components such as canvas don't steal neigbor areas.
            int Vtrim = cmp.height > 10 ? 0 : 5 - cmp.height / 2;
            // Notice that left+width is just outside the client area.
            Point cmpOnScreen;
            if (cmp.canvas != null)
                cmpOnScreen = cmp.canvas.GuiInstance.PointToScreen(new Point(cmp.RelativeX(cmp.left), cmp.RelativeY(cmp.top)));
            else
                cmpOnScreen = cmp.GuiInstance.PointToScreen(new Point(cmp.RelativeX(cmp.left), cmp.RelativeY(cmp.top)));
            return ((point.X >= cmpOnScreen.X - Htrim && point.X < cmpOnScreen.X + VisComponent.PhyS(cmp.width) + Htrim) && (point.Y < cmpOnScreen.Y + VisComponent.PhyS(cmp.height) + Vtrim && point.Y >= cmpOnScreen.Y - Vtrim));
        }

        private static List<Form> SortByZIndex(List<Form> unsortedForms) {
            var forms = new List<Form>();
            var handles = unsortedForms.ToDictionary(form => (form.Handle));

            var hWnd = GetTopWindow(IntPtr.Zero);
            while (hWnd != IntPtr.Zero) {
                hWnd = GetWindow(hWnd, GW_HWNDNEXT);
                if (handles.ContainsKey(hWnd)) {
                    var form = handles[hWnd];
                    forms.Add(form);
                }
            }

            forms.Reverse();
            return forms;
        }

        private static KeyStates GetKeyState(Keys key) {
            var state = KeyStates.None;
            var retVal = GetKeyState((int)key);
            if ((retVal & 0x8000) == 0x8000)
                state |= KeyStates.Down;

            if ((retVal & 1) == 1)
                state |= KeyStates.Toggled;

            return state;
        }

        [StructLayout(LayoutKind.Sequential)]
        private struct POINT {
            public int x;
            public int y;
        }

        [StructLayout(LayoutKind.Sequential)]
        private struct MSLLHOOKSTRUCT {
            public POINT pt;
            public uint mouseData;
            public uint flags;
            public uint time;
            public IntPtr dwExtraInfo;
        }

        [Flags]
        private enum KeyStates {
            None = 0,
            Down = 1,
            Toggled = 2
        }

        [DllImport("user32.dll", SetLastError = true)]
        static extern IntPtr GetWindow(IntPtr hWnd, uint uCmd);

        [DllImport("user32.dll")]
        static extern IntPtr GetTopWindow(IntPtr hWnd);

        [DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        private static extern IntPtr SetWindowsHookEx(int idHook, LowLevelMouseProc lpfn, IntPtr hMod, uint dwThreadId);

        [DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        private static extern IntPtr SetWindowsHookEx(int idHook, LowLevelKeyboardProc lpfn, IntPtr hMod, uint dwThreadId);

        [DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        [return: MarshalAs(UnmanagedType.Bool)]
        private static extern bool UnhookWindowsHookEx(IntPtr hhk);

        [DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        private static extern IntPtr CallNextHookEx(IntPtr hhk, int nCode, IntPtr wParam, IntPtr lParam);

        [DllImport("kernel32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        private static extern IntPtr GetModuleHandle(string lpModuleName);

        [DllImport("kernel32")]
        private static extern uint GetCurrentThreadId();

        [DllImport("user32.dll", CharSet = CharSet.Auto, ExactSpelling = true)]
        private static extern short GetKeyState(int keyCode);
    }
}
