﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Reflection;
using System.IO;
using System.Windows.Forms;
using System.Diagnostics;
using VisTool;
using uVisStudio.VisDesigner;
using uVisStudio.VisExplorer;

namespace uVisStudio {
    public class StudioException : Exception {
        public string Info = string.Empty;
        public StudioException(string info) {
            this.Info = info;
        }
    }

    public interface ICommand {
        bool Execute();
        bool Undo();
    }

    public class UndoManager {
        // See Controllers (class Controller) for an explanation of the Undo mechanisms.
        public static List<ICommand> Commands = new List<ICommand>();
        public static List<ICommand> PoppedCommands = new List<ICommand>();

        public static void AddCommand(ICommand command) {
            PoppedCommands.Clear();
            Commands.Add(command);
            //Debug.Print("AddCommand " + command.GetType().ToString() + " new count: " + Commands.Count);
        }

        public static bool Undo() {
            if (Commands.Count > 0) {
                var command = Commands[Commands.Count - 1];
                if (command.TryUndo()) {
                    PoppedCommands.Add(command);
                    Commands.Remove(command);
                    SelectionBox.DeleteAll();
                    //Debug.Print("Undo " + command.GetType().ToString() + " new count: " + Commands.Count);
                };
            }
            return true;
        }

        public static void Redo() {
            if (PoppedCommands.Count > 0) {
                var command = PoppedCommands[PoppedCommands.Count - 1];
                if (command.TryExecute()) {
                    Commands.Add(command);
                    PoppedCommands.Remove(command);
                    //Debug.Print("Redo " + command.GetType().ToString() + " new count: " + Commands.Count);
                }
            }
        }

        public static void Refresh(Template template) {
            var formTemplate = template.FormTpl;
            if (!(bool)StudioBase.Studio.refreshCB.IsChecked)
                VisEngine.CompileOnly = true;
            VisEngine.FormRefresh(formTemplate); 
            VisEngine.CompileOnly = false;
        }
    }

    #region commands

    //-----------------------------------------------------------------------------------------------
    public class TemplateDrawnCommand : ICommand { // Uses default value for width when width <4. Height similar.
        // Used also when a template is dropped, but then height and width are zero.
        private ComponentInfo componentInfo;
        private Template tpl;
        private ContainerComponent container;
        private Point relativePoint; // For a Form, it is the absolute screen position.
        private int width;
        private int height;
        string newFrmName;

        public TemplateDrawnCommand(ComponentInfo componentInfo, ContainerComponent container, Point relativePoint, int width, int height) {
            // Container may be null for a Form.
            this.componentInfo = componentInfo;
            this.container = container;
            this.relativePoint = relativePoint;
            this.width = width;
            this.height = height;
        }

        public bool Execute() {
            if (componentInfo.Type == typeof(VisForm)) {
                  FileInfo frmFile;
                Directory.SetCurrentDirectory(VisEngine.ApplicationDirectory);
                int number = 1;
                while (true) {
                    newFrmName = "frm" + number;
                    try {
                        frmFile = new FileInfo(newFrmName + ".vis");
                        if (!frmFile.Exists) {
                            StreamWriter file = frmFile.AppendText();
                            file.Write("Text: Name\r\nWidth: "+ (this.width == 0 ? 400 : width > 12 ? width : 12));
                            file.Write("\r\nHeight: " + (this.height == 0 ? 300 : height > 30 ? height : 30) + "\r\n");
                            file.Close();
                            break;
                        }
                    }
                    catch (Exception x) {
                        MessageBox.Show("Cannot create file \"" + newFrmName + ".vis" + " in the application directory: " + x.Message);
                        return false;
                    }
                    number++;
                }
                StudioBase.Explorer.UpdateExplorer();
                int dummy;
                this.tpl = VisEngine.OpenForm(newFrmName, null, out dummy);
                Form frm = this.tpl.RowBundle.Instances[0].cmp.GuiInstance as Form;
                //frm.Top = relativePoint.Y; frm.Left = relativePoint.X;
                //frm.Show();
            } else {
                //template = template.LookUp(); if (template == null) return false; // Try later
                FormTemplate formTemplate = container.tpl.FormTpl;
                var name = HookState.CreateTemplateName(formTemplate, componentInfo.Type.Name);
                this.tpl = VisEngine.CreateTemplate(name, componentInfo, formTemplate);
                if (this.tpl == null) return false; // Because the form is updating or there are multiple clones.

                foreach (var value in componentInfo.PropertyByName.Values) {
                    // Create properties with a default formula.
                    var attribute = value.VisAttribute;
                    if (attribute == null) continue;
                    //if (!attribute.ShowInitially) continue; // Set value even when the formula isn't set.
                    //if (value.Name.ToUpper() == "NAME") continue;

                    if (attribute.DefaultFormula == null || attribute.ReadOnly) continue;
                    var formula = attribute.DefaultFormula.ToString();
                    if (String.IsNullOrEmpty(formula)) continue;
                    VisEngine.CreateOrSetTemplateProperty(tpl, value.Name, formula);
                }

                // Set properties defined by drawing/dropping. Compensate Top and Left for scrolling.
                // Tested for Form updating, etc. above in CreateTemplate.
                ScrollableControl panel = container.GuiInstance as ScrollableControl;
                VisComponent.SetTopFormula(componentInfo, tpl, tpl, VisComponent.LogS(relativePoint.Y + panel.VerticalScroll.Value), 0);
                VisComponent.SetLeftFormula(componentInfo, tpl, tpl, VisComponent.LogS(relativePoint.X + panel.HorizontalScroll.Value), 0);
                if (height > 3) VisComponent.SetHeightFormula(componentInfo, tpl, tpl, VisComponent.LogS(height), 0);
                if (width > 3) VisComponent.SetWidthFormula(componentInfo, tpl, tpl, VisComponent.LogS(width), 0);

                string canvasFormula = container.GetCanvasFormula();
                if (canvasFormula != null)
                    VisEngine.CreateOrSetTemplateProperty(tpl, "Canvas", canvasFormula);

                UndoManager.Refresh(tpl);
                SelectionBox.DeleteAll();
                VisComponent cmp = tpl.RowBundle.Instances[0].cmp;
                SelectionBox.selected = new SelectionBox(cmp);
                TemplateHandler.SelectTemplate(cmp); // Triggers Studio to update Property grids.

                //TemplateHandler.SelectTemplate(tpl);
                //VisPropertyBox.PropertyBoxManager.UpdateAllGrids(true);
                //StudioBase.Explorer.UpdateExplorer();
            }
            return true;
        }

        public bool Undo() {
            if (tpl is FormTemplate) { // The template is a Form. Close it and delete the file.
                SystemItems.CloseForm(tpl.RowBundle.Instances[0].cmp);
                try {
                    Directory.SetCurrentDirectory(VisEngine.ApplicationDirectory);
                    FileInfo frmFile = new FileInfo(newFrmName + ".vis");
                    frmFile.Delete();
                    StudioBase.Explorer.UpdateExplorer(); 
                    return true;
                }
                catch (Exception ex) {
                    MessageBox.Show("Cannot delete file " + tpl.Name + ".vis\n" + ex.Message);
                }
                return false;
            }

            tpl = tpl.LookUp(); if (tpl == null) return false;
            if (VisEngine.DeleteTemplate(tpl)) {
                UndoManager.Refresh(tpl);
                VisPropertyBox.PropertyBoxManager.UpdateAllGrids(true);
                StudioBase.Explorer.UpdateExplorer();
                return true;
            }
            return false;
        }
    }

    public class DeleteTemplateCommand : ICommand {
        private List<Template> templates;

        public DeleteTemplateCommand(List<Template> templates) {
            this.templates = templates;
        }

        public bool Execute() {
            foreach (var template in templates) {
                var updatedTemplate = template.LookUp();
                if (updatedTemplate == null) return false;
                if (!VisEngine.DeleteTemplate(updatedTemplate)) return false;
            }

            var tpl = templates[0];
            UndoManager.Refresh(tpl);
            VisPropertyBox.PropertyBoxManager.UpdateAllGrids(true);
            StudioBase.Explorer.UpdateExplorer();
            return true;
        }

        public bool Undo() {
            FormTemplate frmTpl = null;
            //Template template = null;
            foreach (var tpl in templates) {
                //template = tpl.LookUp(); if (template == null) return false; // Try later
                frmTpl = tpl.FormTpl; // Trust that the references are still valid.
                if (tpl is FormTemplate) { // We cannot undo form delete. Open the vis-file instead.
                    MessageBox.Show("Cannot undo deletion of a Form. The vis-file has been deleted.");
                    return true; // Allow other commands to be undone.
                }
                if (VisEngine.CreateTemplate(tpl, frmTpl) == null) return false; // Maybe multiple clones.
                //var parentTemplate = template.ParentTemplate.LookUp();
                //if (parentTemplate == null) continue;
                //parentTemplate.ChildTemplates.Add(template);
            }
            if (frmTpl != null) UndoManager.Refresh(frmTpl);
            VisPropertyBox.PropertyBoxManager.UpdateAllGrids(true);
            StudioBase.Explorer.UpdateExplorer();
            return true;
        }
    }

    public class TemplatesCopiedCommand : ICommand {
        private List<MoveTransaction> oldTransactions;
        private List<MoveTransaction> newTransactions = new List<MoveTransaction>(0); 
        public TemplatesCopiedCommand(List<MoveTransaction> transactions) {
            this.oldTransactions = transactions;
            for (int i = 0; i < transactions.Count; i++) { // Prepare the copies.
                MoveTransaction old = transactions[i];
                newTransactions.Add(new MoveTransaction(old.Instance, old.newFrmTpl));
            }
        }

        public bool Execute() {
            Template template = null;
            FormTemplate newFrmTpl = null; // All transactions have the same newFrmTpl;

            bool fullyDone = true; // Assumes everything can be changed.
            for (int i = 0; i < oldTransactions.Count; i++) {
                MoveTransaction old = oldTransactions[i];
                template = old.tpl.LookUp(); // Checks that the source form isn't closed, etc. 
                if (template == null) return false;
                if (!VisEngine.LockNoClones(template)) return false; // cannot edit multi-clone form templates or forms being updated.
                VisEngine.UnLockForm(template);
                newFrmTpl = old.newFrmTpl;
                Template newTpl = VisEngine.CopyTemplate(template, newFrmTpl);
                if (newTpl == null)
                    return false;
                newTransactions[i].tpl = newTpl;

                // Borrow compiled expressions in order to generate new formulas.:
                newTpl.propLeft.startChain = template.propLeft.startChain; 
                newTpl.propTop.startChain = template.propTop.startChain; 
                fullyDone = fullyDone & VisComponent.SetLeftFormula(newTpl.Info, newTpl, template, old.NewLeft, old.Instance.left);
                fullyDone = fullyDone & VisComponent.SetTopFormula(newTpl.Info, newTpl, template, old.NewTop, old.Instance.top);
                VisEngine.CreateOrSetTemplateProperty(newTpl, "Canvas", old.CanvasFormula);
            }

            UndoManager.Refresh(newFrmTpl);
            SelectionBox.DeleteAll();
            for (int j = 0; j < newTransactions.Count; j++) {
                Template tpl = newTransactions[j].tpl;
                if (tpl.RowBundle.Instances.Count > 0) {
                        VisComponent cmp = tpl.RowBundle.Instances[0].cmp;
                        SelectionBox.selected = new SelectionBox(cmp);
                        TemplateHandler.SelectTemplate(cmp); // Triggers Studio to update Property grids.
                        break;
                }
            }
            //VisPropertyBox.PropertyBoxManager.UpdateAllGrids(true);
            // StudioBase.Explorer.UpdateExplorer();
            if (!fullyDone)
                System.Windows.Forms.MessageBox.Show("Couldn't change one or more formulas to reflect the new position", "uVis");
            return true;
        }

        public bool Undo() {
            Template template = null;
            foreach (var copy in newTransactions) {
                template = copy.tpl.LookUp();
                if (template == null) return false;
                if (!VisEngine.LockNoClones(template)) return false; // cannot edit multi-clone form templates or forms being updated.
                VisEngine.UnLockForm(template);
                VisEngine.DeleteTemplate(template);
            }

            UndoManager.Refresh(template);
            //VisPropertyBox.PropertyBoxManager.UpdateAllGrids(false);
            // StudioBase.Explorer.UpdateExplorer();
            return true;
        }
    }

    public class TemplatesMovedCommand : ICommand {
        private List<MoveTransaction> transactions;

        public TemplatesMovedCommand(List<MoveTransaction> transactions) {
            this.transactions = transactions;
        }

        public bool Execute() {
            Template template = null;
            bool fullyDone = true; // Assumes everything can be changed.
            foreach (var move in transactions) {
                template = move.tpl.LookUp();
                if (template == null) return false;
                if (!VisEngine.LockNoClones(template)) return false; // cannot edit multi-clone form templates or forms being updated.
                VisEngine.UnLockForm(template);
                if (template.propLeft.expr != null || template.propRight.expr == null && template.propWidth.expr == null)  
                    fullyDone = fullyDone & VisComponent.SetLeftFormula(template.Info, template, move.tpl, move.NewLeft, move.Instance.left);
                if (template.propRight.expr != null) // May adjust Left as well as Right
                    fullyDone = fullyDone & VisComponent.SetRightFormula(template.Info, template, move.tpl, move.NewLeft, move.Instance.left);

                if (template.propTop.expr != null || template.propBottom.expr == null && template.propHeight.expr == null)
                    fullyDone = fullyDone & VisComponent.SetTopFormula(template.Info, template, move.tpl, move.NewTop, move.Instance.top);
                if (template.propBottom.expr != null) // May adjust Top as well as Bottom
                    fullyDone = fullyDone & VisComponent.SetBottomFormula(template.Info, template, move.tpl, move.NewTop, move.Instance.top);

                VisEngine.CreateOrSetTemplateProperty(template, "Canvas", move.CanvasFormula);
                if (template.Info.Type == typeof(VisForm))
                    UndoManager.Refresh(template);
            }
            if (template.Info.Type != typeof(VisForm))
                UndoManager.Refresh(template);

            VisPropertyBox.PropertyBoxManager.UpdateAllGrids(false);
            if (!fullyDone)
                System.Windows.Forms.MessageBox.Show("Couldn't change one or more formulas to reflect the move", "uVis");
            return true;
        }

        public bool Undo() {
            Template template = null;
            foreach (var move in transactions) {
                template = move.tpl.LookUp();
                if (template == null) return false;
                if (!VisEngine.LockNoClones(template)) return false; // cannot edit multi-clone form templates or forms being updated.
                VisEngine.UnLockForm(template);

                for (int i = 0; i < move.oldFormulas.Length; i++) {
                    string oldValProp = move.oldFormulas[i];
                    if (template.valPropArray[i].Formula != oldValProp) // The formula has changed.
                        VisEngine.CreateOrSetTemplateProperty(template, template.valPropArray[i].Name, oldValProp);
                }
                if (template.Info.Type == typeof(VisForm))
                    UndoManager.Refresh(template);
            }
            if (template.Info.Type != typeof(VisForm))
                UndoManager.Refresh(template);

            VisPropertyBox.PropertyBoxManager.UpdateAllGrids(false);
            // StudioBase.Explorer.UpdateExplorer();
            return true;
        }
    }

    public class TemplatesResizedCommand : ICommand {
        private List<ResizeTransaction> transactions; // All components must be on the same canvas.
        // When the components are forms, they are on the same "canvas", i.e. the screen.

        public TemplatesResizedCommand(List<ResizeTransaction> transactions) {
            this.transactions = transactions;
        }

        public bool Execute() {
            bool fullyDone = true; // Assumes everything can be changed.
            Template tpl = null;
            foreach (var resize in transactions) {
                tpl = resize.tpl.LookUp();
                if (tpl == null) return false;
                if (!VisEngine.LockNoClones(tpl)) return false; // cannot edit multi-clone form templates or forms being updated.
                VisEngine.UnLockForm(tpl);

                if (resize.BeforeLeft != resize.NewLeft) { // Left moved: May change Width and maybe Left, depending on where the formulas are.
                    if (tpl.propLeft.expr != null || (tpl.propWidth.expr == null && tpl.propRight.expr == null)) // Set Left when Left expr or no expr.
                        fullyDone = fullyDone & VisComponent.SetLeftFormula(tpl.Info, tpl, tpl, resize.NewLeft, resize.Instance.Left);
                    if (!(tpl.propLeft.expr != null && tpl.propRight.expr != null)) // Set Width unless Left and Right defined.
                        fullyDone = fullyDone & VisComponent.SetWidthFormula(tpl.Info, tpl, tpl, resize.NewWidth, resize.Instance.Width);
                } else if (resize.BeforeWidth != resize.NewWidth) { // Width changed, Left kept: May change Width and maybe Right.
                    if (tpl.propRight.expr != null || (tpl.propWidth.expr == null && tpl.propLeft.expr == null)) // Set Right when Right expr or no expr.
                        fullyDone = fullyDone & VisComponent.SetRightFormula(tpl.Info, tpl, tpl, resize.NewLeft + resize.NewWidth, resize.Instance.Left + resize.Instance.Width);
                    if (!(tpl.propLeft.expr != null && tpl.propRight.expr != null)) // Set Width unless Left and Right defined.
                        fullyDone = fullyDone & VisComponent.SetWidthFormula(tpl.Info, tpl, tpl, resize.NewWidth, resize.Instance.Width);
                }

                if (resize.BeforeTop != resize.NewTop) { // Top moved: May change Height and maybe Top, depending on where the formulas are.
                    if (tpl.propTop.expr != null || (tpl.propHeight.expr == null && tpl.propBottom.expr == null)) // Set Top when Top expr or no expr.
                        fullyDone = fullyDone & VisComponent.SetTopFormula(tpl.Info, tpl, tpl, resize.NewTop, resize.Instance.top);
                    if (!(tpl.propTop.expr != null && tpl.propBottom.expr != null)) // Set Height unless Top and Bottom defined.
                        fullyDone = fullyDone & VisComponent.SetHeightFormula(tpl.Info, tpl, tpl, resize.NewHeight, resize.Instance.height);
                } else if (resize.BeforeHeight != resize.NewHeight) { // Height changed, Top kept: May change Height and maybe Bottom.
                    if (tpl.propBottom.expr != null || (tpl.propHeight.expr == null && tpl.propTop.expr == null)) // Set Bottom when Bottom expr or no expr.
                        fullyDone = fullyDone & VisComponent.SetBottomFormula(tpl.Info, tpl, tpl, resize.NewTop + resize.NewHeight, resize.Instance.top + resize.Instance.height);
                    if (!(tpl.propTop.expr != null && tpl.propBottom.expr != null)) // Set Height unless Top and Bottom defined.
                        fullyDone = fullyDone & VisComponent.SetHeightFormula(tpl.Info, tpl, tpl, resize.NewHeight, resize.Instance.height);
                }

                if (tpl.Info.Type == typeof(VisForm))
                    UndoManager.Refresh(tpl);
            }
            if (tpl.Info.Type != typeof(VisForm))
                UndoManager.Refresh(tpl);

            VisPropertyBox.PropertyBoxManager.UpdateAllGrids(false);
            if (!fullyDone)
                System.Windows.Forms.MessageBox.Show("Couldn't change one or more formulas to reflect the new size", "uVis");
            return true;
        }

        public bool Undo() {
            Template template = null;
            foreach (var move in transactions) {
                template = move.tpl.LookUp();
                if (template == null) continue;
                if (!VisEngine.LockNoClones(template)) continue; // cannot edit multi-clone form templates or forms being updated.
                VisEngine.UnLockForm(template);
                for (int i = 0; i < move.oldFormulas.Length; i++) {
                    string oldValProp = move.oldFormulas[i];
                    if (template.valPropArray[i].Formula != oldValProp) // The formula has changed.
                        VisEngine.CreateOrSetTemplateProperty(template, template.valPropArray[i].Name, oldValProp);
                }
                if (template.Info.Type == typeof(VisForm))
                    UndoManager.Refresh(template);
            }
            if (template.Info.Type != typeof(VisForm))
                UndoManager.Refresh(template);

            VisPropertyBox.PropertyBoxManager.UpdateAllGrids(false);
            // StudioBase.Explorer.UpdateExplorer();
            return true;
        }
    }

    public class RenameFileCommand : ICommand {
        private string newLocation;
        private string oldLocation;

        public RenameFileCommand(string file, string newNameWithExtension) {
            this.oldLocation = file;
            this.newLocation = newNameWithExtension;
        }

        public bool Execute() {
            if (File.Exists(oldLocation)) {
                File.Move(oldLocation, newLocation);
            }
            VisPropertyBox.PropertyBoxManager.UpdateAllGrids(false);
            StudioBase.Explorer.UpdateExplorer();
            VisEngine.UpdateFormDirectories(null, null);
            return true;
        }

        public bool Undo() {
            if (File.Exists(newLocation)) {
                File.Move(newLocation, oldLocation);
            }
            VisPropertyBox.PropertyBoxManager.UpdateAllGrids(false);
            StudioBase.Explorer.UpdateExplorer();
            VisEngine.UpdateFormDirectories(null, null);
            return true;
        }
    }

    public class RemovePropertyCommand : ICommand {
        private string propName; // Don't save tplProp. It may become dangling as a result of TreeChange.
        private Template template;
        private string formula;

        public RemovePropertyCommand(Template tpl, string propName) {
            TplProperty tplProp;
            this.template = tpl;
            this.propName = propName;
            PropInfo propInfo; 
            tplProp = TplProperty.GetTemplateProperty(tpl, propName, out propInfo);
            this.formula = tplProp == null ? null : tplProp.Formula;
        }

        public bool Execute() {
            TplProperty tplProp;
            this.template = this.template.LookUp();
            if (template == null) return false;
            PropInfo propInfo;
            tplProp = TplProperty.GetTemplateProperty(template, propName, out propInfo);
            if (tplProp != null && VisEngine.DeleteTemplateProperty(template, tplProp)) {
                UndoManager.Refresh(template);
                TemplateHandler.SelectTemplate(template);
                VisPropertyBox.PropertyBoxManager.UpdateAllGrids(true);
                // StudioBase.Explorer.UpdateExplorer();
                return true;
            }
            return false;
        }

        public bool Undo() {
            this.template = this.template.LookUp();
            if (template == null) return false;
            if (VisEngine.CreateOrSetTemplateProperty(template, propName, this.formula) != null) {
                UndoManager.Refresh(template);
                TemplateHandler.SelectTemplate(template);
                VisPropertyBox.PropertyBoxManager.UpdateAllGrids(true);
                // StudioBase.Explorer.UpdateExplorer();
                return true;
            }
            return false;
        }
    }

    public class AddPropertyCommand : ICommand {
        private string propName; // Don't save tplProp. It may become dangling as a result of TreeChange.
        private string formula;
        private Template template;

        public AddPropertyCommand(Template template, string propName, string formula) {
            this.template = template;
            this.propName = propName;
            this.formula = formula;
        }

        public bool Execute() {
            TplProperty tplProp;
            this.template = this.template.LookUp();
            if (template == null) return false;
            this.propName = CreatePropertyName(template, propName); // If not unique, add an integer suffix.
            tplProp = VisEngine.CreateOrSetTemplateProperty(template, propName, formula);
            if (tplProp != null) {
                UndoManager.Refresh(template);
                TemplateHandler.SelectTemplate(template);
                VisPropertyBox.PropertyBoxManager.UpdateAllGrids(true);
                // StudioBase.Explorer.UpdateExplorer();
                return true;
            }
            return false;
        }

        public bool Undo() {
            TplProperty tplProp;
            this.template = this.template.LookUp();
            if (template == null) return false;
            PropInfo propInfo;
            tplProp = TplProperty.GetTemplateProperty(template, propName, out propInfo);
            if (tplProp != null && VisEngine.DeleteTemplateProperty(template, tplProp)) {
                UndoManager.Refresh(template);
                TemplateHandler.SelectTemplate(template);
                VisPropertyBox.PropertyBoxManager.UpdateAllGrids(true);
                // StudioBase.Explorer.UpdateExplorer();
                return true;
            }
            return false;
        }

        private string CreatePropertyName(Template template, string propName) { // If propName not unique, add an integer suffix.
            var newProperty = propName;
            for (int i = 1; true; i++) {
                if (!template.DesignProperties.ContainsKey(newProperty)) {
                    return newProperty;
                }
                newProperty = propName + i;
            }
        }
    }

    public class ChangePropertyNameCommand : ICommand {
        private TplProperty tplProp; // After the change, this is a dangling object. A new property has been created.
        private Template template;
        private string oldName;
        private string newName;

        public ChangePropertyNameCommand(TplProperty tplProp, string newName) {
            this.tplProp = tplProp;
            this.template = tplProp.tpl;
            this.oldName = tplProp.Name;
            this.newName = newName;
        }

        public bool Execute() {
            if (oldName.ToUpper() == "NAME" || oldName.ToUpper()=="TYPE")
                return false;
            TplProperty tplProp;
            this.template = this.template.LookUp();
            if (template == null) return false;
            PropInfo propInfo;
            tplProp = TplProperty.GetTemplateProperty(template, oldName, out propInfo);
            return ChangeName(tplProp, oldName, newName);
            // StudioBase.Explorer.UpdateExplorer();
        }

        public bool Undo() {
            PropInfo propInfo;
            this.template = this.template.LookUp();
            if (template == null) return false;
            tplProp = TplProperty.GetTemplateProperty(template, newName, out propInfo);
            //if (!VisEngine.DeleteTemplateProperty(property.tpl, oldName))
            return ChangeName(tplProp, newName, oldName);
            // StudioBase.Explorer.UpdateExplorer();
        }

        private bool ChangeName(TplProperty property, string key, string value) {
            Template tpl = property.tpl;
            if (VisEngine.RenameTemplateProperty(tpl, property, value)) {
                UndoManager.Refresh(tpl);
                VisPropertyBox.PropertyBoxManager.UpdateAllGrids(true);
                TemplateHandler.SelectTemplate(tpl);
                return true;
            }
            return false;
        }
    }

    public class ChangePropertyCommand : ICommand {
        private TplProperty property; // After the change, this may become a dangling object due to TreeChange.
        private string propName;
        private Template template;
        private string oldFormula;
        private string newFormula;

        public ChangePropertyCommand(TplProperty property, string oldFormula, string newFormula) {
            this.property = property;
            this.propName = property.Name;
            this.template = property.tpl;
            this.oldFormula = oldFormula;
            this.newFormula = newFormula;
        }

        public bool Execute() {
            //Debug.Print("ChangePropertyCommand old: " + oldFormula + "  new: " + newFormula);
            PropInfo propInfo;
            this.template = this.template.LookUp();
            if (template == null) return false;
            property = TplProperty.GetTemplateProperty(template, propName, out propInfo); // At redo, the tplProp may be dangling.
            if (property == null || ChangeProperty(property, newFormula) == null) return false;
            UndoManager.Refresh(template);
            VisPropertyBox.PropertyBoxManager.UpdateAllGrids(false);
            VisPropertyBox.PropertyBoxManager.UpdateRichText(property);
            if (template is FormTemplate && propName.ToUpper() == "NAME")
                StudioBase.Explorer.UpdateExplorer();
            // When called through text edit, there is no need to update. But when called through Redo, update is needed.
            return true;
        }

        public bool Undo() {
            //Debug.Print("Undo old: " + oldFormula + "  new: " + newFormula);
            PropInfo propInfo;
            this.template = this.template.LookUp();
            if (template == null) return false;
            property = TplProperty.GetTemplateProperty(property.tpl, propName, out propInfo);
            if (property == null || ChangeProperty(property, oldFormula) == null) return false;
            UndoManager.Refresh(template);
            VisPropertyBox.PropertyBoxManager.UpdateAllGrids(false);
            VisPropertyBox.PropertyBoxManager.UpdateRichText(property);
            if (template is FormTemplate && propName.ToUpper() == "NAME")
                StudioBase.Explorer.UpdateExplorer();
            return true;
        }

        private TplProperty ChangeProperty(TplProperty property, string value) {
            var template = property.tpl;
            var key = property.Name;
            return VisEngine.CreateOrSetTemplateProperty(template, key, value);
        }
    }

    public class ChangeTemplateTypeCommand : ICommand {
        private Template template;
        private string oldValue;
        private string newType;

        public ChangeTemplateTypeCommand(Template property, string oldType, string newType) {
            this.template = property;
            this.oldValue = oldType;
            this.newType = newType;
        }

        public bool Execute() {
            this.template = this.template.LookUp();
            if (template == null) return false;
            return ChangeTemplateType(template, newType);
            // StudioBase.Explorer.UpdateExplorer();
        }

        public bool Undo() {
            this.template = this.template.LookUp();
            if (template == null) return false;
            return ChangeTemplateType(template, oldValue);
            // StudioBase.Explorer.UpdateExplorer();
        }

        private bool ChangeTemplateType(Template template, string value) {
            if (VisEngine.ChangeTemplateType(template, value)) {
                UndoManager.Refresh(template);
                VisPropertyBox.PropertyBoxManager.UpdateAllGrids(true);
                return true;
            }
            return false;
        }
    }

    public class OpenFileCommand : ICommand { // Should be a simple method in Controller.OpenFile. No need to Undo.
        public static event EventHandler MapFileOpened; // Mess - call directly.
        private string file;

        public OpenFileCommand(string file) {
            this.file = file;
        }

        public bool Execute() {
            var name = Path.GetFileNameWithoutExtension(file);
            var extension = Path.GetExtension(file);
            var dummy = 0;

            if (extension.ToUpper().Equals(".VIS")) {
                var formTemplate = VisEngine.OpenForm(name, new Expression[] { }, out dummy);

                //if (formTemplate != null) // Don't select the form. The designer usually wants to do something else.
                //    TemplateHandler.SelectTemplate(formTemplate);
            } else if (extension.ToUpper().Equals(".VISM")) {
                VisEngine.CloseAllForms();
                StudioBase.datamap.Clean();
                if (!Parser.OpenMapFile(file)) // Will remove previous DataMap structures.
                    return false;
                StudioBase.Studio.Title = "Application: " + VisEngine.ApplicationDirectory + "\\" + VisEngine.MapFile + ".vism" + "     uVis: " + VisEngine.ProgramDirectory;
                MapFileOpened.Invoke(this, new EventArgs());
                var formTemplate = VisEngine.OpenForm(VisEngine.startUpForm, null, out dummy);
            }
            return true;
        }

        public bool Undo() {
            return true;
        }
    }

    public class CreateFileCommand : ICommand { // Should be a simple method. No ned for Undo. 
        private string file;

        public CreateFileCommand(string file) {
            this.file = file;
        }

        public bool Execute() {
            var extension = Path.GetExtension(file);
            var fromLocation = VisEngine.ProgramDirectory + "\\default." + extension;
            var toLocation = VisEngine.ApplicationDirectory + "\\" + file;

            File.Copy(fromLocation, toLocation);
            SystemItems.Refresh();
            // VisPropertyBox.PropertyBoxManager.UpdateAllGrids();
            StudioBase.Explorer.UpdateExplorer();
            return true;
        }

        public bool Undo() {
            var location = VisEngine.ApplicationDirectory + "\\" + file;
            File.Delete(location);
            // VisPropertyBox.PropertyBoxManager.UpdateAllGrids();
            StudioBase.Explorer.UpdateExplorer();
            return true;
        }
    }

    public class CopyFileCommand : ICommand {
        private string file;

        public CopyFileCommand(string file) {
            this.file = file;
        }

        public bool Execute() {
            var name = Path.GetFileName(file);
            var copyName = name + "_copy";
            var newFile = VisEngine.ApplicationDirectory + "\\" + copyName;
            File.Copy(file, newFile);
            SystemItems.Refresh();
            // VisPropertyBox.PropertyBoxManager.UpdateAllGrids();
            StudioBase.Explorer.UpdateExplorer();
            VisEngine.UpdateFormDirectories(null, null);
            return true;
        }

        public bool Undo() {
            var name = Path.GetFileName(file);
            var copyName = name + "_copy";
            var newFile = VisEngine.ApplicationDirectory + "\\" + copyName;

            if (File.Exists(newFile)) {
                File.Delete(newFile);
            }
            SystemItems.Refresh();
            // VisPropertyBox.PropertyBoxManager.UpdateAllGrids();
            StudioBase.Explorer.UpdateExplorer();
            VisEngine.UpdateFormDirectories(null, null);
            return true;
        }
    }

    #endregion
}
