﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Windows;
using System.Windows.Forms;
using System.Windows.Forms.Integration;
using VisTool;
using Point = System.Drawing.Point;
using UserControl = System.Windows.Controls.UserControl;

namespace uVisStudio.VisExplorer {
    public partial class VisExplorer : UserControl {
        private Controller controller = Controller.Instance;
        private TreeView treeView;
        private ContextMenu contextMenu;

        public VisExplorer() {
            StudioBase.Explorer = this;
            this.InitializeComponent();
            this.treeView = new TreeView();
            var formsHost = new WindowsFormsHost { Child = this.treeView };
            this.Container.Content = formsHost;

            var imageList = new ImageList();
            imageList.Images.Add(Properties.Resources.NewFolderHS);
            imageList.Images.Add(Properties.Resources.openfolderHS);
            imageList.Images.Add(Properties.Resources.DocumentHS);
            this.treeView.ImageList = imageList;

            this.treeView.ImageIndex = 0;
            this.treeView.SelectedImageIndex = 1;

            this.treeView.NodeMouseDoubleClick += FileViewNodeMouseDoubleClick;
            this.treeView.NodeMouseDoubleClick += TemplateViewNodeMouseDoubleClick;
            this.treeView.KeyPress += FileViewKeyDown;
            this.treeView.KeyPress += TemplateViewKeyDown;

            this.contextMenu = new ContextMenu();
            this.contextMenu.Popup += ContextMenuPopup;
            this.treeView.ContextMenu = this.contextMenu;
            this.treeView.MouseDown += TreeViewMouseDown;

            VisExplorerMenu.DirectoryOpened += OnDirectoryOpened;
            //VisEngine.ShowForm += OnShowForm;
        }

        public void UpdateExplorer() {
            ManageFiles();
            ManageTemplates();
            VisEngine.UpdateFormDirectories(null, null);
        }

        private void OnShowForm(object sender, DesignModeEventArgs e) // Not used
        {
            var form = e.form;
            var formInstance = (VisForm)form.Tag;

            formInstance.FormClosed -= OnFormsClosed;
            formInstance.FormClosed += OnFormsClosed;

            ManageFiles();
            ManageTemplates();
        }

        private void OnDirectoryOpened(object sender, EventArgs e) {
            ManageFiles();
            ManageTemplates();

            this.treeView.Nodes[0].Expand();
        }

        private void OnFormsClosed(object sender, EventArgs eventArgs) {
            ManageFiles();
            ManageTemplates();
        }

        private void ManageFiles() {
            List<String> files = Directory.GetFiles(VisEngine.ApplicationDirectory).ToList();
            files.Sort((a, b) => a.CompareTo(b));
            var uVisFiles = files.Where(x => x.ToUpper().EndsWith(".VIS") || x.ToUpper().EndsWith(".VISM"));
            var directoryName = new FileInfo(VisEngine.ApplicationDirectory + "\\").Directory.Name;

            if (this.treeView.Nodes.Count == 0) this.treeView.Nodes.Add(new ProjectViewItem());

            var project = (ProjectViewItem)this.treeView.Nodes[0];
            if (project.Directory != VisEngine.ApplicationDirectory) {
                project.Text = directoryName;
                project.Directory = VisEngine.ApplicationDirectory;
                project.ImageIndex = 0;
                project.SelectedImageIndex = 1;
            }

            project.Nodes.Clear();
            //for (int i = project.Nodes.Count - 1; i >= 0; i--) {
            //    var fileViewItem = (FileViewItem)project.Nodes[i];
            //    if (!uVisFiles.Contains(fileViewItem.File)) {
            //        project.Nodes.Remove(fileViewItem);
            //    }
            //}

            foreach (var uVisFile in uVisFiles) {
                var fileViewItem = FindFileViewItem(uVisFile, project);
                if (fileViewItem == null) {
                    fileViewItem = new FileViewItem();
                    fileViewItem.Text = Path.GetFileName(uVisFile);
                    fileViewItem.File = uVisFile;
                    fileViewItem.ImageIndex = 2;
                    fileViewItem.SelectedImageIndex = 2;
                    project.Nodes.Add(fileViewItem);
                }
            }
        }

        private void ManageTemplates() {
            //if (this.treeView.Nodes.Count == 0) return;

            //var project = (ProjectViewItem)this.treeView.Nodes[0];
            //foreach (var node in project.Nodes) {
            //    var item = (FileViewItem)node;
            //    if (!item.File.ToUpper().EndsWith(".VIS")) continue;

            //    string formName = Path.GetFileNameWithoutExtension(item.File);
            //    int cloneID = 0;
            //    FormTemplate formTemplate = VisEngine.GetFormTemplate(formName, ref cloneID);
            //    if (formTemplate != null && formTemplate.Templates.Count == item.Templates.Count) {
            //        item.RemoveTemplates();
            //        foreach (var pair in formTemplate.Templates) {
            //            var template = pair.Value;
            //            item.AddTemplate(template);
            //        }
            //        item.Templates.Sort((a, b) => a.Name.CompareTo(b.Name));
            //    } else {
            //        item.RemoveTemplates();
            //    }
            //}
        }

        private FileViewItem FindFileViewItem(string file, ProjectViewItem project) {
            foreach (var item in project.Nodes) {
                var fileViewItem = (FileViewItem)item;
                var name = fileViewItem.File.ToUpper();
                if (name.Equals(file.ToUpper())) {
                    return fileViewItem;
                }
            }

            return null;
        }

        private void TreeViewMouseDown(object sender, MouseEventArgs e) {
            if (e.Button == MouseButtons.Right) {
                var node = this.treeView.GetNodeAt(new Point(e.X, e.Y));
                if (node != null) {
                    this.treeView.SelectedNode = node;
                }
                contextMenu.Show(this.treeView, new Point(e.X, e.Y));
            }
        }

        private void ContextMenuPopup(object sender, EventArgs e) {
            this.contextMenu.MenuItems.Clear();

            var rename = new MenuItem("Rename", RenameItemClick);
            //var separator = new MenuItem("-");
            //var edit = new MenuItem("Edit", EditItemClick);
            var copy = new MenuItem("Copy", CopyItemClick);
            var delete = new MenuItem("Delete", DeleteItemClick);

            this.contextMenu.MenuItems.Add(rename);
            //this.contextMenu.MenuItems.Add(separator);
            //this.contextMenu.MenuItems.Add(edit);
            this.contextMenu.MenuItems.Add(copy);
            this.contextMenu.MenuItems.Add(delete);
        }

        private void FileViewKeyDown(object sender, KeyPressEventArgs e) {
            if (e.KeyChar == (char)Keys.Enter) {
                var selectedNode = this.treeView.SelectedNode as FileViewItem;
                if (selectedNode != null) {
                    var file = selectedNode.File;
                    this.controller.OpenFile(file);
                    treeView.Focus();
                    e.Handled = true;
                }
            }
        }

        private void TemplateViewKeyDown(object sender, KeyPressEventArgs e) {
            if (e.KeyChar == (char)Keys.Enter) {
                var selectedNode = this.treeView.SelectedNode as TemplateViewItem;
                if (selectedNode != null) {
                    var template = selectedNode.Template;
                    TemplateHandler.SelectTemplate(template);
                    treeView.Focus();
                    e.Handled = true;
                }
            }
        }

        private void FileViewNodeMouseDoubleClick(object sender, TreeNodeMouseClickEventArgs e) {
            var selectedNode = this.treeView.SelectedNode as FileViewItem;
            if (selectedNode != null) {
                var file = selectedNode.File;
                this.controller.OpenFile(file);
                treeView.Focus();
            }
        }

        private void TemplateViewNodeMouseDoubleClick(object sender, TreeNodeMouseClickEventArgs e) {
            var selectedNode = this.treeView.SelectedNode as TemplateViewItem;
            if (selectedNode != null) {
                var template = selectedNode.Template;
                TemplateHandler.SelectTemplate(template);
                treeView.Focus();
            }
        }

        private void RenameItemClick(object sender, EventArgs eventArgs) {
            var fileItem = this.treeView.SelectedNode as FileViewItem;
            if (fileItem == null) return;

            var renameFileDialog = new RenameFileDialog();
            renameFileDialog.Title = "Rename file";

            var screen = Screen.PrimaryScreen;
            var screenHeight = screen.Bounds.Height;
            var screenWidth = screen.Bounds.Width;
            renameFileDialog.Left = (screenWidth - renameFileDialog.Width) / 2;
            renameFileDialog.Top = (screenHeight - renameFileDialog.Height) / 2;

            renameFileDialog.ResizeMode = ResizeMode.NoResize;
            renameFileDialog.SetFile(fileItem.File);
            renameFileDialog.ShowDialog();
        }

        //private void EditItemClick(object sender, EventArgs eventArgs)
        //{
        //    var fileViewItem = this.treeView.SelectedNode as FileViewItem;
        //    if (fileViewItem != null)
        //    {
        //        this.controller.EditFile(fileViewItem.File);
        //    }
        //}

        private void CopyItemClick(object sender, EventArgs eventArgs) {
            var fileItem = this.treeView.SelectedNode as FileViewItem;
            if (fileItem != null) {
                this.controller.CopyFile(fileItem.File);
                return;
            }
            var templateItem = this.treeView.SelectedNode as TemplateViewItem;
            if (templateItem != null) {
                //this.controller.CopyTemplate(templateItem.Template);
                return;
            }
        }

        private void DeleteItemClick(object sender, EventArgs eventArgs) {
            var fileItem = this.treeView.SelectedNode as FileViewItem;
            if (fileItem != null) {
                this.controller.DeleteFile(fileItem.File);
                return;
            }
            var templateItem = this.treeView.SelectedNode as TemplateViewItem;
            if (templateItem != null) {
                var tpl = templateItem.Template;
                this.controller.DeleteTemplates(new List<Template>() { tpl });
                return;
            }
        }
    }

    public class TemplateViewItem : System.Windows.Forms.TreeNode {
        public Template Template { get; set; }
    }

    public class FileViewItem : System.Windows.Forms.TreeNode {
        public List<Template> Templates = new List<Template>();
        public string File { get; set; }

        public bool ContainsTemplate(Template template) {
            return Templates.Contains(template);
        }

        public void AddTemplate(Template template) {
            this.Templates.Add(template);
            var templateViewItem = new TemplateViewItem();
            templateViewItem.ImageIndex = 2;
            templateViewItem.SelectedImageIndex = 2;
            templateViewItem.Text = template.Name;
            templateViewItem.Template = template;
            this.Nodes.Add(templateViewItem);
        }

        public void RemoveTemplate(Template template) {
            this.Templates.Remove(template);
            for (int i = 0; i < Nodes.Count; i++) {
                var node = (TemplateViewItem)Nodes[i];
                if (node.Template.Equals(template)) {
                    Nodes.Remove(node);
                }
            }
        }

        public void RemoveTemplates() {
            this.Templates.Clear();
            this.Nodes.Clear();
        }
    }

    public class ProjectViewItem : System.Windows.Forms.TreeNode {
        public string Directory { get; set; }
    }
}